/*---
  Flow*: A Verification Tool for Cyber-Physical Systems.
  Authors: Xin Chen, Sriram Sankaranarayanan, and Erika Abraham.
  Email: Xin Chen <chenxin415@gmail.com> if you have questions or comments.

  The code is released as is under the GNU General Public License (GPL).
---*/

#ifndef CONTINUOUS_H_
#define CONTINUOUS_H_

#include "TaylorModel.h"
#include "Geometry.h"

namespace libflowstar
{
class Flowpipe;
class ContinuousSystem;
}

namespace flowstar
{

class Flowpipe						// A flowpipe is represented by a composition of two Taylor models. The left Taylor model is the preconditioning part.
{
public:
	TaylorModelVec tmvPre;			// preconditioning Taylor model
	TaylorModelVec tmv;
	std::vector<Interval> domain;	// domain of tmv, the first variable is t

public:
	Flowpipe();
	Flowpipe(const TaylorModelVec & tmvPre_input, const TaylorModelVec & tmv_input, const std::vector<Interval> & domain_input);
	Flowpipe(const std::vector<Interval> & box, const Interval & I);								// represent a box
	Flowpipe(const TaylorModelVec & tmv_input, const std::vector<Interval> & domain_input);		// construct a flowpipe from a Taylor model
	Flowpipe(const Flowpipe & flowpipe);
	virtual ~Flowpipe();

	void clear();
	void dump(FILE *fp, const std::vector<std::string> & stateVarNames, const std::vector<std::string> & tmVarNames, const Interval & cutoff_threshold) const;
	void dump_normal(FILE *fp, const std::vector<std::string> & stateVarNames, const std::vector<std::string> & tmVarNames, std::vector<Interval> & step_exp_table, const Interval & cutoff_threshold) const;
	void composition(TaylorModelVec & result, const Interval & cutoff_threshold) const;	// apply the preconditioning part to the Taylor model
	void composition_normal(TaylorModelVec & result, const std::vector<Interval> & step_exp_table, const Interval & cutoff_threshold) const;

	void intEval(std::vector<Interval> & result, const Interval & cutoff_threshold) const;
	void intEvalNormal(std::vector<Interval> & result, const std::vector<Interval> & step_exp_table, const Interval & cutoff_threshold) const;
	void intEvalNormal(std::vector<Interval> & result, const std::vector<int> & varIDs, const std::vector<Interval> & step_exp_table, const int order, const Interval & cutoff_threshold) const;

	void normalize();


	// Taylor model integration by only using Picard operation
	// fixed step sizes and orders
	int advance_picard(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;
	int advance_picard(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;

	// adaptive step sizes and fixed orders
	int advance_picard(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;
	int advance_picard(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;

	// adaptive orders and fixed step sizes
	int advance_picard(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, int & order, const int maxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;
	int advance_picard(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, std::vector<int> & orders, const int localMaxOrder, const std::vector<int> & maxOrders, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;


	// fast integration scheme for low-degree ODEs
	// fixed step sizes and orders
	int advance_low_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & taylorExpansion, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;
	int advance_low_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & taylorExpansion, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const std::vector<int> & orders, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;

	// adaptive step sizes and fixed orders
	int advance_low_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & taylorExpansion, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;
	int advance_low_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & taylorExpansion, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;

	// adaptive orders and fixed step sizes
	int advance_low_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const std::vector<HornerForm> & taylorExpansion, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, int & order, const int maxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;
	int advance_low_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const std::vector<HornerForm> & taylorExpansion, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, std::vector<int> & orders, const std::vector<int> & maxOrders, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;

	// integration scheme for high-degree ODEs
	// fixed step sizes and orders
	int advance_high_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;
	int advance_high_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;

	// adaptive step sizes and fixed orders
	int advance_high_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;
	int advance_high_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;

	// adaptive orders and fixed step sizes
	int advance_high_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, int & order, const int maxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;
	int advance_high_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, std::vector<int> & orders, const int localMaxOrder, const std::vector<int> & maxOrders, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;


	// integration scheme for non-polynomial ODEs (using Taylor approximations)
	// fixed step sizes and orders
	int advance_non_polynomial_taylor(Flowpipe & result, const std::vector<std::string> & strOde, const std::vector<std::string> & strOde_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;
	int advance_non_polynomial_taylor(Flowpipe & result, const std::vector<std::string> & strOde, const std::vector<std::string> & strOde_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;

	// adaptive step sizes and fixed orders
	int advance_non_polynomial_taylor(Flowpipe & result, const std::vector<std::string> & strOde, const std::vector<std::string> & strOde_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;
	int advance_non_polynomial_taylor(Flowpipe & result, const std::vector<std::string> & strOde, const std::vector<std::string> & strOde_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;

	// adaptive orders and fixed step sizes
	int advance_non_polynomial_taylor(Flowpipe & result, const std::vector<std::string> & strOde, const std::vector<std::string> & strOde_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, int & order, const int maxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;
	int advance_non_polynomial_taylor(Flowpipe & result, const std::vector<std::string> & strOde, const std::vector<std::string> & strOde_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, std::vector<int> & orders, const int localMaxOrder, const std::vector<int> & maxOrders, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const;

	// Taylor model flowpipes with the remainder represented by support functions
	// fixed orders and step sizes
	int advance_picard_symbolic_remainder(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const Interval & cutoff_threshold, Interval_matrix & trans_back, std::vector<Interval_matrix> & J, std::vector<Interval_matrix> & Phi_L) const;



	// flowpipes for time-triggered jumps
	int advance_picard_tt(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<Interval> & ode_rem, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const Interval & cutoff_threshold) const;




	int advance_picard_hybridization_with_decomposition(Flowpipe & result, TaylorModelVec & x0_output, std::vector<Interval> & boundOfx0, const std::vector<HornerForm> & ode, const TaylorModelVec & c0, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const std::vector<std::vector<bool> > & substitution, const int order, const std::vector<Interval> & estimation, const Interval & cutoff_threshold, Interval_matrix & trans_back, std::vector<Interval_matrix> & J, std::vector<Interval_matrix> & Phi_L, const bool bsimp) const;
	int refinement_picard_hybridization_with_decomposition(const std::vector<HornerForm> & ode, const TaylorModelVec & x0, const std::vector<std::vector<bool> > & substitution, std::vector<Interval> & step_exp_table, const int order, const std::vector<Interval> & estimation, const Interval & cutoff_threshold);

//	int refinement_picard_hybridization_with_decomposition(const TaylorModelVec & tmvOde, const std::vector<Interval> & hybrid_domain, const std::vector<int> & hybrid_dim, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const Interval & cutoff_threshold);

	int advance_picard_suppfunc_remainder_hyb2(Flowpipe & result, const std::vector<Interval> & boundOfx0, const TaylorModelVec & ode, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const Interval & cutoff_threshold, const TaylorModelVec & initialSet, std::vector<Interval_matrix> & remainders, std::vector<Interval_matrix> & trans_on_local_initial) const;

	// flowpipe construction with dynamics refinement
	int advance_picard_dyn_ref(Flowpipe & result, std::vector<Interval> & range, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const Interval & cutoff_threshold) const;

	Flowpipe & operator = (const Flowpipe & flowpipe);

	friend class ContinuousSystem;
	friend class ContinuousReachability;
	friend class HybridSystem;
	friend class HybridReachability;
	friend class libflowstar::Flowpipe;
	friend class libflowstar::ContinuousSystem;
};

class ContinuousSystem
{
protected:
	TaylorModelVec tmvOde;
	TaylorModelVec tmvOde_centered;
	std::vector<HornerForm> hfOde;		// a Horner form of the ode
	std::vector<HornerForm> hfOde_centered;
	Flowpipe initialSet;			// the initial set
	std::vector<std::string> strOde;
	std::vector<std::string> strOde_centered;

public:
	ContinuousSystem();
	ContinuousSystem(const TaylorModelVec & ode_input, const Flowpipe & initialSet_input);
	ContinuousSystem(const std::vector<std::string> & strOde_input, const Flowpipe & initialSet_input);
	ContinuousSystem(const ContinuousSystem & system);
	virtual ~ContinuousSystem();

	// efficient integration method for linear ODEs
	// since we can always compute a safe remainder, adaptive techniques are not needed
	// since the size of a Taylor series is linear w.r.t. the number of state variables, we may simply use uniform orders
	void reach_linear(std::list<Flowpipe> & results, const double step, const double time, const int order, const bool bPrint, const Interval & cutoff_threshold) const;

	// only use Picard operation
	// fixed step sizes and orders
	void reach_picard(std::list<Flowpipe> & results, const double step, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;
	void reach_picard(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;

	// adaptive step sizes and fixed orders
	void reach_picard(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;
	void reach_picard(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;

	// adaptive orders and fixed step sizes
	void reach_picard(std::list<Flowpipe> & results, const double step, const double time, const int order, const int maxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;
	void reach_picard(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const std::vector<int> & maxOrders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;


	// for low-degree ODEs
	// fixed step sizes and orders
	void reach_low_degree(std::list<Flowpipe> & results, const double step, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;
	void reach_low_degree(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;

	// adaptive step sizes and fixed orders
	void reach_low_degree(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;
	void reach_low_degree(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;

	// adaptive orders and fixed step sizes
	void reach_low_degree(std::list<Flowpipe> & results, const double step, const double time, const int order, const int maxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;
	void reach_low_degree(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const std::vector<int> & maxOrders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;

	// for high-degree ODEs
	// fixed step sizes and orders
	void reach_high_degree(std::list<Flowpipe> & results, const double step, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;
	void reach_high_degree(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;

	// adaptive step sizes and fixed orders
	void reach_high_degree(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;
	void reach_high_degree(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;

	// adaptive orders and fixed step sizes
	void reach_high_degree(std::list<Flowpipe> & results, const double step, const double time, const int order, const int maxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;
	void reach_high_degree(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const std::vector<int> & maxOrders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;

	// for non-polynomial ODEs (using Taylor approximations)
	// fixed step sizes and orders
	void reach_non_polynomial_taylor(std::list<Flowpipe> & results, const double step, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;
	void reach_non_polynomial_taylor(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;

	// adaptive step sizes and fixed orders
	void reach_non_polynomial_taylor(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;
	void reach_non_polynomial_taylor(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;

	// adaptive orders and fixed step sizes
	void reach_non_polynomial_taylor(std::list<Flowpipe> & results, const double step, const double time, const int order, const int maxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;
	void reach_non_polynomial_taylor(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const std::vector<int> & maxOrders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const;

	// Taylor model flowpipes with the remainder represented by support functions
	// fixed orders and step sizes
	void reach_picard_symbolic_remainder(std::list<Flowpipe> & results, const double step, const double time, const int order, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold, const int N) const;

	void reach_picard_hybridization_with_decomposition(std::list<Flowpipe> & results, const double step, const double time, const int order, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold, const std::vector<std::vector<bool> > & substitution, const std::vector<int> & vars2sub, const std::vector<std::vector<int> > & subrule, const int M, const double sigma) const;





	void reach_picard_dynamical_hybridization2(std::list<Flowpipe> & results, const double step, const double time, const int order, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold, const std::vector<int> & vars2sub, const std::vector<std::vector<int> > & subrule, const int N, const double epsilon, const int M) const;

	void reach_linear_dynamical_hybridization(std::vector<TaylorModelVec> & flowpipes, std::vector<Interval> & domain, const std::vector<int> & varIDs, const double step, const double epsilon, const double time, const int order, const bool bPrint, const Interval & cutoff_threshold, const int N) const;


	// flowpipe construction with dynamics refinement
	void reach_picard_dyn_ref(std::list<Flowpipe> & results, const std::vector<int> & varIDs, const double step, const double epsilon, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const Interval & cutoff_threshold) const;

	ContinuousSystem & operator = (const ContinuousSystem & system);

	friend class ContinuousReachability;
};


class ContinuousReachability		// The reachability analysis of continuous systems
{
public:
	ContinuousSystem system;		// the continuous system
	double step;					// the step size used in the reachability analysis
	double time;					// the time horizon for the reachability analysis
	int precondition;				// the preconditioning technique
	std::vector<int> outputAxes;			// the output axes
	int plotSetting;
	int plotFormat;
	int numSections;				// the number of sections in each dimension

	int orderType;
	bool bAdaptiveSteps;
	bool bAdaptiveOrders;

	std::vector<Interval> estimation;	// the remainder estimation for varying time step
	double miniStep;				// the minimum step size
	std::vector<int> orders;				// the order(s)
	std::vector<int> maxOrders;			// the maximum orders
	int globalMaxOrder;

	bool bPrint;
	bool bSafetyChecking;

	int integrationScheme;

	Interval cutoff_threshold;

	std::list<Flowpipe> flowpipes;
	std::list<TaylorModelVec> flowpipesCompo;
	std::list<std::vector<Interval> > domains;

	std::vector<PolynomialConstraint> unsafeSet;

	std::map<std::string,int> stateVarTab;
	std::vector<std::string> stateVarNames;

	std::map<std::string,int> tmVarTab;
	std::vector<std::string> tmVarNames;

	std::map<std::string,int> parTab;
	std::vector<std::string> parNames;
	std::vector<Interval> parRanges;

	char outputFileName[NAME_SIZE];

public:
	ContinuousReachability();
	~ContinuousReachability();

	void dump(FILE *fp) const;

	void run();
	void composition();
	int safetyChecking() const;
	unsigned long numOfFlowpipes() const;

	void dump_potential_counterexample(FILE *fp, const std::list<TaylorModelVec> & flowpipes, const std::list<std::vector<Interval> > & domains, const std::list<Interval> & globalTimes) const;

	void plot_2D() const;

	void plot_2D_GNUPLOT(FILE *fp) const;
	void plot_2D_interval_GNUPLOT(FILE *fp) const;
	void plot_2D_octagon_GNUPLOT(FILE *fp) const;
	void plot_2D_grid_GNUPLOT(FILE *fp) const;

	void plot_2D_MATLAB(FILE *fp) const;
	void plot_2D_interval_MATLAB(FILE *fp) const;
	void plot_2D_octagon_MATLAB(FILE *fp) const;
	void plot_2D_grid_MATLAB(FILE *fp) const;

	bool declareStateVar(const std::string & vName);
	int getIDForStateVar(const std::string & vName) const;
	bool getStateVarName(std::string & vName, const int id) const;

	bool declareTMVar(const std::string & vName);
	int getIDForTMVar(const std::string & vName) const;
	bool getTMVarName(std::string & vName, const int id) const;

	bool declarePar(const std::string & pName, const Interval & range);
	int getIDForPar(const std::string & pName) const;
	bool getParName(std::string & pName, const int id) const;
	bool getRangeForPar(Interval & range, const std::string & pName) const;
};

void computeTaylorExpansion(TaylorModelVec & result, const TaylorModelVec & first_order_deriv, const TaylorModelVec & ode, const int order, const Interval & cutoff_threshold);
void computeTaylorExpansion(TaylorModelVec & result, const TaylorModelVec & first_order_deriv, const TaylorModelVec & ode, const std::vector<int> & orders, const Interval & cutoff_threshold);

void construct_step_exp_table(std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const int order);
void construct_step_exp_table(std::vector<Interval> & step_exp_table, const Interval & step, const int order);

void preconditionQR(Matrix & result, const TaylorModelVec & tmv, const int rangeDim, const int domainDim);

Interval rho(const TaylorModelVec & tmv, const std::vector<Interval> & l, const std::vector<Interval> & domain);
Interval rhoNormal(const TaylorModelVec & tmv, const std::vector<Interval> & l, const std::vector<Interval> & step_end_exp_table);

Interval rho(const TaylorModelVec & tmv, const RowVector & l, const std::vector<Interval> & domain);
Interval rhoNormal(const TaylorModelVec & tmv, const RowVector & l, const std::vector<Interval> & step_end_exp_table);

void templatePolyhedron(Polyhedron & result, const TaylorModelVec & tmv, const std::vector<Interval> & domain);
void templatePolyhedronNormal(Polyhedron & result, const TaylorModelVec & tmv, std::vector<Interval> & step_end_exp_table);


// domain contraction by using interval arithmetic
int contract_interval_arithmetic(TaylorModelVec & flowpipe, std::vector<Interval> & domain, const std::vector<PolynomialConstraint> & pcs, std::vector<bool> & boundary_intersected, const Interval & cutoff_threshold);

//int contract_interval_arithmetic(TaylorModel & tm, std::vector<Interval> & domain, const Interval & intMax, const Interval & intMin, const Interval & cutoff_threshold);

int contract_interval_arithmetic(TaylorModel & tm_mid, std::vector<Interval> & domain_mid,
		TaylorModel & tm_max, std::vector<Interval> & domain_max, TaylorModel & tm_min, std::vector<Interval> & domain_min,
		const TaylorModel & tm, const std::vector<Interval> & domain, const Interval & intMax, const Interval & intMin, const Interval & cutoff_threshold);


int contract_interval_arithmetic(TaylorModelVec & flowpipe, std::vector<Interval> & domain, const Polyhedron & inv, std::vector<bool> & boundary_intersected);

int contract_remainder(const std::vector<Interval> & polyRange, std::vector<Interval> & remainders, const std::vector<HornerForm> & hfs, const std::vector<Interval> & b);

void gridBox(std::list<std::vector<Interval> > & grids, const std::vector<Interval> & box, const int num);

void exp_int_mat(Interval_matrix & result_ts, Interval_matrix & result_rem, const Interval_matrix & A, const int order);
void int_exp_int_mat(Interval_matrix & result_ts, Interval_matrix & result_rem, const Interval_matrix & A, const double step, const int order);

void compute_int_mat_pow(std::vector<Interval_matrix> & result, const Interval_matrix & A, const int order);

void enlarge(std::vector<Interval> & box, const double e);

}


#endif /* CONTINUOUS_H_ */
