
#include "Continuous_lib.h"

using namespace libflowstar;

Flowpipe::Flowpipe() : flowstar::Flowpipe()
{
}

Flowpipe::Flowpipe(const Assignment<Interval> & box)
{
	int rangeDim = box.size() - 1;
	int domainDim = rangeDim + 1;
	flowstar::Interval intUnit(-1,1), intZero;

	flowstar::TaylorModelVec tmvCenter;
	std::vector<double> scalars;

	domain.push_back(intZero);		// time interval

	// normalize the domain to [-1,1]^n
	for(int i=0; i<rangeDim; ++i)
	{
		double midpoint = box.values[i+1].midpoint();
		flowstar::Interval intMid(midpoint);
		flowstar::TaylorModel tmTemp(intMid, domainDim);
		tmvCenter.tms.push_back(tmTemp);

		flowstar::Interval intTemp = (flowstar::Interval)box.values[i+1];
		intTemp.sub_assign(midpoint);
		scalars.push_back( intTemp.sup() );
		domain.push_back(intUnit);
	}

	flowstar::Matrix coefficients_of_tmvPre(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		coefficients_of_tmvPre.set(scalars[i], i, i+1);
	}

	flowstar::TaylorModelVec tmvTemp(coefficients_of_tmvPre);
	tmvTemp.add(tmvPre, tmvCenter);

	flowstar::Matrix coefficients_of_tmv(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		coefficients_of_tmv.set(1, i, i+1);
	}

	flowstar::TaylorModelVec tmvTemp2(coefficients_of_tmv);
	tmv = tmvTemp2;
}

Flowpipe::Flowpipe(const std::vector<TaylorModel> & tm, const std::vector<Interval> & tmDomain)
{
	int rangeDim = tm.size();

	tmv.tms = *(std::vector<flowstar::TaylorModel> *)(&tm);
	domain = *(std::vector<flowstar::Interval> *)(&tmDomain);

	flowstar::Matrix coefficients_of_tmvPre(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		coefficients_of_tmvPre.set(1, i, i+1);
	}

	flowstar::TaylorModelVec tmvTemp(coefficients_of_tmvPre);
	tmvPre = tmvTemp;

	normalize();
}

Flowpipe::Flowpipe(const std::vector<TaylorModel> & tmVector, const Assignment<Interval> & tmVectorDomain)
{
	int rangeDim = tmVector.size();

	for(int i=0; i<rangeDim; ++i)
	{
		tmv.tms.push_back(tmVector[i]);
	}

	domain = *(std::vector<flowstar::Interval> *)(&tmVectorDomain.values);

	flowstar::Matrix coefficients_of_tmvPre(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		coefficients_of_tmvPre.set(1, i, i+1);
	}

	flowstar::TaylorModelVec tmvTemp(coefficients_of_tmvPre);
	tmvPre = tmvTemp;

	normalize();
}

Flowpipe::Flowpipe(const std::vector<TaylorModel> & tmVectorPre, const std::vector<TaylorModel> & tmVector, const Assignment<Interval> & tmVectorDomain)
{
	int n = tmVectorPre.size();

	for(int i=0; i<n; ++i)
	{
		tmvPre.tms.push_back(tmVectorPre[i]);
	}

	n = tmVector.size();

	for(int i=0; i<n; ++i)
	{
		tmv.tms.push_back(tmVector[i]);
	}

	domain = *(std::vector<flowstar::Interval> *)(&tmVectorDomain.values);
}

Flowpipe::Flowpipe(const flowstar::Flowpipe & flowpipe) : flowstar::Flowpipe(flowpipe)
{
}

Flowpipe::~Flowpipe()
{
}

void Flowpipe::getStepsize(Interval & step) const
{
	step = domain[0];
}

void Flowpipe::merge(const std::vector<TaylorModel> & tmFlowpipe, const std::vector<Interval> & domain, const ReachabilitySetting & rs)
{
	// we simply compute an interval union of the two flowpipes

	std::vector<flowstar::Interval> fp_range_1;
	intEvalNormal(fp_range_1, rs.step_exp_table, rs.cutoff_threshold);

	std::vector<flowstar::Interval> fp_range_2;
	for(int i=0; i<tmFlowpipe.size(); ++i)
	{
		flowstar::Interval I;
		tmFlowpipe[i].intEval(I, *(std::vector<flowstar::Interval> *)(&domain));
		fp_range_2.push_back(I);
	}

	for(int i=0; i<fp_range_1.size(); ++i)
	{
		fp_range_1[i].hull_assign(fp_range_2[i]);
	}

	flowstar::Interval zero;
	flowstar::Flowpipe newFp(fp_range_1, zero);
	*this = newFp;
}

void Flowpipe::rangeOfPoly(Assignment<Interval> & range, const ReachabilitySetting & rs) const
{
	flowstar::TaylorModelVec tmvTemp;
	composition_normal(tmvTemp, rs.step_exp_table, rs.cutoff_threshold);

	std::vector<flowstar::Interval> result;
	tmvTemp.polyRangeNormal(result, rs.step_exp_table);

	range.values.push_back(ZERO);
	for(int i=0; i<result.size(); ++i)
	{
		range.values.push_back((libflowstar::Interval)result[i]);
	}
}

void Flowpipe::range(Assignment<Interval> & range, const ReachabilitySetting & rs) const
{
	flowstar::TaylorModelVec tmvTemp;
	composition_normal(tmvTemp, rs.step_exp_table, rs.cutoff_threshold);

	std::vector<flowstar::Interval> result;
	tmvTemp.intEvalNormal(result, rs.step_exp_table);

	range.values.push_back(ZERO);
	for(int i=0; i<result.size(); ++i)
	{
		range.values.push_back((libflowstar::Interval)result[i]);
	}
}

void Flowpipe::output(FILE *fp, const Variables & stateVars, const Variables & tmVars, const Interval & cutoff) const
{
	flowstar::Flowpipe::dump(fp, stateVars.varNames, tmVars.varNames, cutoff);
}

void Flowpipe::toTaylorModel(std::vector<TaylorModel> & tmv, std::vector<Interval> & dom, const ReachabilitySetting & rs) const
{
	flowstar::TaylorModelVec tmVec;
	composition(tmVec, rs.cutoff_threshold);

	tmv = *(std::vector<libflowstar::TaylorModel> *)(&(tmVec.tms));
	dom = *(std::vector<libflowstar::Interval> *)(&domain);
}

void Flowpipe::toTaylorModelNormal(std::vector<TaylorModel> & tmv, std::vector<Interval> & dom, const ReachabilitySetting & rs) const
{
	flowstar::TaylorModelVec tmVec;
	composition_normal(tmVec, rs.step_exp_table, rs.cutoff_threshold);

	tmv = *(std::vector<libflowstar::TaylorModel> *)(&(tmVec.tms));
	dom = *(std::vector<libflowstar::Interval> *)(&domain);
}

bool Flowpipe::next_poly_ode_1(Flowpipe & nextFlowpipe, const ContinuousSystem & system, ReachabilitySetting & rs) const
{
	std::vector<flowstar::PolynomialConstraint> dummy_invariant;
	return advance_picard(nextFlowpipe, system.hfOde, system.hfOde_centered, rs.precondition, rs.step_exp_table, rs.step_end_exp_table, rs.orders[0], rs.estimation, dummy_invariant, rs.cutoff_threshold);
}

bool Flowpipe::next_flowpipe_linear(Assignment<TaylorModel> & tmFlowpipe, Assignment<Interval> & flowpipeDomain, const ContinuousSystem & system, ReachabilitySetting & rs) const
{
	std::vector<flowstar::PolynomialConstraint> dummy_invariant;
	Flowpipe nextFlowpipe;

	bool bvalid = advance_picard(nextFlowpipe, system.hfOde, system.hfOde_centered, rs.precondition, rs.step_exp_table, rs.step_end_exp_table, rs.orders[0], rs.estimation, dummy_invariant, rs.cutoff_threshold);

	if(bvalid)
	{
		flowstar::TaylorModelVec tmv, flowpipe;
		nextFlowpipe.tmvPre.evaluate_t(tmv, rs.step_end_exp_table);
		nextFlowpipe.tmvPre = tmv;
		nextFlowpipe.composition_normal(flowpipe, rs.step_end_exp_table, rs.cutoff_threshold);
		flowpipe.ctrunc_normal(rs.step_end_exp_table, 1);

		TaylorModel empty;
		tmFlowpipe[0] = empty;
		for(int i=0; i<flowpipe.tms.size(); ++i)
		{
			tmFlowpipe.values[i+1] = (libflowstar::TaylorModel)flowpipe.tms[i];
		}

		flowpipeDomain.values = *(std::vector<libflowstar::Interval> *)(&domain);

		return true;
	}
	else
	{
		return false;
	}
}






ODE::ODE(const Variables & vars)
{
	variables = vars;
	Polynomial polynomial;
	flowstar::HornerForm hf;
	flowstar::Interval I;

	for(int i=1; i<vars.varNames.size(); ++i)
	{
		// the derivatives of all variables are zero
		derivatives.push_back(polynomial);
		derivatives_hf.push_back(hf);
		remainders.push_back(I);
	}
}

ODE::ODE(const ODE & ode)
{
	variables = ode.variables;
	derivatives = ode.derivatives;
	derivatives_hf = ode.derivatives_hf;
	remainders = ode.remainders;
}

ODE::~ODE()
{
	variables.clear();
	derivatives.clear();
	derivatives_hf.clear();
	remainders.clear();
}

ODE & ODE::operator = (const ODE & ode)
{
	if(this == &ode)
		return *this;

	variables = ode.variables;
	derivatives = ode.derivatives;
	derivatives_hf = ode.derivatives_hf;
	remainders = ode.remainders;

	return *this;
}

bool ODE::assignDerivative(const std::string & varName, const Polynomial & polynomial)
{
	int id = variables.getIDForVar(varName);

	if(id == -1)
	{
		return false;
	}
	else
	{
		derivatives[id-1] = polynomial;

		flowstar::HornerForm hf;
		polynomial.toHornerForm(hf);
		derivatives_hf[id-1] = hf;

		return true;
	}
}

bool ODE::assignDerivative(const std::string & varName, const TaylorModel & taylormodel)
{
	int id = variables.getIDForVar(varName);

	if(id == -1)
	{
		return false;
	}
	else
	{
		derivatives[id-1] = taylormodel.expansion;
		remainders[id-1] = taylormodel.remainder;

		flowstar::HornerForm hf;
		taylormodel.expansion.toHornerForm(hf);
		derivatives_hf[id-1] = hf;

		return true;
	}
}

bool ODE::tmEval(TaylorModel & result, const std::vector<TaylorModel> & tms, const int order, const std::string & varName, ReachabilitySetting & rs) const
{
	int id = variables.getIDForVar(varName);

	if(id == -1)
	{
		return false;
	}
	else
	{
		flowstar::TaylorModelVec vars;
		vars.tms = *(std::vector<flowstar::TaylorModel> *)(&tms);

		std::vector<flowstar::Interval> varsPolyRange;
		vars.polyRangeNormal(varsPolyRange, rs.step_exp_table);
		derivatives_hf[id-1].insert_ctrunc_normal(*(flowstar::TaylorModel *)(&result), vars, varsPolyRange, rs.step_exp_table, tms.size()+1, order, rs.cutoff_threshold);
		return true;
	}
}

bool ODE::getHF(flowstar::HornerForm & hf, flowstar::Interval & rem, const std::string & varName) const
{
	int id = variables.getIDForVar(varName);

	if(id == -1)
	{
		return false;
	}
	else
	{
		hf = derivatives_hf[id-1];
		rem = remainders[id-1];

		return true;
	}
}

void ODE::output(FILE *fp) const
{
	for(int i=1; i<variables.size(); ++i)
	{
		std::string strPoly;
		derivatives[i-1].toString(strPoly, variables.varNames);
		fprintf(fp, "%s' = %s", variables.varNames[i].c_str(), strPoly.c_str());
		fprintf(fp, " + ");
		remainders[i-1].dump(fp);
		fprintf(fp, "\n");
	}

	fprintf(fp, "\n");
}





ContinuousSystem::ContinuousSystem() : flowstar::ContinuousSystem()
{
}

ContinuousSystem::~ContinuousSystem()
{
}

ContinuousSystem::ContinuousSystem(const Assignment<Polynomial> & ode, const Variables & stateVars)
{
	int rangeDim = ode.size() - 1;
	flowstar::Interval intZero;

	for(int i=0; i<rangeDim; ++i)
	{
		flowstar::TaylorModel tm((flowstar::Polynomial)ode.values[i+1]);
		tmvOde.tms.push_back(tm);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		flowstar::HornerForm hf;
		tmvOde.tms[i].expansion.toHornerForm(hf);
		hfOde.push_back(hf);
	}

	tmvOde_centered = tmvOde;
	tmvOde_centered.center_nc();

	for(int i=0; i<rangeDim; ++i)
	{
		flowstar::HornerForm hf;
		tmvOde_centered.tms[i].expansion.toHornerForm(hf);
		hfOde_centered.push_back(hf);
	}

	variables = stateVars;
}

ContinuousSystem::ContinuousSystem(const ContinuousSystem & system)
{
	tmvOde				= system.tmvOde;
	hfOde				= system.hfOde;
	initialSet			= system.initialSet;
	strOde				= system.strOde;
	tmvOde_centered		= system.tmvOde_centered;
	hfOde_centered		= system.hfOde_centered;
	strOde_centered		= system.strOde_centered;
	variables			= system.variables;
}

ContinuousSystem & ContinuousSystem::operator = (const ContinuousSystem & system)
{
	if(this == &system)
		return *this;

	tmvOde				= system.tmvOde;
	hfOde				= system.hfOde;
	initialSet			= system.initialSet;
	strOde				= system.strOde;
	tmvOde_centered		= system.tmvOde_centered;
	hfOde_centered		= system.hfOde_centered;
	strOde_centered		= system.strOde_centered;
	variables			= system.variables;

	return *this;
}

void ContinuousSystem::output(FILE *fp) const
{
	fprintf(fp, "\nSystem dynamics:\n");

	for(int i=1; i<variables.size(); ++i)
	{
		std::string vName;
		variables.getVarName(vName, i);

		std::string strPoly;
		tmvOde.tms[i-1].expansion.toString(strPoly, variables.varNames);
		fprintf(fp, "%s' = %s\n", vName.c_str(), strPoly.c_str());
	}

	fprintf(fp, "\n");
}

bool ContinuousSystem::setInit(const Flowpipe & flowpipe)
{
	if(variables.size()-1 != flowpipe.tmvPre.tms.size())
	{
		printf("Dimension of the initial set do not match the dimension of the system.\n");
		return false;
	}

	initialSet = (flowstar::Flowpipe)flowpipe;
	return true;
}

void ContinuousSystem::reach_linear(std::list<Flowpipe> & result, const ReachabilitySetting & rs) const
{
	flowstar::ContinuousSystem::reach_linear(*(std::list<flowstar::Flowpipe> *)(&result), rs.step, rs.time, rs.orders[0], rs.bPrint, rs.cutoff_threshold);
}

void ContinuousSystem::reach_poly_1(std::list<Flowpipe> & result, const ReachabilitySetting & rs) const
{
	std::vector<std::string> stateVarNames;

	for(int i=1; i<variables.size(); ++i)
	{
		stateVarNames.push_back(variables.varNames[i]);
	}

	switch(rs.orderType)
	{
	case UNIFORM:
		if(rs.bAdaptiveSteps)
		{
			reach_picard(*(std::list<flowstar::Flowpipe> *)(&result), rs.step, rs.miniStep, rs.time, rs.orders[0], rs.precondition, rs.estimation, rs.bPrint, stateVarNames, rs.cutoff_threshold);
		}
		else if(rs.bAdaptiveOrders)
		{
			reach_picard(*(std::list<flowstar::Flowpipe> *)(&result), rs.step, rs.time, rs.orders[0], rs.maxOrders[0], rs.precondition, rs.estimation, rs.bPrint, stateVarNames, rs.cutoff_threshold);
		}
		else
		{
			reach_picard(*(std::list<flowstar::Flowpipe> *)(&result), rs.step, rs.time, rs.orders[0], rs.precondition, rs.estimation, rs.bPrint, stateVarNames, rs.cutoff_threshold);
		}
		break;
	case MULTI:
		if(rs.bAdaptiveSteps)
		{
			reach_picard(*(std::list<flowstar::Flowpipe> *)(&result), rs.step, rs.miniStep, rs.time, rs.orders, rs.globalMaxOrder, rs.precondition, rs.estimation, rs.bPrint, stateVarNames, rs.cutoff_threshold);
		}
		else if(rs.bAdaptiveOrders)
		{
			reach_picard(*(std::list<flowstar::Flowpipe> *)(&result), rs.step, rs.time, rs.orders, rs.maxOrders, rs.globalMaxOrder, rs.precondition, rs.estimation, rs.bPrint, stateVarNames, rs.cutoff_threshold);
		}
		else
		{
			reach_picard(*(std::list<flowstar::Flowpipe> *)(&result), rs.step, rs.time, rs.orders, rs.globalMaxOrder, rs.precondition, rs.estimation, rs.bPrint, stateVarNames, rs.cutoff_threshold);
		}
		break;
	}
}

void ContinuousSystem::reach_linear_hybridization(std::vector<flowstar::TaylorModelVec> & tmFlowpipes, std::vector<Interval> & domain, const ReachabilitySetting & rs, const std::vector<int> & varIDs, const double epsilon, const int N) const
{
	int rangeDim = tmvOde.tms.size();

	flowstar::ContinuousSystem::reach_linear_dynamical_hybridization(tmFlowpipes, *(std::vector<flowstar::Interval> *)(&domain), varIDs, rs.step, epsilon, rs.time, rs.orders[0], rs.bPrint, rs.cutoff_threshold, N);
}

void ContinuousSystem::reach_symbolic_remainder(std::list<Flowpipe> & result, const ReachabilitySetting & rs, const int N) const
{
	std::vector<std::string> stateVarNames;

	for(int i=1; i<variables.size(); ++i)
	{
		stateVarNames.push_back(variables.varNames[i]);
	}

	flowstar::ContinuousSystem::reach_picard_symbolic_remainder(*(std::list<flowstar::Flowpipe> *)(&result), rs.step, rs.time, rs.orders[0], rs.estimation, rs.bPrint, stateVarNames, rs.cutoff_threshold, N);
}

namespace libflowstar
{

void plot_2D_box_gnuplot(FILE *fp, const std::string & epsFileName, std::list<Flowpipe> & flowpipes, const std::string & x, const std::string & y, const Variables & variables, const Interval & cutoff)
{
	fprintf(fp, "set terminal postscript enhanced color\n");

	char filename[NAME_SIZE+10];
	sprintf(filename, "%s.eps", epsFileName.c_str());
	fprintf(fp, "set output '%s'\n", filename);

	fprintf(fp, "set style line 1 linecolor rgb \"blue\"\n");
	fprintf(fp, "set autoscale\n");
	fprintf(fp, "unset label\n");
	fprintf(fp, "set xtic auto\n");
	fprintf(fp, "set ytic auto\n");
	fprintf(fp, "set xlabel \"%s\"\n", x.c_str());
	fprintf(fp, "set ylabel \"%s\"\n", y.c_str());
	fprintf(fp, "plot '-' notitle with lines ls 1\n");

	int idx = variables.getIDForVar(x) - 1;
	int idy = variables.getIDForVar(y) - 1;

	std::vector<flowstar::Interval> step_exp_table, step_end_exp_table;
	Interval intStep;

	std::list<Flowpipe>::const_iterator fpIter = flowpipes.begin();

	int maxOrder = 0;

	for(; fpIter != flowpipes.end(); ++fpIter)
	{
		int tmp = maxOrder;
		for(int i=0; i<fpIter->tmvPre.tms.size(); ++i)
		{
			int order = fpIter->tmvPre.tms[i].degree();
			if(maxOrder < order)
			{
				maxOrder = order;
			}
		}

		for(int i=0; i<fpIter->tmv.tms.size(); ++i)
		{
			int order = fpIter->tmv.tms[i].degree();
			if(maxOrder < order)
			{
				maxOrder = order;
			}
		}

		if(step_exp_table.size() == 0 || intStep != fpIter->domain[0] || maxOrder > tmp)
		{
			flowstar::construct_step_exp_table(step_exp_table, step_end_exp_table, fpIter->domain[0].sup(), maxOrder);
			intStep = fpIter->domain[0];
		}

		std::vector<flowstar::Interval> box;
		flowstar::TaylorModelVec tmv;
		fpIter->composition_normal(tmv, step_exp_table, (flowstar::Interval)cutoff);

		tmv.intEvalNormal(box, step_exp_table);

		Interval X(box[idx]), Y(box[idy]);

		fprintf(fp, "%lf %lf\n", X.inf(), Y.inf());
		fprintf(fp, "%lf %lf\n", X.sup(), Y.inf());
		fprintf(fp, "%lf %lf\n", X.sup(), Y.sup());
		fprintf(fp, "%lf %lf\n", X.inf(), Y.sup());
		fprintf(fp, "%lf %lf\n", X.inf(), Y.inf());
		fprintf(fp, "\n\n");
	}

	fprintf(fp, "e\n");
}

void plot_2D_box_matlab(FILE *fp, std::list<Flowpipe> & flowpipes, const std::string & x, const std::string & y, const Variables & variables, const Interval & cutoff)
{
	int idx = variables.getIDForVar(x) - 1;
	int idy = variables.getIDForVar(y) - 1;

	std::vector<flowstar::Interval> step_exp_table, step_end_exp_table;
	Interval intStep;

	std::list<Flowpipe>::const_iterator fpIter = flowpipes.begin();

	int maxOrder = 0;

	fprintf(fp, "figure\n");

	for(; fpIter != flowpipes.end(); ++fpIter)
	{
		int tmp = maxOrder;
		for(int i=0; i<fpIter->tmvPre.tms.size(); ++i)
		{
			int order = fpIter->tmvPre.tms[i].degree();
			if(maxOrder < order)
			{
				maxOrder = order;
			}
		}

		for(int i=0; i<fpIter->tmv.tms.size(); ++i)
		{
			int order = fpIter->tmv.tms[i].degree();
			if(maxOrder < order)
			{
				maxOrder = order;
			}
		}

		if(step_exp_table.size() == 0 || intStep != fpIter->domain[0] || maxOrder > tmp)
		{
			flowstar::construct_step_exp_table(step_exp_table, step_end_exp_table, fpIter->domain[0].sup(), maxOrder);
			intStep = fpIter->domain[0];
		}

		std::vector<flowstar::Interval> box;
		flowstar::TaylorModelVec tmv;
		fpIter->composition_normal(tmv, step_exp_table, (flowstar::Interval)cutoff);

		tmv.intEvalNormal(box, step_exp_table);

		Interval X(box[idx]), Y(box[idy]);

//		fprintf(fp,"plot( [%e,%e,%e,%e,%e] , [%e,%e,%e,%e,%e] , 'k');\nhold on;\nclear;\n",
//				X.inf(), X.sup(), X.sup(), X.inf(), X.inf(), Y.inf(), Y.inf(), Y.sup(), Y.sup(), Y.inf());

		fprintf(fp,"fill( [%e,%e,%e,%e,%e] , [%e,%e,%e,%e,%e] , [0.6,0.6,0.6], 'EdgeColor', [0.6,0.6,0.6]);\nhold on;\nclear;\n",
				X.inf(), X.sup(), X.sup(), X.inf(), X.inf(), Y.inf(), Y.inf(), Y.sup(), Y.sup(), Y.inf());
	}
}

void plot_2D_box_gnuplot_linear(FILE *fp, const std::string & epsFileName, const std::vector<flowstar::TaylorModelVec> & flowpipes, const std::vector<Interval> & domain, const int order, const std::string & x, const std::string & y, const Variables & variables)
{
	fprintf(fp, "set terminal postscript enhanced color\n");

	char filename[NAME_SIZE+10];
	sprintf(filename, "%s.eps", epsFileName.c_str());
	fprintf(fp, "set output '%s'\n", filename);

	fprintf(fp, "set style line 1 linecolor rgb \"blue\"\n");
	fprintf(fp, "set autoscale\n");
	fprintf(fp, "unset label\n");
	fprintf(fp, "set xtic auto\n");
	fprintf(fp, "set ytic auto\n");
	fprintf(fp, "set xlabel \"%s\"\n", x.c_str());
	fprintf(fp, "set ylabel \"%s\"\n", y.c_str());
	fprintf(fp, "plot '-' notitle with lines ls 1\n");

	int idx = variables.getIDForVar(x) - 1;
	int idy = variables.getIDForVar(y) - 1;

	std::vector<flowstar::Interval> step_exp_table, step_end_exp_table;
	flowstar::construct_step_exp_table(step_exp_table, step_end_exp_table, domain[0].sup(), order);

	for(int i=0; i<flowpipes.size(); ++i)
	{
		std::vector<flowstar::Interval> box;

		flowpipes[i].intEvalNormal(box, step_exp_table);

		Interval X(box[idx]), Y(box[idy]);

		fprintf(fp, "%lf %lf\n", X.inf(), Y.inf());
		fprintf(fp, "%lf %lf\n", X.sup(), Y.inf());
		fprintf(fp, "%lf %lf\n", X.sup(), Y.sup());
		fprintf(fp, "%lf %lf\n", X.inf(), Y.sup());
		fprintf(fp, "%lf %lf\n", X.inf(), Y.inf());
		fprintf(fp, "\n\n");
	}

	fprintf(fp, "e\n");
}

void plot_2D_box_matlab_linear(FILE *fp, const std::vector<flowstar::TaylorModelVec> & flowpipes, const std::vector<Interval> & domain, const int order, const std::string & x, const std::string & y, const Variables & variables)
{
	int idx = variables.getIDForVar(x) - 1;
	int idy = variables.getIDForVar(y) - 1;

	std::vector<flowstar::Interval> step_exp_table, step_end_exp_table;
	flowstar::construct_step_exp_table(step_exp_table, step_end_exp_table, domain[0].sup(), order);

	for(int i=0; i<flowpipes.size(); ++i)
	{
		std::vector<flowstar::Interval> box;

		flowpipes[i].intEvalNormal(box, step_exp_table);

		Interval X(box[idx]), Y(box[idy]);

		fprintf(fp,"plot( [%e,%e,%e,%e,%e] , [%e,%e,%e,%e,%e] , 'b');\nhold on;\nclear;\n",
				X.inf(), X.sup(), X.sup(), X.inf(), X.inf(), Y.inf(), Y.inf(), Y.sup(), Y.sup(), Y.inf());
	}
}

bool simulate(std::list<Flowpipe> & flowpipes, Flowpipe & last_flowpipe, const Flowpipe & init, const ODE & ode, ReachabilitySetting & rs, const int N)
{
	flowstar::Flowpipe fp_current = init;
//	flowpipes.clear();

	std::vector<flowstar::Interval> step_exp_table = rs.step_exp_table;
	std::vector<flowstar::Interval> step_end_exp_table = rs.step_end_exp_table;

	int M = 0;

	std::vector<flowstar::Interval> ode_rem;

	for(int i=0; i<ode.remainders.size(); ++i)
	{
		ode_rem.push_back(ode.remainders[i] * rs.step_exp_table[1]);
	}

	for(int i=N; i>0; )
	{
		flowstar::Flowpipe fp_next;
		bool bvalid = fp_current.advance_picard_tt(fp_next, ode.derivatives_hf, ode_rem, rs.precondition, step_exp_table, step_end_exp_table, rs.orders[0], rs.estimation, rs.cutoff_threshold);

		if(!bvalid)
		{
			if(M < REF_DEPTH)
			{
				double curStep = step_exp_table[1].sup() / 2;
				flowstar::construct_step_exp_table(step_exp_table, step_end_exp_table, curStep, 2*rs.globalMaxOrder);
				++M;
				i *= 2;
			}
			else
			{
				fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
				return false;
			}
		}
		else
		{
			flowpipes.push_back(fp_next);
			fp_current = fp_next;
			--i;

//			printf("stepsize = %lf\n", step_exp_table[1].sup());
		}
	}

	flowstar::TaylorModelVec tmvTemp;
	fp_current.tmvPre.evaluate_t(tmvTemp, rs.step_end_exp_table);

	fp_current.tmvPre = tmvTemp;
	last_flowpipe = (libflowstar::Flowpipe)fp_current;

//	flowpipes.push_back(last_flowpipe);

	return true;
}


bool getRange(Interval & range, const std::vector<TaylorModel> & tmFlowpipe, const std::vector<Interval> & domain, const std::string & vName, const Variables & vars)
{
	int id = vars.getIDForVar(vName);

	if(id == -1)
	{
		return false;
	}
	else
	{
		tmFlowpipe[id-1].intEval(range, *(std::vector<flowstar::Interval> *)(&domain));
		return true;
	}
}

bool getRangeNormal(Interval & range, const std::vector<TaylorModel> & tmFlowpipe, ReachabilitySetting & rs, const std::string & vName, const Variables & vars)
{
	int id = vars.getIDForVar(vName);

	if(id == -1)
	{
		return false;
	}
	else
	{
		tmFlowpipe[id-1].intEvalNormal(range, rs.step_exp_table);
		return true;
	}
}


int contract_flowpipe(std::vector<TaylorModel> & tmFlowpipe, std::vector<Interval> & domain, const std::vector<PolynomialConstraint> & constraints, ReachabilitySetting & rs)
{
	flowstar::TaylorModelVec tmv;
	tmv.tms = *(std::vector<flowstar::TaylorModel> *)(&tmFlowpipe);

	std::vector<bool> boundary_intersected;

	int res = contract_interval_arithmetic(tmv, *(std::vector<flowstar::Interval> *)(&domain), *(std::vector<flowstar::PolynomialConstraint> *)(&constraints), boundary_intersected, rs.cutoff_threshold);

	// -1: no intersection
	// 0:  no contraction
	// 1:  contracted, but not the time interval
	// 2:  contracted, also the time interval

	if(res != -1)
	{
		tmv.normalize(*(std::vector<flowstar::Interval> *)(&domain));
		tmFlowpipe = *(std::vector<libflowstar::TaylorModel> *)(&tmv.tms);
	}

	return res;
}
/*
int contract_TaylorModel(TaylorModel & tm, std::vector<Interval> & domain, const Interval & intMax, const Interval & intMin, ReachabilitySetting & rs)
{
	int res = contract_interval_arithmetic(tm, *(std::vector<flowstar::Interval> *)(&domain), *(flowstar::Interval *)(&intMax), *(flowstar::Interval *)(&intMin), rs.cutoff_threshold);

	if(res > 0)
	{
		tm.normalize(*(std::vector<flowstar::Interval> *)(&domain));
	}

	return res;
}
*/
int contract_TaylorModel(TaylorModel & tm_mid, std::vector<Interval> & domain_mid,
		TaylorModel & tm_max, std::vector<Interval> & domain_max, TaylorModel & tm_min, std::vector<Interval> & domain_min,
		const TaylorModel & tm, const std::vector<Interval> & domain, const Interval & intMax, const Interval & intMin, ReachabilitySetting & rs)
{
	int res = contract_interval_arithmetic(tm_mid, *(std::vector<flowstar::Interval> *)(&domain_mid),
			tm_max, *(std::vector<flowstar::Interval> *)(&domain_max), tm_min, *(std::vector<flowstar::Interval> *)(&domain_min),
			tm, *(std::vector<flowstar::Interval> *)(&domain), *(flowstar::Interval *)(&intMax), *(flowstar::Interval *)(&intMin), rs.cutoff_threshold);

	return res;
}

}
