
#ifndef CONTINUOUS_LIB_H_
#define CONTINUOUS_LIB_H_


#include "Variables_lib.h"
#include "DataTypes_lib.h"

#define REF_DEPTH 5

namespace libflowstar
{

class ContinuousSystem;
class ReachabilitySetting;
class ODE;

class Flowpipe : public flowstar::Flowpipe
{
public:
	Flowpipe();
	Flowpipe(const Assignment<Interval> & box);
	Flowpipe(const std::vector<TaylorModel> & tm, const std::vector<Interval> & tmDomain);
	Flowpipe(const std::vector<TaylorModel> & tmVector, const Assignment<Interval> & tmVectorDomain);
	Flowpipe(const std::vector<TaylorModel> & tmVectorPre, const std::vector<TaylorModel> & tmVector, const Assignment<Interval> & tmVectorDomain);
	Flowpipe(const flowstar::Flowpipe & flowpipe);
	~Flowpipe();

	void rangeOfPoly(Assignment<Interval> & range, const ReachabilitySetting & rs) const;
	void range(Assignment<Interval> & range, const ReachabilitySetting & rs) const;

	void getStepsize(Interval & step) const;

	void merge(const std::vector<TaylorModel> & tmFlowpipe, const std::vector<Interval> & domain, const ReachabilitySetting & rs);

	void output(FILE *fp, const Variables & stateVars, const Variables & tmVars, const Interval & cutoff = ZERO) const;
	void toTaylorModel(std::vector<TaylorModel> & tmv, std::vector<Interval> & dom, const ReachabilitySetting & rs) const;
	void toTaylorModelNormal(std::vector<TaylorModel> & tmv, std::vector<Interval> & dom, const ReachabilitySetting & rs) const;

	bool next_poly_ode_1(Flowpipe & nextFlowpipe, const ContinuousSystem & system, ReachabilitySetting & rs) const;

	bool next_flowpipe_linear(Assignment<TaylorModel> & tmFlowpipe, Assignment<Interval> & flowpipeDomain, const ContinuousSystem & system, ReachabilitySetting & rs) const;

	friend class ContinuousSystem;

	friend void plot_2D_box_gnuplot(FILE *fp, const std::string & epsFileName, std::list<Flowpipe> & flowpipes, const std::string & x, const std::string & y, const Variables & variables, const Interval & cutoff);
	friend void plot_2D_box_matlab(FILE *fp, std::list<Flowpipe> & flowpipes, const std::string & x, const std::string & y, const Variables & variables, const Interval & cutoff);
	friend bool simulate(std::list<Flowpipe> & flowpipes, Flowpipe & last_flowpipe, const Flowpipe & init, const ODE & ode, ReachabilitySetting & rs, const int N);
};



class ODE
{
protected:
	Variables variables;
	std::vector<Polynomial> derivatives;
	std::vector<flowstar::HornerForm> derivatives_hf;
	std::vector<flowstar::Interval> remainders;

public:
	ODE(const Variables & vars);
	ODE(const ODE & ode);
	~ODE();

	ODE & operator = (const ODE & ode);

	bool assignDerivative(const std::string & varName, const Polynomial & polynomial);
	bool assignDerivative(const std::string & varName, const TaylorModel & taylormodel);

	bool tmEval(TaylorModel & result, const std::vector<TaylorModel> & tms, const int order, const std::string & varName, ReachabilitySetting & rs) const;

	bool getHF(flowstar::HornerForm & hf, flowstar::Interval & rem, const std::string & varName) const;

	void output(FILE *fp) const;

	friend bool simulate(std::list<Flowpipe> & flowpipes, Flowpipe & last_flowpipe, const Flowpipe & init, const ODE & ode, ReachabilitySetting & rs, const int N);
};


class ContinuousSystem : public flowstar::ContinuousSystem
{
private:
	Variables variables;								// the state variables

public:
	ContinuousSystem();
	~ContinuousSystem();
	ContinuousSystem(const Assignment<Polynomial> & ode, const Variables & stateVars);	// the first variable of the polynomials should be reserved by t
	ContinuousSystem(const ContinuousSystem & system);

	ContinuousSystem & operator = (const ContinuousSystem & system);

	void output(FILE *fp) const;
	bool setInit(const Flowpipe & flowpipe);

	void reach_linear(std::list<Flowpipe> & result, const ReachabilitySetting & rs) const;
	void reach_poly_1(std::list<Flowpipe> & result, const ReachabilitySetting & rs) const;
	void reach_linear_hybridization(std::vector<flowstar::TaylorModelVec> & tmFlowpipes, std::vector<Interval> & domain, const ReachabilitySetting & rs, const std::vector<int> & varIDs, const double epsilon, const int N) const;
	void reach_symbolic_remainder(std::list<Flowpipe> & result, const ReachabilitySetting & rs, const int N) const;

	friend class Flowpipe;
};

void plot_2D_box_gnuplot(FILE *fp, const std::string & epsFileName, std::list<Flowpipe> & flowpipes, const std::string & x, const std::string & y, const Variables & variables, const Interval & cutoff = ZERO);
void plot_2D_box_matlab(FILE *fp, std::list<Flowpipe> & flowpipes, const std::string & x, const std::string & y, const Variables & variables, const Interval & cutoff = ZERO);
void plot_2D_box_gnuplot_linear(FILE *fp, const std::string & epsFileName, const std::vector<flowstar::TaylorModelVec> & flowpipes, const std::vector<Interval> & domain, const int order, const std::string & x, const std::string & y, const Variables & variables);
void plot_2D_box_matlab_linear(FILE *fp, const std::vector<flowstar::TaylorModelVec> & flowpipes, const std::vector<Interval> & domain, const int order, const std::string & x, const std::string & y, const Variables & variables);

bool simulate(std::list<Flowpipe> & flowpipes, Flowpipe & last_flowpipe, const Flowpipe & init, const ODE & ode, ReachabilitySetting & rs, const int N);

bool getRange(Interval & range, const std::vector<TaylorModel> & tmFlowpipe, const std::vector<Interval> & domain, const std::string & vName, const Variables & vars);
bool getRangeNormal(Interval & range, const std::vector<TaylorModel> & tmFlowpipe, ReachabilitySetting & rs, const std::string & vName, const Variables & vars);

int contract_flowpipe(std::vector<TaylorModel> & tmFlowpipe, std::vector<Interval> & domain, const std::vector<PolynomialConstraint> & constraints, ReachabilitySetting & rs);

/* return values:
 * -2: tm <= min
 * -1: tm >= max
 *  0: min <= tm <= max
 *  1: min is contained in tm but not max
 *  2: max is contained in tm but not min
 *  3: tm strictly contains min and max
 */
//int contract_TaylorModel(TaylorModel & tm, std::vector<Interval> & domain, const Interval & intMax, const Interval & intMin, ReachabilitySetting & rs);

int contract_TaylorModel(TaylorModel & tm_mid, std::vector<Interval> & domain_mid,
		TaylorModel & tm_max, std::vector<Interval> & domain_max, TaylorModel & tm_min, std::vector<Interval> & domain_min,
		const TaylorModel & tm, const std::vector<Interval> & domain, const Interval & intMax, const Interval & intMin, ReachabilitySetting & rs);

}

#endif /* CONTINUOUS_LIB_H_ */
