
#include "DataTypes_lib.h"

libflowstar::ParseSetting parseSettingLib;
flowstar::ParseResult parseResultLib;

namespace libflowstar
{
Interval ZERO;
}

using namespace libflowstar;

Interval::Interval() : flowstar::Interval()
{
}

Interval::Interval(const double c) : flowstar::Interval(c)
{
}

Interval::Interval(const double l, const double u) : flowstar::Interval(l, u)
{
}

Interval::Interval(const flowstar::Interval & I) : flowstar::Interval(I)
{
}

Interval::~Interval()
{
}

void Interval::output(FILE *fp) const
{
	dump(fp);
}



ParseSetting::ParseSetting() : flowstar::ParseSetting()
{
}

ParseSetting::ParseSetting(const ParseSetting & setting)
{
	strODE = setting.strODE;

	ranges = setting.ranges;
	iterRange = setting.iterRange;

	cutoff_threshold = setting.cutoff_threshold;

	step_exp_table = setting.step_exp_table;
	flowpipe = setting.flowpipe;
	order = setting.order;

	variables = setting.variables;
	strPolynomial = setting.strPolynomial;
}

ParseSetting::~ParseSetting()
{
	ranges.clear();
	step_exp_table.clear();
	variables.clear();
}

void ParseSetting::clear()
{
	ranges.clear();
	iterRange = ranges.begin();

	step_exp_table.clear();

	flowstar::Interval intZero;
	cutoff_threshold = intZero;

	variables.clear();
}

ParseSetting & ParseSetting::operator = (const ParseSetting & setting)
{
	if(this == &setting)
		return *this;

	strODE = setting.strODE;

	ranges = setting.ranges;
	iterRange = setting.iterRange;

	cutoff_threshold = setting.cutoff_threshold;

	step_exp_table = setting.step_exp_table;
	flowpipe = setting.flowpipe;
	order = setting.order;

	variables = setting.variables;
	strPolynomial = setting.strPolynomial;

	return *this;
}



PolynomialConstraint::PolynomialConstraint() : flowstar::PolynomialConstraint()
{
}

PolynomialConstraint::PolynomialConstraint(const std::string & strPolynomial, const Variables & vars)
{
	Polynomial polynomial(strPolynomial, vars);

	flowstar::Interval zero(0);
	p = (flowstar::Polynomial)polynomial;

	p.toHornerForm(hf);
	B = zero;
}

void PolynomialConstraint::output(FILE *fp, const Variables & vars) const
{
	flowstar::PolynomialConstraint::dump(fp, vars.varNames);
}




Polynomial::Polynomial() : flowstar::Polynomial()
{
}

Polynomial::Polynomial(const std::string & strPolynomial, const Variables & variables)
{
	parseSettingLib.clear();

	std::string prefix(str_prefix_polynomial_lib);
	std::string suffix(str_suffix);

	parseSettingLib.variables = variables;
	parseSettingLib.strPolynomial = prefix + strPolynomial + suffix;

	parsePolynomialLib();		// call the parser

	*this = parseResultLib.expansion;
}

Polynomial::Polynomial(const flowstar::Polynomial & polynomial) : flowstar::Polynomial(polynomial)
{
}

Polynomial::~Polynomial()
{
}

void Polynomial::substitute(const std::string & vName, const Variables & vars, const Interval & I)
{
	int id = vars.getIDForVar(vName);

	if(id >= 0)
	{
		flowstar::Polynomial::substitute(id, I);
	}
}

void Polynomial::output(FILE *fp, const Variables & variables) const
{
	dump_interval(fp, variables.varNames);
	fprintf(fp, "\n");
}



ReachabilitySetting::ReachabilitySetting()
{
	step				= 0;
	time				= 0;
	precondition		= ID_PRE;
	orderType			= -1;
	bAdaptiveSteps		= false;
	bAdaptiveOrders		= false;
	miniStep			= 0;
	globalMaxOrder		= 0;
	bPrint				= true;
	integrationScheme	= false;

	Interval cutoff(-1e-12,1e-12);
	cutoff_threshold = cutoff;
}

ReachabilitySetting::~ReachabilitySetting()
{
	estimation.clear();
	orders.clear();
	maxOrders.clear();
	step_exp_table.clear();
	step_end_exp_table.clear();
}

ReachabilitySetting::ReachabilitySetting(const ReachabilitySetting & rs)
{
	step				= rs.step;
	time				= rs.time;
	precondition		= rs.precondition;
	orderType			= rs.orderType;
	bAdaptiveSteps		= rs.bAdaptiveSteps;
	bAdaptiveOrders		= rs.bAdaptiveOrders;
	miniStep			= rs.miniStep;
	globalMaxOrder		= rs.globalMaxOrder;
	bPrint				= rs.bPrint;
	integrationScheme	= rs.integrationScheme;
	estimation			= rs.estimation;
	orders				= rs.orders;
	maxOrders			= rs.maxOrders;
	cutoff_threshold	= rs.cutoff_threshold;
	step_exp_table		= rs.step_exp_table;
	step_end_exp_table	= rs.step_end_exp_table;
}

ReachabilitySetting & ReachabilitySetting::operator = (const ReachabilitySetting & rs)
{
	if(this == &rs)
		return *this;

	step				= rs.step;
	time				= rs.time;
	precondition		= rs.precondition;
	orderType			= rs.orderType;
	bAdaptiveSteps		= rs.bAdaptiveSteps;
	bAdaptiveOrders		= rs.bAdaptiveOrders;
	miniStep			= rs.miniStep;
	globalMaxOrder		= rs.globalMaxOrder;
	bPrint				= rs.bPrint;
	integrationScheme	= rs.integrationScheme;
	estimation			= rs.estimation;
	orders				= rs.orders;
	maxOrders			= rs.maxOrders;
	cutoff_threshold	= rs.cutoff_threshold;
	step_exp_table		= rs.step_exp_table;
	step_end_exp_table	= rs.step_end_exp_table;

	return *this;
}

bool ReachabilitySetting::setTime(const double t)
{
	if(t <= 0)
	{
		printf("Time horizon must be a positive number.\n");
		return false;
	}
	else
	{
		time = t;
		return true;
	}
}

bool ReachabilitySetting::setFixedStepsize(const double stepsize)
{
	if(stepsize > 0)
	{
		step = stepsize;
		bAdaptiveSteps = false;
		return true;
	}
	else
	{
		printf("A time step-size should be larger than 0.\n");
		return false;
	}
}

bool ReachabilitySetting::setAdaptiveStepsize(const double minStepsize, const double maxStepsize)
{
	if(minStepsize <= 0 || maxStepsize <= 0)
	{
		printf("A time step-size should be larger than 0.\n");
		return false;
	}

	if(minStepsize <= maxStepsize)
	{
		step = maxStepsize;
		miniStep = minStepsize;
		bAdaptiveSteps = true;
		return true;
	}
	else
	{
		printf("MIN step should be no larger than MAX step.\n");
		return false;
	}
}

bool ReachabilitySetting::setFixedOrder(const int order)
{
	if(order > 0)
	{
		orders.push_back(order);
		globalMaxOrder = order;
		bAdaptiveOrders = false;
		orderType = UNIFORM;
		return true;
	}
	else
	{
		printf("Orders should be larger than zero.\n");
		return false;
	}
}

bool ReachabilitySetting::setFixedOrder(const std::vector<int> & order)
{
	int maxOrder = 0;

	for(int i=1; i<order.size(); ++i)
	{
		if(order[i] > 0)
		{
			orders.push_back(order[i]);

			if(maxOrder < order[i])
				maxOrder = order[i];
		}
		else
		{
			printf("Orders should be larger than zero.\n");
			return false;
		}
	}

	bAdaptiveOrders = false;
	orderType = MULTI;

	return true;
}

bool ReachabilitySetting::setAdaptiveOrder(const int k, const int max_k)
{
	if(k > 0 && max_k > 0)
	{
		if(k <= max_k)
		{
			orders.push_back(k);
			maxOrders.push_back(max_k);
			globalMaxOrder = max_k;
			bAdaptiveOrders = true;
			orderType = UNIFORM;
			return true;
		}
		else
		{
			printf("No order less than %d and greater than %d can be found.\n", max_k, k);
			return false;
		}
	}
	else
	{
		printf("Orders should be larger than zero.\n");
		return false;
	}
}

bool ReachabilitySetting::setAdaptiveOrder(const Assignment<int> & k, const Assignment<int> & max_k)
{
	if(k.size() != max_k.size())
	{
		printf("The dimensions of MIN and MAX orders do not match.\n");
		return false;
	}

	int maxOrder = 0;

	for(int i=1; i<k.size(); ++i)
	{
		if(k.values[i] <= max_k.values[i])
		{
			orders.push_back(k.values[i]);
			maxOrders.push_back(max_k.values[i]);

			if(maxOrder < max_k.values[i])
				maxOrder = max_k.values[i];
		}
		else
		{
			printf("No order less than %d and greater than %d can be found.\n", max_k.values[i], k.values[i]);
			return false;
		}
	}

	bAdaptiveOrders = true;
	orderType = MULTI;
	globalMaxOrder = maxOrder;

	return true;
}

bool ReachabilitySetting::setRemainderEstimation(const Assignment<Interval> & intervals)
{
	for(int i=1; i<intervals.size(); ++i)
	{
		if(!intervals.values[i].valid())
		{
			printf("Remainder estimation is not valid.\n");
			return false;
		}
		else
		{
			estimation.push_back((flowstar::Interval)intervals.values[i]);
		}
	}

	return true;
}

bool ReachabilitySetting::setPrecondition(const int type)
{
	if(type != QR_PRE && type != ID_PRE)
	{
		printf("Unknown preconditioning specification.\n");
		return false;
	}
	else
	{
		precondition = type;
		return true;
	}
}

bool ReachabilitySetting::setCutoff(const Interval & cutoff)
{
	if(cutoff.valid())
	{
		cutoff_threshold = (flowstar::Interval)cutoff;
		return true;
	}
	else
	{
		printf("Cutoff threshold is not a valid interval.\n");
		return false;
	}
}

bool ReachabilitySetting::setPrecision(const int prec)
{
	if(prec >= 53)
	{
		intervalNumPrecision = prec;
		return true;
	}
	else
	{
		printf("Precision should be large than 53.\n");
		return false;
	}
}

void ReachabilitySetting::printon()
{
	bPrint = true;
}

void ReachabilitySetting::printoff()
{
	bPrint = false;
}

void ReachabilitySetting::prepareForReachability()
{
	flowstar::compute_factorial_rec(globalMaxOrder+5);
	flowstar::compute_power_4(globalMaxOrder+5);
	flowstar::compute_double_factorial(2*globalMaxOrder+10);

	flowstar::construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);
}






TaylorModel::TaylorModel() : flowstar::TaylorModel()
{
}

TaylorModel::TaylorModel(const std::string & strPolynomial, const Variables & variables)
{
	Polynomial polynomial(strPolynomial, variables);

	expansion = (flowstar::Polynomial)polynomial;
}

TaylorModel::TaylorModel(const std::string & strPolynomial, const Variables & variables, const Interval & I)
{
	Polynomial polynomial(strPolynomial, variables);

	expansion = (flowstar::Polynomial)polynomial;
	remainder = (flowstar::Interval)I;
}

TaylorModel::TaylorModel(const Polynomial & polynomial)
{
	expansion = (flowstar::Polynomial)polynomial;
}

TaylorModel::TaylorModel(const flowstar::TaylorModel & tm)
{
	expansion = tm.expansion;
	remainder = tm.remainder;
}

TaylorModel::~TaylorModel()
{
}

void TaylorModel::output(FILE *fp, const Variables & variables) const
{
	dump_interval(fp, variables.varNames);
	fprintf(fp, "\n");
}

TaylorModel & TaylorModel::operator += (const TaylorModel & tm)
{
	expansion += tm.expansion;
	remainder += tm.remainder;

	return *this;
}

TaylorModel & TaylorModel::operator -= (const TaylorModel & tm)
{
	expansion -= tm.expansion;
	remainder -= tm.remainder;

	return *this;
}

TaylorModel & TaylorModel::operator *= (const double d)
{
	Interval I(d);
	expansion *= I;
	remainder *= I;

	return *this;
}

TaylorModel & TaylorModel::operator *= (const Interval & I)
{
	expansion *= I;
	remainder *= I;

	return *this;
}

TaylorModel TaylorModel::operator + (const TaylorModel & tm) const
{
	TaylorModel result = *this;
	result += tm;

	return result;
}

TaylorModel TaylorModel::operator - (const TaylorModel & tm) const
{
	TaylorModel result = *this;
	result -= tm;

	return result;
}

TaylorModel TaylorModel::operator * (const double d) const
{
	TaylorModel result = *this;
	result *= d;

	return result;
}

TaylorModel TaylorModel::operator * (const Interval & I) const
{
	TaylorModel result = *this;
	result *= I;

	return result;
}

void TaylorModel::ctrunc(const Assignment<Interval> & domain, const int order)
{
	flowstar::TaylorModel::ctrunc( *(std::vector<flowstar::Interval> *)(&domain.values) , order);
}

void TaylorModel::nctrunc(const int order)
{
	flowstar::TaylorModel::nctrunc(order);
}

void TaylorModel::mul(TaylorModel & result, const TaylorModel & tm, const Assignment<Interval> & domain, const int order, const Interval & cutoff_threshold) const
{
	flowstar::TaylorModel::mul_ctrunc(result, tm, *(std::vector<flowstar::Interval> *)(&domain.values), order, cutoff_threshold);
}

void TaylorModel::mul_no_remainder(TaylorModel & result, const TaylorModel & tm, const int order, const Interval & cutoff_threshold) const
{
	flowstar::TaylorModel::mul_no_remainder(result, tm, order, cutoff_threshold);
}


namespace libflowstar
{

void combine(TaylorModel & result, const flowstar::HornerForm & hf, const Interval & rem, const std::vector<TaylorModel> & replacement, const std::vector<Interval> & domain, const int order, const Interval & cutoff_threshold)
{
	std::vector<flowstar::Interval> polyRange;
	flowstar::TaylorModelVec vars;

	for(int i=0; i<replacement.size(); ++i)
	{
		Interval I;
		replacement[i].polyRange(I, *(std::vector<flowstar::Interval> *)(&domain));
		polyRange.push_back(I);

		vars.tms.push_back(replacement[i]);
	}

	hf.insert_ctrunc(result, vars, polyRange, *(std::vector<flowstar::Interval> *)(&domain), order, cutoff_threshold);
	result.remainder += rem;
}

void normalize(std::vector<TaylorModel> & tms, std::vector<Interval> & domain)
{
	flowstar::TaylorModelVec tmv;
	tmv.tms = *(std::vector<flowstar::TaylorModel> *)(&tms);

	tmv.normalize(*(std::vector<flowstar::Interval> *)(&domain));

	tms = *(std::vector<libflowstar::TaylorModel> *)(&tmv.tms);
}

}
