
#ifndef DATATYPES_LIB_H_
#define DATATYPES_LIB_H_

#include "Variables_lib.h"
#include "TaylorModel.h"
#include "Constraints.h"
#include "Continuous.h"

namespace libflowstar
{

class Interval : public flowstar::Interval
{
public:
	Interval();
	Interval(const double c);
	Interval(const double l, const double u);
	Interval(const flowstar::Interval & I);
	~Interval();

	void output(FILE *fp) const;
};

extern Interval ZERO;

class ParseSetting : public flowstar::ParseSetting
{
public:
	Variables variables;
	std::string strPolynomial;

	ParseSetting();
	ParseSetting(const ParseSetting & setting);
	~ParseSetting();

	virtual void clear();

	virtual ParseSetting & operator = (const ParseSetting & setting);
};

// p(x) <= 0
class PolynomialConstraint : flowstar::PolynomialConstraint
{
public:
	PolynomialConstraint();
	PolynomialConstraint(const std::string & strPolynomial, const Variables & variables);

	void output(FILE *fp, const Variables & vars) const;
};

template <typename V>
class Assignment
{
protected:
	std::vector<V> values;

public:
	Assignment(const int n);
	~Assignment();

	void clear();

	bool assign(const std::string & vName, const V & value, const Variables & variables);
	bool getValue(V & value, const std::string & vName, const Variables & variables) const;
	int size() const;

	V & operator [] (const int i);
	Assignment & operator = (const Assignment & assignment);

	friend class Polynomial;
	friend class TaylorModel;
	friend class Flowpipe;
	friend class ReachabilitySetting;
	friend class ContinuousSystem;
};

template <typename V>
Assignment<V>::Assignment(const int n)
{
	V defaultValue;

	for(int i=0; i<n; ++i)
	{
		values.push_back(defaultValue);
	}
}

template <typename V>
Assignment<V>::~Assignment()
{
	values.clear();
}

template <typename V>
void Assignment<V>::clear()
{
	values.clear();
}

template <typename V>
bool Assignment<V>::assign(const std::string & vName, const V & value, const Variables & variables)
{
	int id = variables.getIDForVar(vName);

	if(id >= 0)
	{
		values[id] = value;
		return true;
	}
	else
	{
		return false;
	}
}

template <typename V>
bool Assignment<V>::getValue(V & value, const std::string & vName, const Variables & variables) const
{
	int id = variables.getIDForVar(vName);

	if(id >= 0)
	{
		value = values[id];
		return true;
	}
	else
	{
		return false;
	}
}

template <typename V>
int Assignment<V>::size() const
{
	return (int)values.size();
}

template <typename V>
V & Assignment<V>::operator [] (const int i)
{
	return values[i];
}

template <typename V>
Assignment<V> & Assignment<V>::operator = (const Assignment<V> & assignment)
{
	if(this == &assignment)
		return *this;

	values = assignment.values;
}


class Polynomial : public flowstar::Polynomial
{
public:
	Polynomial();
	Polynomial(const std::string & strPolynomial, const Variables & variables);
	Polynomial(const flowstar::Polynomial & polynomial);
	~Polynomial();

	void substitute(const std::string & vName, const Variables & vars, const Interval & I);	// replace a variable by an interval

	void output(FILE *fp, const Variables & variables) const;
};


class ReachabilitySetting
{
public:
	double step;						// step-size
	double time;						// time horizon
	int precondition;					// preconditioning technique
	int orderType;
	bool bAdaptiveSteps;
	bool bAdaptiveOrders;

	std::vector<flowstar::Interval> estimation;	// remainder estimation for varying time step
	double miniStep;					// minimum step size
	std::vector<int> orders;			// order(s)
	std::vector<int> maxOrders;			// maximum order(s)
	int globalMaxOrder;

	bool bPrint;
	int integrationScheme;
	flowstar::Interval cutoff_threshold;

	std::vector<flowstar::Interval> step_exp_table;
	std::vector<flowstar::Interval> step_end_exp_table;

public:
	ReachabilitySetting();
	~ReachabilitySetting();
	ReachabilitySetting(const ReachabilitySetting & rs);

	ReachabilitySetting & operator = (const ReachabilitySetting & rs);

	bool setTime(const double t);
	bool setFixedStepsize(const double stepsize);
	bool setAdaptiveStepsize(const double minStepsize, const double maxStepsize);
	bool setFixedOrder(const int order);
	bool setFixedOrder(const std::vector<int> & order);
	bool setAdaptiveOrder(const int k, const int max_k);
	bool setAdaptiveOrder(const Assignment<int> & k, const Assignment<int> & max_k);	// k and max_k should be assigned based on the same variable definition
	bool setRemainderEstimation(const Assignment<Interval> & intervals);
	bool setPrecondition(const int type);
	bool setCutoff(const Interval & cutoff);
	bool setPrecision(const int prec);
	void printon();
	void printoff();

	void prepareForReachability();	// call this function after all parameters are properly set

	friend class Flowpipe;
	friend class ContinuousSystem;
};


class TaylorModel : public flowstar::TaylorModel
{
public:
	TaylorModel();
	TaylorModel(const std::string & strPolynomial, const Variables & variables);
	TaylorModel(const std::string & strPolynomial, const Variables & variables, const Interval & I);
	TaylorModel(const Polynomial & polynomial);
	TaylorModel(const flowstar::TaylorModel & tm);
	~TaylorModel();

	void output(FILE *fp, const Variables & variables) const;

	TaylorModel & operator += (const TaylorModel & tm);
	TaylorModel & operator -= (const TaylorModel & tm);
	TaylorModel & operator *= (const double d);
	TaylorModel & operator *= (const Interval & I);

	TaylorModel operator + (const TaylorModel & tm) const;
	TaylorModel operator + (const Interval & I) const;
	TaylorModel operator - (const TaylorModel & tm) const;
	TaylorModel operator * (const double d) const;
	TaylorModel operator * (const Interval & I) const;

	void ctrunc(const Assignment<Interval> & domain, const int order);		// conservative truncation
	void nctrunc(const int order);											// non-conservative truncation
	void mul(TaylorModel & result, const TaylorModel & tm, const Assignment<Interval> & domain, const int order, const Interval & cutoff_threshold = ZERO) const;
	void mul_no_remainder(TaylorModel & result, const TaylorModel & tm, const int order, const Interval & cutoff_threshold = ZERO) const;
};


void combine(TaylorModel & result, const flowstar::HornerForm & hf, const Interval & rem, const std::vector<TaylorModel> & replacement, const std::vector<Interval> & domain, const int order, const Interval & cutoff_threshold);

void normalize(std::vector<TaylorModel> & tms, std::vector<Interval> & domain);

}

extern libflowstar::ParseSetting parseSettingLib;
extern flowstar::ParseResult parseResultLib;


#endif /* DATATYPES_LIB_H_ */
