/*---
  Flow*: A Verification Tool for Cyber-Physical Systems.
  Authors: Xin Chen, Sriram Sankaranarayanan, and Erika Abraham.
  Email: Xin Chen <chenxin415@gmail.com> if you have questions or comments.

  The code is released as is under the GNU General Public License (GPL).
---*/

#include "Continuous.h"

using namespace flowstar;

Flowpipe::Flowpipe()
{
}

Flowpipe::Flowpipe(const TaylorModelVec & tmvPre_input, const TaylorModelVec & tmv_input, const std::vector<Interval> & domain_input):
		tmvPre(tmvPre_input), tmv(tmv_input), domain(domain_input)
{
}

Flowpipe::Flowpipe(const std::vector<Interval> & box, const Interval & I)
{
	int rangeDim = box.size();
	int domainDim = rangeDim + 1;
	Interval intUnit(-1,1), intZero;

	TaylorModelVec tmvCenter;
	std::vector<double> scalars;

	domain.push_back(I);		// time interval

	// normalize the domain to [-1,1]^n
	for(int i=0; i<rangeDim; ++i)
	{
		double midpoint = box[i].midpoint();
		Interval intMid(midpoint);
		TaylorModel tmTemp(intMid, domainDim);
		tmvCenter.tms.push_back(tmTemp);

		Interval intTemp = box[i];
		intTemp.sub_assign(midpoint);
		scalars.push_back( intTemp.sup() );
		domain.push_back(intUnit);
	}

	Matrix coefficients_of_tmvPre(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		coefficients_of_tmvPre.set(scalars[i], i, i+1);
	}

	TaylorModelVec tmvTemp(coefficients_of_tmvPre);
	tmvTemp.add(tmvPre, tmvCenter);

	Matrix coefficients_of_tmv(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		coefficients_of_tmv.set(1, i, i+1);
	}

	TaylorModelVec tmvTemp2(coefficients_of_tmv);
	tmv = tmvTemp2;
}

Flowpipe::Flowpipe(const TaylorModelVec & tmv_input, const std::vector<Interval> & domain_input)
{
	int rangeDim = tmv_input.tms.size();

	tmv = tmv_input;
	domain = domain_input;

	Matrix coefficients_of_tmvPre(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		coefficients_of_tmvPre.set(1, i, i+1);
	}

	TaylorModelVec tmvTemp(coefficients_of_tmvPre);
	tmvPre = tmvTemp;

	normalize();
}

Flowpipe::Flowpipe(const Flowpipe & flowpipe) : tmvPre(flowpipe.tmvPre), tmv(flowpipe.tmv), domain(flowpipe.domain)
{
}

Flowpipe::~Flowpipe()
{
	clear();
}

void Flowpipe::clear()
{
	tmvPre.clear();
	tmv.clear();
	domain.clear();
}

void Flowpipe::dump(FILE *fp, const std::vector<std::string> & stateVarNames, const std::vector<std::string> & tmVarNames, const Interval & cutoff_threshold) const
{
	TaylorModelVec tmvTemp;

	composition(tmvTemp, cutoff_threshold);

	// dump the Taylor model
	tmvTemp.dump_interval(fp, stateVarNames, tmVarNames);

	//dump the domain
	for(int i=0; i<domain.size(); ++i)
	{
		fprintf(fp, "%s in ", tmVarNames[i].c_str());
		domain[i].dump(fp);
		fprintf(fp, "\n");
	}
}

void Flowpipe::dump_normal(FILE *fp, const std::vector<std::string> & stateVarNames, const std::vector<std::string> & tmVarNames, std::vector<Interval> & step_exp_table, const Interval & cutoff_threshold) const
{
	TaylorModelVec tmvTemp;

	composition_normal(tmvTemp, step_exp_table, cutoff_threshold);

	// dump the Taylor model
	tmvTemp.dump_interval(fp, stateVarNames, tmVarNames);

	//dump the domain
	for(int i=0; i<domain.size(); ++i)
	{
		fprintf(fp, "%s in ", tmVarNames[i].c_str());
		domain[i].dump(fp);
		fprintf(fp, "\n");
	}

	fprintf(fp, "\n");
}

void Flowpipe::composition(TaylorModelVec & result, const Interval & cutoff_threshold) const
{
	std::vector<int> orders;

	for(int i=0; i<tmv.tms.size(); ++i)
	{
		int d1 = tmv.tms[i].degree();
		int d2 = tmvPre.tms[i].degree();

		if(d1 > d2)
		{
			orders.push_back(d1);
		}
		else
		{
			orders.push_back(d2);
		}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRange(tmvPolyRange, domain);
	tmvPre.insert_ctrunc(result, tmv, tmvPolyRange, domain, orders, cutoff_threshold);
}

void Flowpipe::composition_normal(TaylorModelVec & result, const std::vector<Interval> & step_exp_table, const Interval & cutoff_threshold) const
{
	int domainDim = domain.size();

	std::vector<int> orders;

	for(int i=0; i<tmv.tms.size(); ++i)
	{
		int d1 = tmv.tms[i].degree();
		int d2 = tmvPre.tms[i].degree();

		if(d1 > d2)
		{
			orders.push_back(d1);
		}
		else
		{
			orders.push_back(d2);
		}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_exp_table);
	tmvPre.insert_ctrunc_normal(result, tmv, tmvPolyRange, step_exp_table, domainDim, orders, cutoff_threshold);
}

void Flowpipe::intEval(std::vector<Interval> & result, const Interval & cutoff_threshold) const
{
	TaylorModelVec tmvTemp;
	composition(tmvTemp, cutoff_threshold);

	tmvTemp.intEval(result, domain);
}

void Flowpipe::intEvalNormal(std::vector<Interval> & result, const std::vector<Interval> & step_exp_table, const Interval & cutoff_threshold) const
{
	TaylorModelVec tmvTemp;
	composition_normal(tmvTemp, step_exp_table, cutoff_threshold);
	tmvTemp.intEvalNormal(result, step_exp_table);
}

void Flowpipe::intEvalNormal(std::vector<Interval> & result, const std::vector<int> & varIDs, const std::vector<Interval> & step_exp_table, const int order, const Interval & cutoff_threshold) const
{
	result.clear();

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_exp_table);

	for(int i=0; i<varIDs.size(); ++i)
	{
		TaylorModel tm;
		tmvPre.tms[varIDs[i]-1].insert_ctrunc_normal(tm, tmv, tmvPolyRange, step_exp_table, domain.size(), order, cutoff_threshold);

		Interval I;
		tm.intEvalNormal(I, step_exp_table);
		result.push_back(I);
	}
}

void Flowpipe::normalize()
{
	Interval intZero;

	// we first normalize the Taylor model tmv
	tmv.normalize(domain);

	int rangeDim = tmv.tms.size();

	// compute the center point of tmv
	std::vector<Interval> intVecCenter;
	tmv.constant(intVecCenter);
	tmv.rmConstant();

	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		tmv.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	std::vector<Interval> tmvRange;
	tmv.intEval(tmvRange, domain);

	std::vector<std::vector<Interval> > coefficients;
	std::vector<Interval> row;

	for(int i=0; i<rangeDim+1; ++i)
	{
		row.push_back(intZero);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		coefficients.push_back(row);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		Interval intScalor;
		tmvRange[i].mag(intScalor);

		if(intScalor.subseteq(intZero))
		{
			coefficients[i][i+1] = intZero;
		}
		else
		{
			coefficients[i][i+1] = intScalor;
			tmv.tms[i].div_assign(intScalor);
		}
	}

	TaylorModelVec newVars(coefficients);
	for(int i=0; i<rangeDim; ++i)
	{
		TaylorModel tmTemp(intVecCenter[i], rangeDim+1);
		newVars.tms[i].add_assign(tmTemp);
	}

	for(int i=0; i<tmvPre.tms.size(); ++i)
	{
		TaylorModel tmTemp;
		tmvPre.tms[i].insert_no_remainder_no_cutoff(tmTemp, newVars, rangeDim+1, tmvPre.tms[i].degree());
		tmvPre.tms[i].expansion = tmTemp.expansion;
	}
}











// Taylor model integration by only using Picard iteration
int Flowpipe::advance_picard(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);

	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;

		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);

		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}


	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(0,i,i);
			invS.set(1,i,i);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);


	TaylorModelVec x, x0;

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec c0_plus_Ar0;
	Ar0.add(c0_plus_Ar0, c0);

	x0 = c0_plus_Ar0;
	x = x0;

	for(int i=1; i<=order; ++i)
	{
		x.Picard_no_remainder_assign(x0, ode_centered, rangeDim+1, i, cutoff_threshold);
	}

	x.cutoff(cutoff_threshold);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i]; // + step_uncertainties[i];		// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);

		intDifferences.push_back(intTemp);
	}

	// add the uncertainties and the cutoff intervals onto the result
	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		// add the uncertainties and the cutoff intervals onto the result
		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{

			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

int Flowpipe::advance_picard(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), orders, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}


	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(0,i,i);
			invS.set(1,i,i);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);


	TaylorModelVec x, x0;

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec c_plus_Ar0;
	Ar0.add(c_plus_Ar0, c0);

	x0 = c_plus_Ar0;
	x = x0;

	for(int i=1; i<=globalMaxOrder; ++i)
	{
		x.Picard_no_remainder_assign(x0, ode_centered, rangeDim+1, i, cutoff_threshold);
	}

	x.nctrunc(orders);

	x.cutoff(cutoff_threshold);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i]; // + step_uncertainties[i];		// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, orders, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);

		intDifferences.push_back(intTemp);
	}

	// add the uncertainties and the cutoff intervals onto the result
	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		// add the uncertainties and the cutoff intervals onto the result
		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{

			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}


// adaptive step sizes and fixed orders
int Flowpipe::advance_picard(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;

		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}


	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(0, i, i);
			invS.set(1, i, i);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	TaylorModelVec x, x0;

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec c0_plus_Ar0;
	Ar0.add(c0_plus_Ar0, c0);

	x0 = c0_plus_Ar0;
	x = x0;

	for(int i=1; i<=order; ++i)
	{
		x.Picard_no_remainder_assign(x0, ode_centered, rangeDim+1, i, cutoff_threshold);
	}

	x.cutoff(cutoff_threshold);

	if(step > THRESHOLD_HIGH)		// step size is changed
	{
		construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);
	}

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];// + step_uncertainties[i];		// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	std::vector<Polynomial> polyDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;
		polyDifferences.push_back(polyTemp);

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	for(; !bfound;)
	{
		bfound = true;
		double newStep = step_exp_table[1].sup() * LAMBDA_DOWN;	// reduce the time step size

		if(newStep < miniStep)
		{
			return 0;
		}

		construct_step_exp_table(step_exp_table, step_end_exp_table, newStep, 2*order);

		x.polyRangeNormal(xPolyRange, step_exp_table);
		x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

		// recompute the interval evaluation of the polynomial differences
		for(int i=0; i<rangeDim; ++i)
		{
			polyDifferences[i].intEvalNormal(intDifferences[i], step_exp_table);
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				bfound = false;
				break;
			}
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

int Flowpipe::advance_picard(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), orders, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}


	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(0, i, i);
			invS.set(1, i, i);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	TaylorModelVec x, x0;

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec c_plus_Ar0;
	Ar0.add(c_plus_Ar0, c0);

	x0 = c_plus_Ar0;
	x = x0;

	for(int i=1; i<=globalMaxOrder; ++i)
	{
		x.Picard_no_remainder_assign(x0, ode_centered, rangeDim+1, i, cutoff_threshold);
	}

	x.nctrunc(orders);

	x.cutoff(cutoff_threshold);

	if(step > THRESHOLD_HIGH)		// step size is changed
	{
		construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);
	}

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];// + step_uncertainties[i];		// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, orders, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	std::vector<Polynomial> polyDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;
		polyDifferences.push_back(polyTemp);

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	for(; !bfound;)
	{
		bfound = true;
		double newStep = step_exp_table[1].sup() * LAMBDA_DOWN;	// reduce the time step size

		if(newStep < miniStep)
		{
			return -1;
		}

		construct_step_exp_table(step_exp_table, step_end_exp_table, newStep, 2*globalMaxOrder);

		x.polyRangeNormal(xPolyRange, step_exp_table);
		x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, orders, cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)
		{
			polyDifferences[i].intEvalNormal(intDifferences[i], step_exp_table);
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				bfound = false;
				break;
			}
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

// adaptive orders and fixed step sizes
int Flowpipe::advance_picard(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, int & order, const int maxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}


	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(0, i, i);
			invS.set(1, i, i);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);


	TaylorModelVec x, x0;

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec c0_plus_Ar0;
	Ar0.add(c0_plus_Ar0, c0);

	x0 = c0_plus_Ar0;
	x = x0;

	for(int i=1; i<=order; ++i)
	{
		x.Picard_no_remainder_assign(x0, ode_centered, rangeDim+1, i, cutoff_threshold);
	}

	x.cutoff(cutoff_threshold);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	int newOrder = order;

	for(; !bfound;)
	{
		++newOrder;

		if(newOrder > maxOrder)
		{
			return 0;
		}

		// increase the approximation orders by 1
		x.Picard_no_remainder_assign(x0, ode_centered, rangeDim+1, newOrder, cutoff_threshold);
		x.cutoff(cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)	// apply the estimation again
		{
			x.tms[i].remainder = estimation[i];// + step_uncertainties[i];
		}

		// compute the Picard operation again
		x.polyRangeNormal(xPolyRange, step_exp_table);
		x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, newOrder, cutoff_threshold);

		// Update the irreducible part
		for(int i=0; i<rangeDim; ++i)
		{
			Polynomial polyTemp;
			polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

			Interval intTemp;
			polyTemp.intEvalNormal(intTemp, step_exp_table);
			intDifferences[i] = intTemp;
		}

		// add the uncertainties onto the result
		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		bfound = true;
		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				bfound = false;
				break;
			}
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		// add the uncertainties
		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	order = newOrder;
	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

int Flowpipe::advance_picard(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, std::vector<int> & orders, const int localMaxOrder, const std::vector<int> & maxOrders, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), orders, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}


	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(0, i, i);
			invS.set(1, i, i);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);


	TaylorModelVec x, x0;

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec c_plus_Ar0;
	Ar0.add(c_plus_Ar0, c0);

	x0 = c_plus_Ar0;
	x = x0;

	for(int i=1; i<=localMaxOrder; ++i)
	{
		x.Picard_no_remainder_assign(x0, ode_centered, rangeDim+1, i, cutoff_threshold);
	}

	x.nctrunc(orders);

	x.cutoff(cutoff_threshold);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];// + step_uncertainties[i];		// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, orders, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	std::vector<bool> bIncrease;
	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bIncrease.push_back(true);
			if(bfound)
				bfound = false;
		}
		else
		{
			bIncrease.push_back(false);
		}
	}

	std::vector<int> newOrders = orders;
	bool bIncreaseOthers = false;
	int numIncrease = 0;

	std::vector<bool> bIncreased;
	for(int i=0; i<rangeDim; ++i)
	{
		bIncreased.push_back(false);
	}

	for(; !bfound;)
	{
		bool bChanged = false;

		if(bIncreaseOthers)
		{
			for(int i=0; i<bIncrease.size(); ++i)
			{
				if(!bIncrease[i] && newOrders[i] < maxOrders[i])
				{
					++newOrders[i];
					bIncreased[i] = true;

					if(!bChanged)
						bChanged = true;
				}
			}

			bIncreaseOthers = false;
		}
		else
		{
			numIncrease = 0;
			for(int i=0; i<bIncrease.size(); ++i)
			{
				if(bIncrease[i] && newOrders[i] < maxOrders[i])
				{
					++newOrders[i];
					++numIncrease;

					bIncreased[i] = true;

					if(!bChanged)
						bChanged = true;
				}
			}

			if(numIncrease < newOrders.size())
				bIncreaseOthers = true;
		}

		if(!bChanged)
		{
			return 0;
		}

		// increase the approximation orders
		x.Picard_no_remainder_assign(x0, ode_centered, rangeDim+1, newOrders, bIncreased, cutoff_threshold);
		x.cutoff(cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)
		{
			x.tms[i].remainder = estimation[i];
		}

		// compute the Picard operation again
		x.polyRangeNormal(xPolyRange, step_exp_table);
		x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, newOrders, cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)
		{
			// update the irreducible part if necessary
			if(bIncreased[i])
			{
				Polynomial polyTemp;
				polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

				Interval intTemp;
				polyTemp.intEvalNormal(intTemp, step_exp_table);
				intDifferences[i] = intTemp;
			}
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		bfound = true;

		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				if(!bIncreaseOthers)
				{
					bIncrease[i] = true;
					if(bfound)
						bfound = false;
				}
				else
				{
					bfound = false;
					break;
				}
			}
			else
			{
				if(!bIncreaseOthers)
				{
					bIncrease[i] = false;
				}
			}
		}

		for(int i=0; i<rangeDim; ++i)
		{
			bIncreased[i] = false;
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	orders = newOrders;
	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}


// for low-degree ODEs
// fixed step sizes and orders

int Flowpipe::advance_low_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & taylorExpansion, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);

	// the center point of the remainder
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}


	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(0, i, i);
			invS.set(1, i, i);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	TaylorModelVec x;

	for(int i=0; i<taylorExpansion.size(); ++i)
	{
		TaylorModel tmTemp;
		taylorExpansion[i].insert_no_remainder(tmTemp, x0, rangeDim+1, order, cutoff_threshold);
		x.tms.push_back(tmTemp);
	}

	x.cutoff(cutoff_threshold);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);

		intDifferences.push_back(intTemp);
	}

	// add the uncertainties and the cutoff intervals onto the result
	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders.push_back(tmvTemp.tms[i].remainder);
		}

		// add the uncertainties and the cutoff intervals onto the result
		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{

			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

int Flowpipe::advance_low_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & taylorExpansion, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const std::vector<int> & orders, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), orders, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}


	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(0, i, i);
			invS.set(1, i, i);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	TaylorModelVec x;

	for(int i=0; i<taylorExpansion.size(); ++i)
	{
		TaylorModel tmTemp;
		taylorExpansion[i].insert_no_remainder(tmTemp, x0, rangeDim+1, orders[i], cutoff_threshold);
		x.tms.push_back(tmTemp);
	}

	x.cutoff(cutoff_threshold);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];// + step_uncertainties[i];		// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, orders, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}


// adaptive step sizes and fixed orders
int Flowpipe::advance_low_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & taylorExpansion, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}


	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(0, i, i);
			invS.set(1, i, i);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	TaylorModelVec x;

	for(int i=0; i<taylorExpansion.size(); ++i)
	{
		TaylorModel tmTemp;
		taylorExpansion[i].insert_no_remainder(tmTemp, x0, rangeDim+1, order, cutoff_threshold);
		x.tms.push_back(tmTemp);
	}

	x.cutoff(cutoff_threshold);

	if(step > THRESHOLD_HIGH)		// step size is changed
	{
		construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);
	}

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	std::vector<Polynomial> polyDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;
		polyDifferences.push_back(polyTemp);

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	for(; !bfound;)
	{
		bfound = true;
		double newStep = step_exp_table[1].sup() * LAMBDA_DOWN;	// reduce the time step size

		if(newStep < miniStep)
		{
			return 0;
		}

		construct_step_exp_table(step_exp_table, step_end_exp_table, newStep, 2*order);

		x.polyRangeNormal(xPolyRange, step_exp_table);
		x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

		// recompute the interval evaluation of the polynomial differences
		for(int i=0; i<rangeDim; ++i)
		{
			polyDifferences[i].intEvalNormal(intDifferences[i], step_exp_table);
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				bfound = false;
				break;
			}
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

int Flowpipe::advance_low_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & taylorExpansion, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), orders, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}


	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(0, i, i);
			invS.set(1, i, i);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	TaylorModelVec x;

	for(int i=0; i<taylorExpansion.size(); ++i)
	{
		TaylorModel tmTemp;
		taylorExpansion[i].insert_no_remainder(tmTemp, x0, rangeDim+1, orders[i], cutoff_threshold);
		x.tms.push_back(tmTemp);
	}

	x.cutoff(cutoff_threshold);

	if(step > THRESHOLD_HIGH)		// step size is changed
	{
		construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);
	}

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];// + step_uncertainties[i];		// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, orders, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	std::vector<Polynomial> polyDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;
		polyDifferences.push_back(polyTemp);

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	for(; !bfound;)
	{
		bfound = true;
		double newStep = step_exp_table[1].sup() * LAMBDA_DOWN;	// reduce the time step size

		if(newStep < miniStep)
		{
			return 0;
		}

		construct_step_exp_table(step_exp_table, step_end_exp_table, newStep, 2*globalMaxOrder);

		x.polyRangeNormal(xPolyRange, step_exp_table);
		x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, orders, cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)
		{
			polyDifferences[i].intEvalNormal(intDifferences[i], step_exp_table);
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				bfound = false;
				break;
			}
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

// adaptive orders and fixed step sizes
int Flowpipe::advance_low_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const std::vector<HornerForm> & taylorExpansion, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, int & order, const int maxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}


	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(0, i, i);
			invS.set(1, i, i);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	TaylorModelVec x;

	for(int i=0; i<taylorExpansion.size(); ++i)
	{
		TaylorModel tmTemp;
		taylorExpansion[i].insert_no_remainder(tmTemp, x0, rangeDim+1, order, cutoff_threshold);
		x.tms.push_back(tmTemp);
	}

	x.cutoff(cutoff_threshold);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];// + step_uncertainties[i];		// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	int newOrder = order;

	for(; !bfound;)
	{
		++newOrder;

		if(newOrder > maxOrder)
		{
			return 0;
		}

		// increase the approximation orders by 1
		x.Picard_no_remainder_assign(x0, ode_centered, rangeDim+1, newOrder, cutoff_threshold);
		x.cutoff(cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)
		{
			x.tms[i].remainder = estimation[i];
		}

		// compute the Picard operation again
		x.polyRangeNormal(xPolyRange, step_exp_table);
		x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, newOrder, cutoff_threshold);

		// Update the irreducible part
		for(int i=0; i<rangeDim; ++i)
		{
			Polynomial polyTemp;
			polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

			Interval intTemp;
			polyTemp.intEvalNormal(intTemp, step_exp_table);
			intDifferences[i] = intTemp;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		bfound = true;
		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				bfound = false;
				break;
			}
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		// add the uncertainties
		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	order = newOrder;
	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

int Flowpipe::advance_low_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const std::vector<HornerForm> & taylorExpansion, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, std::vector<int> & orders, const std::vector<int> & maxOrders, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), orders, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}


	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(0, i, i);
			invS.set(1, i, i);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	TaylorModelVec x;

	for(int i=0; i<taylorExpansion.size(); ++i)
	{
		TaylorModel tmTemp;
		taylorExpansion[i].insert_no_remainder(tmTemp, x0, rangeDim+1, orders[i], cutoff_threshold);
		x.tms.push_back(tmTemp);
	}

	x.cutoff(cutoff_threshold);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, orders, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	std::vector<bool> bIncrease;
	for(int i=0; i<rangeDim; ++i)
	{
		if( !tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bIncrease.push_back(true);
			if(bfound)
				bfound = false;
		}
		else
		{
			bIncrease.push_back(false);
		}
	}

	std::vector<int> newOrders = orders;
	bool bIncreaseOthers = false;
	int numIncrease = 0;

	std::vector<bool> bIncreased;
	for(int i=0; i<rangeDim; ++i)
	{
		bIncreased.push_back(false);
	}

	for(; !bfound;)
	{
		bool bChanged = false;

		if(bIncreaseOthers)
		{
			for(int i=0; i<bIncrease.size(); ++i)
			{
				if(!bIncrease[i] && newOrders[i] < maxOrders[i])
				{
					++newOrders[i];
					bIncreased[i] = true;

					if(!bChanged)
						bChanged = true;
				}
			}

			bIncreaseOthers = false;
		}
		else
		{
			numIncrease = 0;
			for(int i=0; i<bIncrease.size(); ++i)
			{
				if(bIncrease[i] && newOrders[i] < maxOrders[i])
				{
					++newOrders[i];
					++numIncrease;

					bIncreased[i] = true;

					if(!bChanged)
						bChanged = true;
				}
			}

			if(numIncrease < newOrders.size())
				bIncreaseOthers = true;
		}

		if(!bChanged)
		{
			return 0;
		}

		// increase the approximation orders
		x.Picard_no_remainder_assign(x0, ode, rangeDim+1, newOrders, bIncreased, cutoff_threshold);
		x.cutoff(cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)
		{
			x.tms[i].remainder = estimation[i];
		}

		// compute the Picard operation again
		x.polyRangeNormal(xPolyRange, step_exp_table);
		x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, newOrders, cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)
		{
			// update the irreducible part if necessary
			if(bIncreased[i])
			{
				Polynomial polyTemp;
				polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

				Interval intTemp;
				polyTemp.intEvalNormal(intTemp, step_exp_table);
				intDifferences[i] = intTemp;
			}
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		bfound = true;

		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				if(!bIncreaseOthers)
				{
					bIncrease[i] = true;
					if(bfound)
						bfound = false;
				}
				else
				{
					bfound = false;
					break;
				}
			}
			else
			{
				if(!bIncreaseOthers)
				{
					bIncrease[i] = false;
				}
			}
		}

		for(int i=0; i<rangeDim; ++i)
		{
			bIncreased[i] = false;
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	orders = newOrders;
	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}























// for high-degree ODEs
// fixed step sizes and orders

int Flowpipe::advance_high_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}


	// vector storing the zero ranges
	std::vector<int> zeroIndices;

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(1,i,i);
			invS.set(1,i,i);

			zeroIndices.push_back(i+1);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	// compute the Taylor expansion of r(t)
	// since r(t) = A^{-1} * (x(t) - c(t)), we have that r'(t) = A^{-1} * (x'(t) - c'(t)) = A^{-1} * (f(c(t) + A*r(t), t) - c'(t))

	TaylorModelVec c = c0;
	for(int i=1; i<=order; ++i)
	{
		c.Picard_no_remainder_assign(c0, ode_centered, rangeDim+1, i, cutoff_threshold);	// compute c(t)
	}

	TaylorModelVec dcdt;
	c.derivative(dcdt, 0);	// compute dc/dt

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);
	TaylorModelVec Ar = Ar0;

	TaylorModelVec c_plus_Ar;
	Ar.add(c_plus_Ar, c);

	// compute the Taylor expansion of the ODE of A*r(t)
	TaylorModelVec Adrdt;
	for(int i=0; i<rangeDim; ++i)
	{
		TaylorModel tmTemp;
		ode_centered[i].insert_no_remainder(tmTemp, c_plus_Ar, rangeDim+1, order - 1, cutoff_threshold);
		Adrdt.tms.push_back(tmTemp);
	}

	Adrdt.sub_assign(dcdt);

	TaylorModelVec drdt;
	Adrdt.linearTrans(drdt, invA);

	TaylorModelVec taylorExp_Ar;

	computeTaylorExpansion(taylorExp_Ar, Adrdt, drdt, order, cutoff_threshold);

	// remove the zero terms
	taylorExp_Ar.rmZeroTerms(zeroIndices);
	Ar0.rmZeroTerms(zeroIndices);

	taylorExp_Ar.add_assign(Ar0);

	TaylorModelVec x = c;
	x.add_assign(taylorExp_Ar);		// the Taylor expansion of x(t)

	x.cutoff(cutoff_threshold);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);

		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{

			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}


	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		x.tms[i].remainder.remove_midpoint(M);
		Polynomial P(M, rangeDim+1);
		x.tms[i].expansion += P;
	}


	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

int Flowpipe::advance_high_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), orders, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}

	// vector storing the zero ranges
	std::vector<int> zeroIndices;

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(1, i, i);
			invS.set(1, i, i);

			zeroIndices.push_back(i+1);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	// compute the Taylor expansion of r(t)
	// since r(t) = A^{-1} * (x(t) - c(t)), we have that r'(t) = A^{-1} * (x'(t) - c'(t)) = A^{-1} * (f(c(t) + A*r(t), t) - c'(t))

	TaylorModelVec c = c0;
	for(int i=1; i<=globalMaxOrder; ++i)
	{
		c.Picard_no_remainder_assign(c0, ode_centered, rangeDim+1, i, cutoff_threshold);	// compute c(t)
	}

	c.nctrunc(orders);

	TaylorModelVec dcdt;
	c.derivative(dcdt, 0);	// compute dc/dt

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);
	TaylorModelVec Ar = Ar0;

	TaylorModelVec c_plus_Ar;
	Ar.add(c_plus_Ar, c);

	// compute the Taylor expansion of the ODE of A*r(t)
	TaylorModelVec Adrdt;
	for(int i=0; i<rangeDim; ++i)
	{
		TaylorModel tmTemp;
		ode_centered[i].insert_no_remainder(tmTemp, c_plus_Ar, rangeDim+1, orders[i] - 1, cutoff_threshold);
		Adrdt.tms.push_back(tmTemp);
	}

	Adrdt.sub_assign(dcdt);

	TaylorModelVec drdt;
	Adrdt.linearTrans(drdt, invA);

	TaylorModelVec taylorExp_Ar;

	computeTaylorExpansion(taylorExp_Ar, Adrdt, drdt, orders, cutoff_threshold);

	// remove the zero terms
	taylorExp_Ar.rmZeroTerms(zeroIndices);
	Ar0.rmZeroTerms(zeroIndices);

	taylorExp_Ar.add_assign(Ar0);

	TaylorModelVec x = c;
	x.add_assign(taylorExp_Ar);		// the Taylor expansion of x(t)

	x.cutoff(cutoff_threshold);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];// + step_uncertainties[i];		// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, orders, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

// adaptive step sizes and fixed orders
int Flowpipe::advance_high_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}

	// vector storing the zero ranges
	std::vector<int> zeroIndices;

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(1, i, i);
			invS.set(1, i, i);

			zeroIndices.push_back(i+1);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	// compute the Taylor expansion of r(t)
	// since r(t) = A^{-1} * (x(t) - c(t)), we have that r'(t) = A^{-1} * (x'(t) - c'(t)) = A^{-1} * (f(c(t) + A*r(t), t) - c'(t))

	TaylorModelVec c = c0;
	for(int i=1; i<=order; ++i)
	{
		c.Picard_no_remainder_assign(c0, ode_centered, rangeDim+1, i, cutoff_threshold);	// compute c(t)
	}

	TaylorModelVec dcdt;
	c.derivative(dcdt, 0);	// compute dc/dt

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);
	TaylorModelVec Ar = Ar0;

	TaylorModelVec c_plus_Ar;
	Ar.add(c_plus_Ar, c);

	// compute the Taylor expansion of the ODE of A*r(t)
	TaylorModelVec Adrdt;
	for(int i=0; i<rangeDim; ++i)
	{
		TaylorModel tmTemp;
		ode_centered[i].insert_no_remainder(tmTemp, c_plus_Ar, rangeDim+1, order - 1, cutoff_threshold);
		Adrdt.tms.push_back(tmTemp);
	}

	Adrdt.sub_assign(dcdt);

	TaylorModelVec drdt;
	Adrdt.linearTrans(drdt, invA);

	TaylorModelVec taylorExp_Ar;

	computeTaylorExpansion(taylorExp_Ar, Adrdt, drdt, order, cutoff_threshold);

	// remove the zero terms
	taylorExp_Ar.rmZeroTerms(zeroIndices);
	Ar0.rmZeroTerms(zeroIndices);

	taylorExp_Ar.add_assign(Ar0);

	TaylorModelVec x = c;
	x.add_assign(taylorExp_Ar);		// the Taylor expansion of x(t)

	x.cutoff(cutoff_threshold);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	if(step > THRESHOLD_HIGH)		// step size is changed
	{
		construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);
	}

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	std::vector<Polynomial> polyDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;
		polyDifferences.push_back(polyTemp);

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	for(; !bfound;)
	{
		bfound = true;
		double newStep = step_exp_table[1].sup() * LAMBDA_DOWN;	// reduce the time step size

		if(newStep < miniStep)
		{
			return 0;
		}

		construct_step_exp_table(step_exp_table, step_end_exp_table, newStep, 2*order);

		x.polyRangeNormal(xPolyRange, step_exp_table);
		x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

		// recompute the interval evaluation of the polynomial differences
		for(int i=0; i<rangeDim; ++i)
		{
			polyDifferences[i].intEvalNormal(intDifferences[i], step_exp_table);
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				bfound = false;
				break;
			}
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

int Flowpipe::advance_high_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), orders, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}

	// vector storing the zero ranges
	std::vector<int> zeroIndices;

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(1, i, i);
			invS.set(1, i, i);

			zeroIndices.push_back(i+1);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	// compute the Taylor expansion of r(t)
	// since r(t) = A^{-1} * (x(t) - c(t)), we have that r'(t) = A^{-1} * (x'(t) - c'(t)) = A^{-1} * (f(c(t) + A*r(t), t) - c'(t))

	TaylorModelVec c = c0;
	for(int i=1; i<=globalMaxOrder; ++i)
	{
		c.Picard_no_remainder_assign(c0, ode_centered, rangeDim+1, i, cutoff_threshold);	// compute c(t)
	}

	TaylorModelVec dcdt;
	c.derivative(dcdt, 0);	// compute dc/dt

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);
	TaylorModelVec Ar = Ar0;

	TaylorModelVec c_plus_Ar;
	Ar.add(c_plus_Ar, c);

	// compute the Taylor expansion of the ODE of A*r(t)
	TaylorModelVec Adrdt;
	for(int i=0; i<rangeDim; ++i)
	{
		TaylorModel tmTemp;
		ode_centered[i].insert_no_remainder(tmTemp, c_plus_Ar, rangeDim+1, orders[i] - 1, cutoff_threshold);
		Adrdt.tms.push_back(tmTemp);
	}

	Adrdt.sub_assign(dcdt);

	TaylorModelVec drdt;
	Adrdt.linearTrans(drdt, invA);

	TaylorModelVec taylorExp_Ar;

	computeTaylorExpansion(taylorExp_Ar, Adrdt, drdt, orders, cutoff_threshold);

	// remove the zero terms
	taylorExp_Ar.rmZeroTerms(zeroIndices);
	Ar0.rmZeroTerms(zeroIndices);

	taylorExp_Ar.add_assign(Ar0);

	TaylorModelVec x = c;
	x.add_assign(taylorExp_Ar);		// the Taylor expansion of x(t)

	x.cutoff(cutoff_threshold);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	if(step > THRESHOLD_HIGH)		// step size is changed
	{
		construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);
	}

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];// + step_uncertainties[i];		// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, orders, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	std::vector<Polynomial> polyDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;
		polyDifferences.push_back(polyTemp);

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	for(; !bfound;)
	{
		bfound = true;
		double newStep = step_exp_table[1].sup() * LAMBDA_DOWN;	// reduce the time step size

		if(newStep < miniStep)
		{
			return 0;
		}

		construct_step_exp_table(step_exp_table, step_end_exp_table, newStep, 2*globalMaxOrder);

		x.polyRangeNormal(xPolyRange, step_exp_table);
		x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, orders, cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)
		{
			polyDifferences[i].intEvalNormal(intDifferences[i], step_exp_table);
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				bfound = false;
				break;
			}
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}


// adaptive orders and fixed step sizes
int Flowpipe::advance_high_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, int & order, const int maxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}

	// vector storing the zero ranges
	std::vector<int> zeroIndices;

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(1, i, i);
			invS.set(1, i, i);

			zeroIndices.push_back(i+1);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	// compute the Taylor expansion of r(t)
	// since r(t) = A^{-1} * (x(t) - c(t)), we have that r'(t) = A^{-1} * (x'(t) - c'(t)) = A^{-1} * (f(c(t) + A*r(t), t) - c'(t))

	TaylorModelVec c = c0;
	for(int i=1; i<=order; ++i)
	{
		c.Picard_no_remainder_assign(c0, ode_centered, rangeDim+1, i, cutoff_threshold);	// compute c(t)
	}

	TaylorModelVec dcdt;
	c.derivative(dcdt, 0);	// compute dc/dt

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);
	TaylorModelVec Ar = Ar0;

	TaylorModelVec c_plus_Ar;
	Ar.add(c_plus_Ar, c);

	// compute the Taylor expansion of the ODE of A*r(t)
	TaylorModelVec Adrdt;
	for(int i=0; i<rangeDim; ++i)
	{
		TaylorModel tmTemp;
		ode_centered[i].insert_no_remainder(tmTemp, c_plus_Ar, rangeDim+1, order - 1, cutoff_threshold);
		Adrdt.tms.push_back(tmTemp);
	}

	Adrdt.sub_assign(dcdt);

	TaylorModelVec drdt;
	Adrdt.linearTrans(drdt, invA);

	TaylorModelVec taylorExp_Ar;

	computeTaylorExpansion(taylorExp_Ar, Adrdt, drdt, order, cutoff_threshold);

	// remove the zero terms
	taylorExp_Ar.rmZeroTerms(zeroIndices);
	Ar0.rmZeroTerms(zeroIndices);

	taylorExp_Ar.add_assign(Ar0);

	TaylorModelVec x = c;
	x.add_assign(taylorExp_Ar);		// the Taylor expansion of x(t)

	x.cutoff(cutoff_threshold);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	// add the uncertainties onto the reault
	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	int newOrder = order;

	for(; !bfound;)
	{
		++newOrder;

		if(newOrder > maxOrder)
		{
			return 0;
		}

		// increase the approximation orders by 1
		x.Picard_no_remainder_assign(x0, ode_centered, rangeDim+1, newOrder, cutoff_threshold);
		x.cutoff(cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)	// apply the estimation again
		{
			x.tms[i].remainder = estimation[i];// + step_uncertainties[i];
		}

		// compute the Picard operation again
		x.polyRangeNormal(xPolyRange, step_exp_table);
		x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, newOrder, cutoff_threshold);

		// Update the irreducible part
		for(int i=0; i<rangeDim; ++i)
		{
			Polynomial polyTemp;
			polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

			Interval intTemp;
			polyTemp.intEvalNormal(intTemp, step_exp_table);
			intDifferences[i] = intTemp;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		bfound = true;
		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				bfound = false;
				break;
			}
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	order = newOrder;
	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

int Flowpipe::advance_high_degree(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, std::vector<int> & orders, const int localMaxOrder, const std::vector<int> & maxOrders, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), orders, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}

	// vector storing the zero ranges
	std::vector<int> zeroIndices;

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(1, i, i);
			invS.set(1, i, i);

			zeroIndices.push_back(i+1);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	// compute the Taylor expansion of r(t)
	// since r(t) = A^{-1} * (x(t) - c(t)), we have that r'(t) = A^{-1} * (x'(t) - c'(t)) = A^{-1} * (f(c(t) + A*r(t), t) - c'(t))

	TaylorModelVec c = c0;
	for(int i=1; i<=localMaxOrder; ++i)
	{
		c.Picard_no_remainder_assign(c0, ode_centered, rangeDim+1, i, cutoff_threshold);	// compute c(t)
	}

	c.nctrunc(orders);

	TaylorModelVec dcdt;
	c.derivative(dcdt, 0);	// compute dc/dt

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);
	TaylorModelVec Ar = Ar0;

	TaylorModelVec c_plus_Ar;
	Ar.add(c_plus_Ar, c);

	// compute the Taylor expansion of the ODE of A*r(t)
	TaylorModelVec Adrdt;
	for(int i=0; i<rangeDim; ++i)
	{
		TaylorModel tmTemp;
		ode_centered[i].insert_no_remainder(tmTemp, c_plus_Ar, rangeDim+1, orders[i] - 1, cutoff_threshold);
		Adrdt.tms.push_back(tmTemp);
	}

	Adrdt.sub_assign(dcdt);

	TaylorModelVec drdt;
	Adrdt.linearTrans(drdt, invA);

	TaylorModelVec taylorExp_Ar;

	computeTaylorExpansion(taylorExp_Ar, Adrdt, drdt, orders, cutoff_threshold);

	// remove the zero terms
	taylorExp_Ar.rmZeroTerms(zeroIndices);
	Ar0.rmZeroTerms(zeroIndices);

	taylorExp_Ar.add_assign(Ar0);

	TaylorModelVec x = c;
	x.add_assign(taylorExp_Ar);		// the Taylor expansion of x(t)

	x.cutoff(cutoff_threshold);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];// + step_uncertainties[i];		// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, orders, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	std::vector<bool> bIncrease;
	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bIncrease.push_back(true);
			if(bfound)
				bfound = false;
		}
		else
		{
			bIncrease.push_back(false);
		}
	}

	std::vector<int> newOrders = orders;
	bool bIncreaseOthers = false;
	int numIncrease = 0;

	std::vector<bool> bIncreased;
	for(int i=0; i<rangeDim; ++i)
	{
		bIncreased.push_back(false);
	}

	for(; !bfound;)
	{
		bool bChanged = false;

		if(bIncreaseOthers)
		{
			for(int i=0; i<bIncrease.size(); ++i)
			{
				if(!bIncrease[i] && newOrders[i] < maxOrders[i])
				{
					++newOrders[i];
					bIncreased[i] = true;

					if(!bChanged)
						bChanged = true;
				}
			}

			bIncreaseOthers = false;
		}
		else
		{
			numIncrease = 0;
			for(int i=0; i<bIncrease.size(); ++i)
			{
				if(bIncrease[i] && newOrders[i] < maxOrders[i])
				{
					++newOrders[i];
					++numIncrease;

					bIncreased[i] = true;

					if(!bChanged)
						bChanged = true;
				}
			}

			if(numIncrease < newOrders.size())
				bIncreaseOthers = true;
		}

		if(!bChanged)
		{
			return 0;
		}

		// increase the approximation orders
		x.Picard_no_remainder_assign(x0, ode_centered, rangeDim+1, newOrders, bIncreased, cutoff_threshold);
		x.cutoff(cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)
		{
			x.tms[i].remainder = estimation[i];
		}

		// compute the Picard operation again
		x.polyRangeNormal(xPolyRange, step_exp_table);
		x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, newOrders, cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)
		{
			// update the irreducible part if necessary
			if(bIncreased[i])
			{
				Polynomial polyTemp;
				polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

				Interval intTemp;
				polyTemp.intEvalNormal(intTemp, step_exp_table);
				intDifferences[i] = intTemp;
			}
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		bfound = true;

		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				if(!bIncreaseOthers)
				{
					bIncrease[i] = true;
					if(bfound)
						bfound = false;
				}
				else
				{
					bfound = false;
					break;
				}
			}
			else
			{
				if(!bIncreaseOthers)
				{
					bIncrease[i] = false;
				}
			}
		}

		for(int i=0; i<rangeDim; ++i)
		{
			bIncreased[i] = false;
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	orders = newOrders;
	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}





// integration scheme for non-polynomial ODEs (using Taylor approximations)
// fixed step sizes and orders
int Flowpipe::advance_non_polynomial_taylor(Flowpipe & result, const std::vector<std::string> & strOde, const std::vector<std::string> & strOde_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = strOde.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);

	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}

	// vector storing the zero ranges
	std::vector<int> zeroIndices;

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(1, i, i);
			invS.set(1, i, i);

			zeroIndices.push_back(i+1);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	// compute the Taylor expansion of r(t)
	// since r(t) = A^{-1} * (x(t) - c(t)), we have that r'(t) = A^{-1} * (x'(t) - c'(t)) = A^{-1} * (f(c(t) + A*r(t), t) - c'(t))

	TaylorModelVec c = c0;
	for(int i=1; i<=order; ++i)
	{
		c.Picard_non_polynomial_taylor_no_remainder_assign(c0, strOde_centered, i, cutoff_threshold);
	}

	TaylorModelVec dcdt;
	c.derivative(dcdt, 0);	// compute dc/dt

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);
	TaylorModelVec Ar = Ar0;

	TaylorModelVec c_plus_Ar;
	Ar.add(c_plus_Ar, c);

	// compute the Taylor expansion of the ODE of A*r(t)

	parseSetting.clear();
	parseSetting.order = order - 1;
	parseSetting.flowpipe = c_plus_Ar;
	parseSetting.cutoff_threshold = cutoff_threshold;

	std::string prefix(str_prefix_taylor_polynomial);
	std::string suffix(str_suffix);

	TaylorModelVec Adrdt;
	for(int i=0; i<rangeDim; ++i)
	{
		parseSetting.strODE = prefix + strOde_centered[i] + suffix;

		parseODE();		// call the parser

		TaylorModel tmTemp(parseResult.expansion, intZero);
		Adrdt.tms.push_back(tmTemp);
	}

	Adrdt.sub_assign(dcdt);

	TaylorModelVec drdt;
	Adrdt.linearTrans(drdt, invA);

	TaylorModelVec taylorExp_Ar;

	computeTaylorExpansion(taylorExp_Ar, Adrdt, drdt, order, cutoff_threshold);

	// remove the zero terms
	taylorExp_Ar.rmZeroTerms(zeroIndices);
	Ar0.rmZeroTerms(zeroIndices);

	taylorExp_Ar.add_assign(Ar0);

	TaylorModelVec x = c;
	x.add_assign(taylorExp_Ar);		// the Taylor expansion of x(t)

	x.cutoff(cutoff_threshold);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	x.Picard_non_polynomial_taylor_ctrunc_normal(tmvTemp, x0, strOde, step_exp_table, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);

		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_non_polynomial_taylor_only_remainder(newRemainders, x0, strOde, step_exp_table[1], order);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders.push_back(tmvTemp.tms[i].remainder);
		}

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	return 1;
}

int Flowpipe::advance_non_polynomial_taylor(Flowpipe & result, const std::vector<std::string> & strOde, const std::vector<std::string> & strOde_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = strOde.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);

	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), orders, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}

	// vector storing the zero ranges
	std::vector<int> zeroIndices;

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(1, i, i);
			invS.set(1, i, i);

			zeroIndices.push_back(i+1);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	// compute the Taylor expansion of r(t)
	// since r(t) = A^{-1} * (x(t) - c(t)), we have that r'(t) = A^{-1} * (x'(t) - c'(t)) = A^{-1} * (f(c(t) + A*r(t), t) - c'(t))

	TaylorModelVec c = c0;
	for(int i=1; i<=globalMaxOrder; ++i)
	{
		c.Picard_non_polynomial_taylor_no_remainder_assign(c0, strOde_centered, i, cutoff_threshold);
	}

	TaylorModelVec dcdt;
	c.derivative(dcdt, 0);	// compute dc/dt

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);
	TaylorModelVec Ar = Ar0;

	TaylorModelVec c_plus_Ar;
	Ar.add(c_plus_Ar, c);

	// compute the Taylor expansion of the ODE of A*r(t)

	parseSetting.clear();
	parseSetting.flowpipe = c_plus_Ar;
	parseSetting.cutoff_threshold = cutoff_threshold;

	std::string prefix(str_prefix_taylor_polynomial);
	std::string suffix(str_suffix);

	TaylorModelVec Adrdt;
	for(int i=0; i<rangeDim; ++i)
	{
		parseSetting.strODE = prefix + strOde_centered[i] + suffix;
		parseSetting.order = orders[i] - 1;

		parseODE();		// call the parser

		TaylorModel tmTemp(parseResult.expansion, intZero);
		Adrdt.tms.push_back(tmTemp);
	}

	Adrdt.sub_assign(dcdt);

	TaylorModelVec drdt;
	Adrdt.linearTrans(drdt, invA);

	TaylorModelVec taylorExp_Ar;

	computeTaylorExpansion(taylorExp_Ar, Adrdt, drdt, orders, cutoff_threshold);

	// remove the zero terms
	taylorExp_Ar.rmZeroTerms(zeroIndices);
	Ar0.rmZeroTerms(zeroIndices);

	taylorExp_Ar.add_assign(Ar0);

	TaylorModelVec x = c;
	x.add_assign(taylorExp_Ar);		// the Taylor expansion of x(t)

	x.cutoff(cutoff_threshold);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];// + step_uncertainties[i];	// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	x.Picard_non_polynomial_taylor_ctrunc_normal(tmvTemp, x0, strOde, step_exp_table, orders, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);

		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_non_polynomial_taylor_only_remainder(newRemainders, x0, strOde, step_exp_table[1], orders);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders.push_back(tmvTemp.tms[i].remainder);
		}

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	return -1;
}


// adaptive step sizes and fixed orders
int Flowpipe::advance_non_polynomial_taylor(Flowpipe & result, const std::vector<std::string> & strOde, const std::vector<std::string> & strOde_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const int order, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = strOde.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);

	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}

	// vector storing the zero ranges
	std::vector<int> zeroIndices;

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(1, i, i);
			invS.set(1, i, i);

			zeroIndices.push_back(i+1);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	// compute the Taylor expansion of r(t)
	// since r(t) = A^{-1} * (x(t) - c(t)), we have that r'(t) = A^{-1} * (x'(t) - c'(t)) = A^{-1} * (f(c(t) + A*r(t), t) - c'(t))

	TaylorModelVec c = c0;
	for(int i=1; i<=order; ++i)
	{
		c.Picard_non_polynomial_taylor_no_remainder_assign(c0, strOde_centered, i, cutoff_threshold);
	}

	TaylorModelVec dcdt;
	c.derivative(dcdt, 0);	// compute dc/dt

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);
	TaylorModelVec Ar = Ar0;

	TaylorModelVec c_plus_Ar;
	Ar.add(c_plus_Ar, c);

	// compute the Taylor expansion of the ODE of A*r(t)

	parseSetting.clear();
	parseSetting.flowpipe = c_plus_Ar;
	parseSetting.order = order - 1;
	parseSetting.cutoff_threshold = cutoff_threshold;

	std::string prefix(str_prefix_taylor_polynomial);
	std::string suffix(str_suffix);

	TaylorModelVec Adrdt;
	for(int i=0; i<rangeDim; ++i)
	{
		parseSetting.strODE = prefix + strOde_centered[i] + suffix;

		parseODE();		// call the parser

		TaylorModel tmTemp(parseResult.expansion, intZero);
		Adrdt.tms.push_back(tmTemp);
	}

	Adrdt.sub_assign(dcdt);

	TaylorModelVec drdt;
	Adrdt.linearTrans(drdt, invA);

	TaylorModelVec taylorExp_Ar;

	computeTaylorExpansion(taylorExp_Ar, Adrdt, drdt, order, cutoff_threshold);

	// remove the zero terms
	taylorExp_Ar.rmZeroTerms(zeroIndices);
	Ar0.rmZeroTerms(zeroIndices);

	taylorExp_Ar.add_assign(Ar0);

	TaylorModelVec x = c;
	x.add_assign(taylorExp_Ar);		// the Taylor expansion of x(t)

	x.cutoff(cutoff_threshold);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	if(step > THRESHOLD_HIGH)		// step size is changed
	{
		construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);
	}

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	x.Picard_non_polynomial_taylor_ctrunc_normal(tmvTemp, x0, strOde, step_exp_table, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	std::vector<Polynomial> polyDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;
		polyDifferences.push_back(polyTemp);

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	for(; !bfound;)
	{
		bfound = true;
		double newStep = step_exp_table[1].sup() * LAMBDA_DOWN;	// reduce the time step size

		if(newStep < miniStep)
		{
			return 0;
		}

		construct_step_exp_table(step_exp_table, step_end_exp_table, newStep, 2*order);

		x.Picard_non_polynomial_taylor_ctrunc_normal(tmvTemp, x0, strOde, step_exp_table, order, cutoff_threshold);

		// recompute the interval evaluation of the polynomial differences
		for(int i=0; i<rangeDim; ++i)
		{
			polyDifferences[i].intEvalNormal(intDifferences[i], step_exp_table);
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				bfound = false;
				break;
			}
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_non_polynomial_taylor_only_remainder(newRemainders, x0, strOde, step_exp_table[1], order);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	return 1;
}

int Flowpipe::advance_non_polynomial_taylor(Flowpipe & result, const std::vector<std::string> & strOde, const std::vector<std::string> & strOde_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const double miniStep, const std::vector<int> & orders, const int globalMaxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = strOde.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);

	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), orders, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}

	// vector storing the zero ranges
	std::vector<int> zeroIndices;

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(1, i, i);
			invS.set(1, i, i);

			zeroIndices.push_back(i+1);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	// compute the Taylor expansion of r(t)
	// since r(t) = A^{-1} * (x(t) - c(t)), we have that r'(t) = A^{-1} * (x'(t) - c'(t)) = A^{-1} * (f(c(t) + A*r(t), t) - c'(t))

	TaylorModelVec c = c0;
	for(int i=1; i<=globalMaxOrder; ++i)
	{
		c.Picard_non_polynomial_taylor_no_remainder_assign(c0, strOde_centered, i, cutoff_threshold);
	}

	TaylorModelVec dcdt;
	c.derivative(dcdt, 0);	// compute dc/dt

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);
	TaylorModelVec Ar = Ar0;

	TaylorModelVec c_plus_Ar;
	Ar.add(c_plus_Ar, c);

	// compute the Taylor expansion of the ODE of A*r(t)

	parseSetting.clear();
	parseSetting.flowpipe = c_plus_Ar;
	parseSetting.cutoff_threshold = cutoff_threshold;

	std::string prefix(str_prefix_taylor_polynomial);
	std::string suffix(str_suffix);

	TaylorModelVec Adrdt;
	for(int i=0; i<rangeDim; ++i)
	{
		parseSetting.strODE = prefix + strOde_centered[i] + suffix;
		parseSetting.order = orders[i] - 1;

		parseODE();		// call the parser

		TaylorModel tmTemp(parseResult.expansion, intZero);
		Adrdt.tms.push_back(tmTemp);
	}

	Adrdt.sub_assign(dcdt);

	TaylorModelVec drdt;
	Adrdt.linearTrans(drdt, invA);

	TaylorModelVec taylorExp_Ar;

	computeTaylorExpansion(taylorExp_Ar, Adrdt, drdt, orders, cutoff_threshold);

	// remove the zero terms
	taylorExp_Ar.rmZeroTerms(zeroIndices);
	Ar0.rmZeroTerms(zeroIndices);

	taylorExp_Ar.add_assign(Ar0);

	TaylorModelVec x = c;
	x.add_assign(taylorExp_Ar);		// the Taylor expansion of x(t)

	x.cutoff(cutoff_threshold);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	if(step > THRESHOLD_HIGH)		// step size is changed
	{
		construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);
	}

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	x.Picard_non_polynomial_taylor_ctrunc_normal(tmvTemp, x0, strOde, step_exp_table, orders, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	std::vector<Polynomial> polyDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;
		polyDifferences.push_back(polyTemp);

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);
		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	for(; !bfound;)
	{
		bfound = true;
		double newStep = step_exp_table[1].sup() * LAMBDA_DOWN;	// reduce the time step size

		if(newStep < miniStep)
		{
			return 0;
		}

		construct_step_exp_table(step_exp_table, step_end_exp_table, newStep, 2*globalMaxOrder);

		x.Picard_non_polynomial_taylor_ctrunc_normal(tmvTemp, x0, strOde, step_exp_table, orders, cutoff_threshold);

		// recompute the interval evaluation of the polynomial differences
		for(int i=0; i<rangeDim; ++i)
		{
			polyDifferences[i].intEvalNormal(intDifferences[i], step_exp_table);
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				bfound = false;
				break;
			}
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_non_polynomial_taylor_only_remainder(newRemainders, x0, strOde, step_exp_table[1], orders);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	return 1;
}


// adaptive orders and fixed step sizes
int Flowpipe::advance_non_polynomial_taylor(Flowpipe & result, const std::vector<std::string> & strOde, const std::vector<std::string> & strOde_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, int & order, const int maxOrder, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = strOde.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);

	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}

	// vector storing the zero ranges
	std::vector<int> zeroIndices;

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(1, i, i);
			invS.set(1, i, i);

			zeroIndices.push_back(i+1);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	// compute the Taylor expansion of r(t)
	// since r(t) = A^{-1} * (x(t) - c(t)), we have that r'(t) = A^{-1} * (x'(t) - c'(t)) = A^{-1} * (f(c(t) + A*r(t), t) - c'(t))

	TaylorModelVec c = c0;
	for(int i=1; i<=order; ++i)
	{
		c.Picard_non_polynomial_taylor_no_remainder_assign(c0, strOde_centered, i, cutoff_threshold);
	}

	TaylorModelVec dcdt;
	c.derivative(dcdt, 0);	// compute dc/dt

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);
	TaylorModelVec Ar = Ar0;

	TaylorModelVec c_plus_Ar;
	Ar.add(c_plus_Ar, c);

	// compute the Taylor expansion of the ODE of A*r(t)

	parseSetting.clear();
	parseSetting.order = order - 1;
	parseSetting.flowpipe = c_plus_Ar;
	parseSetting.cutoff_threshold = cutoff_threshold;

	std::string prefix(str_prefix_taylor_polynomial);
	std::string suffix(str_suffix);

	TaylorModelVec Adrdt;
	for(int i=0; i<rangeDim; ++i)
	{
		parseSetting.strODE = prefix + strOde_centered[i] + suffix;

		parseODE();		// call the parser

		TaylorModel tmTemp(parseResult.expansion, intZero);
		Adrdt.tms.push_back(tmTemp);
	}

	Adrdt.sub_assign(dcdt);

	TaylorModelVec drdt;
	Adrdt.linearTrans(drdt, invA);

	TaylorModelVec taylorExp_Ar;

	computeTaylorExpansion(taylorExp_Ar, Adrdt, drdt, order, cutoff_threshold);

	// remove the zero terms
	taylorExp_Ar.rmZeroTerms(zeroIndices);
	Ar0.rmZeroTerms(zeroIndices);

	taylorExp_Ar.add_assign(Ar0);

	TaylorModelVec x = c;
	x.add_assign(taylorExp_Ar);		// the Taylor expansion of x(t)

	x.cutoff(cutoff_threshold);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	x.Picard_non_polynomial_taylor_ctrunc_normal(tmvTemp, x0, strOde, step_exp_table, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);

		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	int newOrder = order;

	for(; !bfound;)
	{
		++newOrder;

		if(newOrder > maxOrder)
		{
			return 0;
		}

		// increase the approximation orders by 1
		x.Picard_non_polynomial_taylor_no_remainder_assign(x0, strOde_centered, newOrder, cutoff_threshold);
		x.cutoff(cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)
		{
			x.tms[i].remainder = estimation[i];
		}

		// compute the Picard operation again
		x.Picard_non_polynomial_taylor_ctrunc_normal(tmvTemp, x0, strOde, step_exp_table, newOrder, cutoff_threshold);

		// Update the irreducible part
		for(int i=0; i<rangeDim; ++i)
		{
			Polynomial polyTemp;
			polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

			Interval intTemp;
			polyTemp.intEvalNormal(intTemp, step_exp_table);
			intDifferences[i] = intTemp;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		bfound = true;
		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				bfound = false;
				break;
			}
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_non_polynomial_taylor_only_remainder(newRemainders, x0, strOde, step_exp_table[1], newOrder);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	order = newOrder;
	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	return 1;
}

int Flowpipe::advance_non_polynomial_taylor(Flowpipe & result, const std::vector<std::string> & strOde, const std::vector<std::string> & strOde_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, std::vector<int> & orders, const int localMaxOrder, const std::vector<int> & maxOrders, const std::vector<Interval> & estimation, const std::vector<PolynomialConstraint> & invariant, const Interval & cutoff_threshold) const
{
	int rangeDim = strOde.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);

	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;
		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);
		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), orders, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	// contract the remainder part of the initial set
	if(invariant.size() > 0)
	{
		std::vector<Interval> polyRangeOfr0;
		result.tmv.polyRangeNormal(polyRangeOfr0, step_end_exp_table);

		Matrix matPre_constraints(rangeDim, rangeDim+1);

		switch(precondition)
		{
		case ID_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				matPre_constraints.set( 1 , i, i+1);
			}

			break;
		}

		case QR_PRE:
		{
			for(int i=0; i<rangeDim; ++i)
			{
				for(int j=0; j<rangeDim; ++j)
				{
					matPre_constraints.set( A.get(i,j) , i, j+1);
				}
			}

			break;
		}
		}

		TaylorModelVec tmvPrecond(matPre_constraints);
		tmvPrecond.add_assign(c0);

		std::vector<HornerForm> preconditioned_constraints;
		std::vector<Interval> b;
		for(int i=0; i<invariant.size(); ++i)
		{
			Polynomial polyTemp;
			invariant[i].hf.insert_no_remainder_no_cutoff(polyTemp, tmvPrecond, rangeDim+1);

			HornerForm hf;
			polyTemp.toHornerForm(hf);
			preconditioned_constraints.push_back(hf);

			b.push_back(invariant[i].B);
		}

		std::vector<Interval> contracted_remainders;
		for(int i=0; i<rangeDim; ++i)
		{
			contracted_remainders.push_back(result.tmv.tms[i].remainder);
		}

		int res = contract_remainder(polyRangeOfr0, contracted_remainders, preconditioned_constraints, b);

		if(res < 0)
		{
			return -1;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = contracted_remainders[i];
			boundOfr0.push_back(polyRangeOfr0[i] + contracted_remainders[i]);
		}
	}
	else
	{
		result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);
	}

	// vector storing the zero ranges
	std::vector<int> zeroIndices;

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(1, i, i);
			invS.set(1, i, i);

			zeroIndices.push_back(i+1);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	// compute the Taylor expansion of r(t)
	// since r(t) = A^{-1} * (x(t) - c(t)), we have that r'(t) = A^{-1} * (x'(t) - c'(t)) = A^{-1} * (f(c(t) + A*r(t), t) - c'(t))

	TaylorModelVec c = c0;
	for(int i=1; i<=localMaxOrder; ++i)
	{
		c.Picard_non_polynomial_taylor_no_remainder_assign(c0, strOde_centered, i, cutoff_threshold);
	}

	TaylorModelVec dcdt;
	c.derivative(dcdt, 0);	// compute dc/dt

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);
	TaylorModelVec Ar = Ar0;

	TaylorModelVec c_plus_Ar;
	Ar.add(c_plus_Ar, c);

	// compute the Taylor expansion of the ODE of A*r(t)

	parseSetting.clear();
	parseSetting.flowpipe = c_plus_Ar;
	parseSetting.cutoff_threshold = cutoff_threshold;

	std::string prefix(str_prefix_taylor_polynomial);
	std::string suffix(str_suffix);

	TaylorModelVec Adrdt;
	for(int i=0; i<rangeDim; ++i)
	{
		parseSetting.strODE = prefix + strOde_centered[i] + suffix;
		parseSetting.order = orders[i] - 1;

		parseODE();		// call the parser

		TaylorModel tmTemp(parseResult.expansion, intZero);
		Adrdt.tms.push_back(tmTemp);
	}

	Adrdt.sub_assign(dcdt);

	TaylorModelVec drdt;
	Adrdt.linearTrans(drdt, invA);

	TaylorModelVec taylorExp_Ar;

	computeTaylorExpansion(taylorExp_Ar, Adrdt, drdt, orders, cutoff_threshold);

	// remove the zero terms
	taylorExp_Ar.rmZeroTerms(zeroIndices);
	Ar0.rmZeroTerms(zeroIndices);

	taylorExp_Ar.add_assign(Ar0);

	TaylorModelVec x = c;
	x.add_assign(taylorExp_Ar);		// the Taylor expansion of x(t)

	x.cutoff(cutoff_threshold);

	TaylorModelVec x0;
	Ar0.add(x0, c0);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];// + step_uncertainties[i];	// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	x.Picard_non_polynomial_taylor_ctrunc_normal(tmvTemp, x0, strOde, step_exp_table, orders, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);

		intDifferences.push_back(intTemp);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	std::vector<bool> bIncrease;
	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bIncrease.push_back(true);
			if(bfound)
				bfound = false;
		}
		else
		{
			bIncrease.push_back(false);
		}
	}

	std::vector<int> newOrders = orders;
	bool bIncreaseOthers = false;
	int numIncrease = 0;

	std::vector<bool> bIncreased;
	for(int i=0; i<rangeDim; ++i)
	{
		bIncreased.push_back(false);
	}

	for(; !bfound;)
	{
		bool bChanged = false;

		if(bIncreaseOthers)
		{
			for(int i=0; i<bIncrease.size(); ++i)
			{
				if(!bIncrease[i] && newOrders[i] < maxOrders[i])
				{
					++newOrders[i];
					bIncreased[i] = true;

					if(!bChanged)
						bChanged = true;
				}
			}

			bIncreaseOthers = false;
		}
		else
		{
			numIncrease = 0;
			for(int i=0; i<bIncrease.size(); ++i)
			{
				if(bIncrease[i] && newOrders[i] < maxOrders[i])
				{
					++newOrders[i];
					++numIncrease;

					bIncreased[i] = true;

					if(!bChanged)
						bChanged = true;
				}
			}

			if(numIncrease < newOrders.size())
				bIncreaseOthers = true;
		}

		if(!bChanged)
		{
			return 0;
		}

		// increase the approximation orders
		x.Picard_non_polynomial_taylor_no_remainder_assign(x0, strOde_centered, newOrders, bIncreased, cutoff_threshold);
		x.cutoff(cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)
		{
			x.tms[i].remainder = estimation[i];
		}

		// compute the Picard operation again
		x.Picard_non_polynomial_taylor_ctrunc_normal(tmvTemp, x0, strOde, step_exp_table, newOrders, cutoff_threshold);

		for(int i=0; i<rangeDim; ++i)
		{
			// update the irreducible part if necessary
			if(bIncreased[i])
			{
				Polynomial polyTemp;
				polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

				Interval intTemp;
				polyTemp.intEvalNormal(intTemp, step_exp_table);
				intDifferences[i] = intTemp;
			}
		}

		for(int i=0; i<rangeDim; ++i)
		{
			tmvTemp.tms[i].remainder += intDifferences[i];
		}

		bfound = true;

		for(int i=0; i<rangeDim; ++i)
		{
			if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
			{
				if(!bIncreaseOthers)
				{
					bIncrease[i] = true;
					if(bfound)
						bfound = false;
				}
				else
				{
					bfound = false;
					break;
				}
			}
			else
			{
				if(!bIncreaseOthers)
				{
					bIncrease[i] = false;
				}
			}
		}

		for(int i=0; i<rangeDim; ++i)
		{
			bIncreased[i] = false;
		}
	}

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_non_polynomial_taylor_only_remainder(newRemainders, x0, strOde, step_exp_table[1], newOrders);

		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{
			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	orders = newOrders;
	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	return 1;
}




int Flowpipe::advance_picard_tt(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<Interval> & ode_rem, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);

	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_x0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfx0;
	result.tmv.intEvalNormal(boundOfx0, step_end_exp_table);


	// Compute the scaling matrix S.
	std::vector<Interval> S, invS;

	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfx0[i].mag(intSup);

		if(intSup.subseteq(intZero))
		{
			S.push_back(intZero);
			invS.push_back(intOne);
		}
		else
		{
			S.push_back(intSup);
			Interval intRecSup;
			intSup.rec(intRecSup);
			invS.push_back(intRecSup);
			boundOfx0[i] = intUnit;
		}
	}

	result.tmv.scale_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);


	TaylorModelVec x0(S);
	x0.add_assign(c0);
	TaylorModelVec x = x0;

	for(int i=1; i<=order; ++i)
	{
		x.Picard_no_remainder_assign(x0, ode, rangeDim+1, i, cutoff_threshold);
	}

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];		// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{

			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}










int Flowpipe::advance_picard_symbolic_remainder(Flowpipe & result, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const Interval & cutoff_threshold, Interval_matrix & trans_back, std::vector<Interval_matrix> & J, std::vector<Interval_matrix> & Phi_L) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// evaluate the the initial set x0
	TaylorModelVec range_of_x0;
	tmvPre.evaluate_t(range_of_x0, step_end_exp_table);

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);
/*
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}
*/
	TaylorModelVec c0(intVecCenter, rangeDim+1);


	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// decompose the linear and nonlinear part
	TaylorModelVec x0_linear, x0_other;
	range_of_x0.decompose(x0_linear, x0_other);

	std::vector<std::vector<Interval> > linear_coefficients;
	std::vector<Interval> zeroVec;
	for(int i=0; i<domain.size(); ++i)
	{
		zeroVec.push_back(intZero);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		linear_coefficients.push_back(zeroVec);
	}

	x0_linear.linearCoefficients(linear_coefficients);

	for(int i=0; i<rangeDim; ++i)
	{
		linear_coefficients[i].erase( linear_coefficients[i].begin() );
	}

	Interval_matrix Phi_L_i(linear_coefficients), local_trans_linear(linear_coefficients);

	Phi_L_i = Phi_L_i * trans_back;

	// compute the remainder part under the linear transformation
	Interval_matrix J_i(rangeDim, 1);

	for(int i=0; i<Phi_L.size(); ++i)
	{
		Phi_L[i] = Phi_L_i * Phi_L[i];
	}

	Phi_L.push_back(Phi_L_i);

	for(int i=1; i<Phi_L.size(); ++i)
	{
		J_i += Phi_L[i] * J[i-1];
	}

	Interval_matrix J_ip1(rangeDim, 1);

	// compute the local initial set
	if(J.size() > 0)
	{
		// compute the polynomial part under the linear transformation
		TaylorModelVec local_initial_linear;
		local_trans_linear.linear_trans(local_initial_linear, tmv);

		// compute the other part
		std::vector<Interval> tmvPolyRange;
		tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
		x0_other.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

		result.tmv.Remainder(J_ip1);

		Interval_matrix x0_rem;
		range_of_x0.Remainder(x0_rem);
		J_ip1 += x0_rem;

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].expansion += local_initial_linear.tms[i].expansion;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = J_ip1[i][0] + J_i[i][0];
		}
	}
	else
	{
		std::vector<Interval> tmvPolyRange;
		tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
		range_of_x0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);
		result.tmv.Remainder(J_ip1);
	}

	J.push_back(J_ip1);

	std::vector<Interval> boundOfx0;
	result.tmv.intEvalNormal(boundOfx0, step_exp_table);

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);

	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfx0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);

			Interval rec_intSup(1/dSup);
			trans_back.set(rec_intSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfx0[i] = intUnit;
		}
		else
		{
			S.set(0,i,i);
			trans_back.set(intZero, i, i);
			invS.set(1,i,i);
		}
	}

	// apply the scaling matrix S
	result.tmv.linearTrans_assign(invS);

	TaylorModelVec x, x0;

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( S.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec c0_plus_Ar0;
	Ar0.add(c0_plus_Ar0, c0);

	x0 = c0_plus_Ar0;
	x = x0;

	for(int i=1; i<=order; ++i)
	{
		x.Picard_no_remainder_assign(x0, ode_centered, rangeDim+1, i, cutoff_threshold);
	}

	x.cutoff(cutoff_threshold);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);

		intDifferences.push_back(intTemp);
	}

	// add the uncertainties and the cutoff intervals onto the result
	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		// add the uncertainties and the cutoff intervals onto the result
		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{

			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

int Flowpipe::advance_picard_hybridization_with_decomposition(Flowpipe & result, TaylorModelVec & x0_output, std::vector<Interval> & boundOfx0, const std::vector<HornerForm> & ode, const TaylorModelVec & c0, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const std::vector<std::vector<bool> > & substitution, const int order, const std::vector<Interval> & estimation, const Interval & cutoff_threshold, Interval_matrix & trans_back, std::vector<Interval_matrix> & J, std::vector<Interval_matrix> & Phi_L, const bool bsimp) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// t has already been replaced by delta
	TaylorModelVec range_of_x0 = tmvPre;

	// decompose the linear and nonlinear part
	TaylorModelVec x0_linear, x0_other;
	range_of_x0.decompose(x0_linear, x0_other);

	std::vector<std::vector<Interval> > linear_coefficients;
	std::vector<Interval> zeroVec;
	for(int i=0; i<domain.size(); ++i)
	{
		zeroVec.push_back(intZero);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		linear_coefficients.push_back(zeroVec);
	}

	x0_linear.linearCoefficients(linear_coefficients);

	for(int i=0; i<rangeDim; ++i)
	{
		linear_coefficients[i].erase( linear_coefficients[i].begin() );
	}

	Interval_matrix Phi_L_i(linear_coefficients), local_trans_linear(linear_coefficients);

	Phi_L_i = Phi_L_i * trans_back;

	// compute the remainder part under the linear transformation
	Interval_matrix J_i(rangeDim, 1);

	for(int i=0; i<Phi_L.size(); ++i)
	{
		Phi_L[i] = Phi_L_i * Phi_L[i];
	}

	Phi_L.push_back(Phi_L_i);

	for(int i=1; i<Phi_L.size(); ++i)
	{
		J_i += Phi_L[i] * J[i-1];
	}

	Interval_matrix J_ip1(rangeDim, 1);

	// compute the local initial set
	if(J.size() > 0)
	{
		// compute the polynomial part under the linear transformation
		TaylorModelVec local_initial_linear;
		local_trans_linear.linear_trans(local_initial_linear, tmv);

		// compute the other part
		std::vector<Interval> tmvPolyRange;
		tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
		x0_other.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

		if(!bsimp)
		{
			result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);
			result.tmv.substitute_with_precond(substitution, step_end_exp_table);
		}
		else
		{
			for(int i=0; i<result.tmv.tms.size(); ++i)
			{
				result.tmv.tms[i].expansion.simplification_in_decomposition(substitution[i]);
			}

			result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);
		}

		result.tmv.Remainder(J_ip1);

		Interval_matrix x0_rem;
		range_of_x0.Remainder(x0_rem);
		J_ip1 += x0_rem;

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].expansion += local_initial_linear.tms[i].expansion;
		}

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = J_ip1[i][0] + J_i[i][0];
		}
	}
	else
	{
		std::vector<Interval> tmvPolyRange;
		tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
		range_of_x0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

		if(!bsimp)
		{
			result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);
			result.tmv.substitute_with_precond(substitution, step_end_exp_table);
		}
		else
		{
			for(int i=0; i<result.tmv.tms.size(); ++i)
			{
				result.tmv.tms[i].expansion.simplification_in_decomposition(substitution[i]);
			}

			result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);
		}

		result.tmv.Remainder(J_ip1);
	}

	J.push_back(J_ip1);

	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);

	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfx0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			Interval rec_intSup;
			intSup.rec(rec_intSup);

			double dSup = intSup.sup();
			S.set(dSup, i, i);
			trans_back.set(rec_intSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfx0[i] = intUnit;
		}
		else
		{
			S.set(0,i,i);
			trans_back.set(intZero, i, i);
			invS.set(1,i,i);
		}
	}

	// apply the scaling matrix S
	result.tmv.linearTrans_assign(invS);

	TaylorModelVec x, x0;

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( S.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec c0_plus_Ar0;
	Ar0.add(c0_plus_Ar0, c0);

	x0 = c0_plus_Ar0;
	x0_output = x0;
	x = x0;

	for(int i=1; i<=order; ++i)
	{
		x.Picard_no_remainder_no_cutoff_assign(x0, ode, rangeDim+1, i);
	}

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal_no_cutoff(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order);

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= 1); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{

			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	x.cutoff_normal(step_exp_table, cutoff_threshold);

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();
	return 1;
}

int Flowpipe::refinement_picard_hybridization_with_decomposition(const std::vector<HornerForm> & ode, const TaylorModelVec & x0, const std::vector<std::vector<bool> > & substitution, std::vector<Interval> & step_exp_table, const int order, const std::vector<Interval> & estimation, const Interval & cutoff_threshold)
{
	int rangeDim = tmvPre.tms.size();

	TaylorModelVec x = x0;

	for(int i=1; i<=order; ++i)
	{
		x.Picard_no_remainder_assign(x0, ode, substitution, rangeDim+1, i, cutoff_threshold);
	}

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;
	std::vector<std::vector<Interval> > trunc_parts;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, trunc_parts, x0, xPolyRange, ode, substitution, step_exp_table, rangeDim+1, order, cutoff_threshold);

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, trunc_parts, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{

			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	x.cutoff_normal(step_exp_table, cutoff_threshold);

	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		x.tms[i].remainder.remove_midpoint(M);
		Polynomial P(M, rangeDim+1);
		x.tms[i].expansion += P;
	}

	tmvPre = x;

	trees.clear();
	return 1;
}
/*
int Flowpipe::refinement_picard_hybridization_with_decomposition(const TaylorModelVec & tmvOde, const std::vector<Interval> & hybrid_domain, const std::vector<int> & hybrid_dim, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const Interval & cutoff_threshold)
{
	// compute the new time-varying ODE


}
*/

/*
int Flowpipe::refinement_picard_hybridization_with_decomposition(const std::vector<HornerForm> & ode, const TaylorModelVec & x0, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const Interval & cutoff_threshold)
{
	int rangeDim = tmvPre.tms.size();

	TaylorModelVec x = x0;

	for(int i=1; i<=order; ++i)
	{
		x.Picard_no_remainder_no_cutoff_assign(x0, ode, rangeDim+1, i);
	}

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i];
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal_no_cutoff(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order);

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		for(int i=0; i<rangeDim; ++i)
		{

			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

//	x.cutoff_normal(step_exp_table, cutoff_threshold);

	tmvPre = x;

	trees.clear();
	return 1;
}
*/

int Flowpipe::advance_picard_suppfunc_remainder_hyb2(Flowpipe & result, const std::vector<Interval> & boundOfx0, const TaylorModelVec & ode, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const Interval & cutoff_threshold, const TaylorModelVec & initialSet, std::vector<Interval_matrix> & remainders, std::vector<Interval_matrix> & trans_on_local_initial) const
{
	int rangeDim = ode.tms.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	// compute the remainder part accumulated by the computed flowpipes
	std::vector<std::vector<Interval> > coefficients;
	std::vector<Interval> zeroVec;
	for(int i=0; i<domain.size(); ++i)
	{
		zeroVec.push_back(intZero);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		coefficients.push_back(zeroVec);
	}

	// decompose the linear and the other parts in x0
	TaylorModelVec x0_linear, x0_other;
	tmvPre.decompose(x0_linear, x0_other);

	std::vector<std::vector<Interval> > linear_coefficients = coefficients;
	x0_linear.linearCoefficients(linear_coefficients);

	for(int i=0; i<rangeDim; ++i)
	{
		linear_coefficients[i].erase( linear_coefficients[i].begin() );
	}

	Interval_matrix linear_trans(linear_coefficients);
	linear_trans.center();
/*
	// compute the remainder part under the linear transformation
	Interval_matrix remainder_from_computed_flowpipes(rangeDim, 1);

	for(int i=0; i<remainders.size(); ++i)
	{
		remainder_from_computed_flowpipes += trans_on_local_initial[i] * remainders[i];
	}

	for(int i=0; i<trans_on_local_initial.size(); ++i)
	{
		trans_on_local_initial[i] = linear_trans * trans_on_local_initial[i];
	}
*/
	Interval_matrix remainder_from_computed_flowpipes(rangeDim, 1);

	for(int i=0; i<trans_on_local_initial.size(); ++i)
	{
		trans_on_local_initial[i] = linear_trans * trans_on_local_initial[i];
	}

	trans_on_local_initial.push_back(linear_trans);

	for(int i=0; i<remainders.size(); ++i)
	{
		remainder_from_computed_flowpipes += trans_on_local_initial[i+1] * remainders[i];
	}


	Interval_matrix last_rem(rangeDim, 1);

	// compute the local initial set
	if(trans_on_local_initial.size() > 0)
	{
		// compute the polynomial part under the linear transformation

		tmvPre.insert_no_remainder_no_cutoff(result.tmv, tmv, rangeDim+1, order);

		Interval_matrix Im;
		tmvPre.Remainder(Im);
		last_rem += Im;

		for(int i=0; i<rangeDim; ++i)
		{
			result.tmv.tms[i].remainder = remainder_from_computed_flowpipes[i][0];
		}
	}
	else
	{
		std::vector<Interval> tmvPolyRange;
		tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
		tmvPre.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);
		result.tmv.Remainder(last_rem);
	}

	remainders.push_back(last_rem);

	std::vector<std::vector<Interval> > tempCoe = coefficients;
	ode.linearCoefficients(tempCoe);

	for(int i=0; i<rangeDim; ++i)
	{
		tempCoe[i].erase( tempCoe[i].begin() );
	}

	Interval_matrix A(tempCoe);

	// compute A^n for 1 <= n <= k
	std::vector<Interval_matrix> A_exp_table;
	compute_int_mat_pow(A_exp_table, A, order+1);

	// compute the expansion for exp(At)
	Polynomial_matrix expAt(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyOne(intOne, rangeDim+1);
		expAt[i][i] = polyOne;
	}

	for(int i=1; i<=order; ++i)
	{
		Polynomial_matrix A_i(A_exp_table[i], rangeDim+1);
		Polynomial t_i(0, i, rangeDim+1);

		A_i.mul_assign(t_i);
		A_i.mul_assign(factorial_rec[i]);

		expAt += A_i;
	}

	// compute the expansion for \int_0^t exp(-At)
	Polynomial_matrix int_exp_At(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial poly(0, 1, rangeDim+1);
		int_exp_At[i][i] = poly;
	}

	Interval intMOne(-1);
	for(int i=1; i<order; ++i)
	{
		Polynomial_matrix A_i(A_exp_table[i], rangeDim+1);
		Polynomial t_i(0, i+1, rangeDim+1);

		A_i.mul_assign(t_i);
		A_i.mul_assign(factorial_rec[i+1]);

		if(i%2 == 1)
		{
			A_i.mul_assign(intMOne);
		}

		int_exp_At += A_i;
	}

	Interval_matrix constant_part;
	ode.constant(constant_part);

	Polynomial_matrix constant_poly(rangeDim, 1);
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial poly(constant_part[i][0], rangeDim+1);
		constant_poly[i][0] = poly;
	}

	Polynomial_matrix pm_x0(rangeDim, 1);
	for(int i=1; i<=rangeDim; ++i)
	{
		Polynomial x_i(i, 1, rangeDim+1);
		pm_x0[i-1][0] = x_i;
	}

	// compute the interval Taylor series of the solution
	// x(t) = exp(At) x0 + exp(At) \int_0^t exp(-As) ds c
	Polynomial_matrix x_t = expAt * pm_x0 + expAt * int_exp_At * constant_poly;
	x_t.nctrunc(order);

	TaylorModelVec x;
	for(int i=0; i<rangeDim; ++i)
	{
		TaylorModel tm(x_t[i][0]);
		x.tms.push_back(tm);
	}

	// compute the remainder
	// compute the remainder for expAt wherein t in [0,delta]

	double max_A = A.max_norm();
	Interval intTemp(max_A);
	intTemp *= step_end_exp_table[1];
	intTemp.exp_assign();
	double maxNorm = intTemp.mag();

	Interval_matrix expAt_rem = A_exp_table[order+1];
	expAt_rem.mul_assign(step_exp_table[order+1]);
	expAt_rem.mul_assign(factorial_rec[order+1]);

	Interval remainder(-maxNorm, maxNorm);

	Interval_matrix interval_expAt(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			interval_expAt.set(remainder, i, j);
		}
	}

	expAt_rem *= interval_expAt;

	Interval_matrix interval_int_expAt = interval_expAt;
	interval_int_expAt.mul_assign(step_exp_table[1]);

	// compute the remainder for int_0^t exp(-As) ds wherein t in [0,delta]
	Interval_matrix int_expAt_rem = A_exp_table[order+1];
	int_expAt_rem.mul_assign(step_exp_table[order+2]);
	int_expAt_rem.mul_assign(factorial_rec[order+2]);
	int_expAt_rem *= interval_expAt;


	Interval_matrix remainder_x = expAt_rem * boundOfx0 + (interval_expAt * int_expAt_rem + expAt_rem * interval_int_expAt) * constant_part;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = remainder_x[i][0];
	}

	std::vector<Interval> ext_boundOfx0;
	ext_boundOfx0.push_back(step_exp_table[1]);
	for(int i=0; i<rangeDim; ++i)
	{
		ext_boundOfx0.push_back(boundOfx0[i]);
	}

	x.cutoff(ext_boundOfx0, cutoff_threshold);

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	return 1;
}




int Flowpipe::advance_picard_dyn_ref(Flowpipe & result, std::vector<Interval> & range, const std::vector<HornerForm> & ode, const std::vector<HornerForm> & ode_centered, const int precondition, std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const int order, const std::vector<Interval> & estimation, const Interval & cutoff_threshold) const
{
	int rangeDim = ode.size();
	Interval intZero, intOne(1,1), intUnit(-1,1);
	result.clear();

	TaylorModelVec range_of_x0 = tmvPre;

	// the center point of x0's polynomial part
	std::vector<Interval> intVecCenter;
	range_of_x0.constant(intVecCenter);

	// the center point of the remainder of x0
	for(int i=0; i<rangeDim; ++i)
	{
		Interval M;
		range_of_x0.tms[i].remainder.remove_midpoint(M);
		intVecCenter[i] += M;
	}

	TaylorModelVec c0(intVecCenter, rangeDim+1);

	// introduce a new variable r0 such that x0 = c0 + A*r0, then r0 is origin-centered
	range_of_x0.rmConstant();

	// compute the preconditioning matrix
	Matrix A(rangeDim,rangeDim), invA(rangeDim,rangeDim);
	TaylorModelVec range_of_r0;

	switch(precondition)
	{
	case ID_PRE:
	{
		range_of_r0 = range_of_x0;

		break;
	}

	case QR_PRE:
	{
		preconditionQR(A, range_of_x0, rangeDim, rangeDim+1);
		A.transpose(invA);
		range_of_x0.linearTrans(range_of_r0, invA);

		break;
	}
	}

	std::vector<Interval> tmvPolyRange;
	tmv.polyRangeNormal(tmvPolyRange, step_end_exp_table);
	range_of_r0.insert_ctrunc_normal(result.tmv, tmv, tmvPolyRange, step_end_exp_table, domain.size(), order, cutoff_threshold);

	std::vector<Interval> boundOfr0;

	result.tmv.intEvalNormal(boundOfr0, step_end_exp_table);


	// Compute the scaling matrix S.
	Matrix S(rangeDim, rangeDim);
	Matrix invS(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Interval intSup;
		boundOfr0[i].mag(intSup);

		if(!intSup.subseteq(intZero))
		{
			double dSup = intSup.sup();
			S.set(dSup, i, i);
			invS.set(1/dSup, i, i);
			boundOfr0[i] = intUnit;
		}
		else
		{
			S.set(0,i,i);
			invS.set(1,i,i);
		}
	}

	// apply the scaling matrix S, i.e., A S S^(-1) A^(-1), then the interval bound of r0 is normalized to [-1,1]^n

	switch(precondition)
	{
	case ID_PRE:
	{
		A = S;
		invA = invS;
		break;
	}

	case QR_PRE:
	{
		A = A * S;
		invA = invS * invA;
		break;
	}
	}

	result.tmv.linearTrans_assign(invS);
	result.tmv.cutoff_normal(step_end_exp_table, cutoff_threshold);

	TaylorModelVec x, x0;

	Matrix matCoefficients_Ar0(rangeDim, rangeDim+1);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			matCoefficients_Ar0.set( A.get(i,j) , i, j+1);
		}
	}

	TaylorModelVec Ar0(matCoefficients_Ar0);

	TaylorModelVec c0_plus_Ar0;
	Ar0.add(c0_plus_Ar0, c0);

	x0 = c0_plus_Ar0;
	x = x0;

	for(int i=1; i<=order; ++i)
	{
		x.Picard_no_remainder_assign(x0, ode_centered, rangeDim+1, i, cutoff_threshold);
	}

	x.cutoff(cutoff_threshold);

	bool bfound = true;

	for(int i=0; i<rangeDim; ++i)
	{
		x.tms[i].remainder = estimation[i]; // + step_uncertainties[i];		// apply the remainder estimation
	}

	TaylorModelVec tmvTemp;
	std::vector<RangeTree *> trees;

	std::vector<Interval> xPolyRange;
	x.polyRangeNormal(xPolyRange, step_exp_table);
	x.Picard_ctrunc_normal(tmvTemp, trees, x0, xPolyRange, ode, step_exp_table, rangeDim+1, order, cutoff_threshold);

	// compute the interval evaluation of the polynomial difference, this part is not able to be reduced by Picard iteration
	std::vector<Interval> intDifferences;
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyTemp;
		polyTemp = tmvTemp.tms[i].expansion - x.tms[i].expansion;

		Interval intTemp;
		polyTemp.intEvalNormal(intTemp, step_exp_table);

		intDifferences.push_back(intTemp);
	}

	// add the uncertainties and the cutoff intervals onto the result
	for(int i=0; i<rangeDim; ++i)
	{
		tmvTemp.tms[i].remainder += intDifferences[i];
	}

	for(int i=0; i<rangeDim; ++i)
	{
		if( ! tmvTemp.tms[i].remainder.subseteq(x.tms[i].remainder) )
		{
			bfound = false;
			break;
		}
	}

	if(!bfound)
	{
		return 0;
	}
	else
	{
		for(int i=0; i<rangeDim; ++i)
			x.tms[i].remainder = tmvTemp.tms[i].remainder;
	}

	bool bfinished = false;
	for(int rSteps = 0; !bfinished && (rSteps <= MAX_REFINEMENT_STEPS); ++rSteps)
	{
		bfinished = true;

		std::vector<Interval> newRemainders;
		x.Picard_only_remainder(newRemainders, trees, x0, ode, step_exp_table[1]);

		// add the uncertainties and the cutoff intervals onto the result
		for(int i=0; i<rangeDim; ++i)
		{
			newRemainders[i] += intDifferences[i];
		}

		for(int i=0; i<rangeDim; ++i)
		{

			if(newRemainders[i].subseteq(x.tms[i].remainder))
			{
				if(x.tms[i].remainder.widthRatio(newRemainders[i]) <= STOP_RATIO)
				{
					bfinished = false;
				}
			}
			else
			{
				bfinished = true;
				break;
			}

			x.tms[i].remainder = newRemainders[i];
		}
	}

	range.clear();
	for(int i=0; i<x.tms.size(); ++i)
	{
		range.push_back(xPolyRange[i] + x.tms[i].remainder);
//		range[i].dump(stdout);printf("\t");
	}
//	printf("\n");

	result.tmvPre = x;
	result.domain = domain;
	result.domain[0] = step_exp_table[1];

	trees.clear();


/*
	std::vector<Interval> flowpiperange;
	result.intEvalNormal(flowpiperange, step_exp_table, cutoff_threshold);
	for(int i=0; i<x.tms.size(); ++i)
	{
		flowpiperange[i].dump(stdout);printf("\t");
	}
	printf("\n");
*/
	return 1;
}

Flowpipe & Flowpipe::operator = (const Flowpipe & flowpipe)
{
	if(this == &flowpipe)
		return *this;

	tmvPre = flowpipe.tmvPre;
	tmv = flowpipe.tmv;
	domain = flowpipe.domain;
	return *this;
}








































// class Continuous_system

ContinuousSystem::ContinuousSystem()
{
}

ContinuousSystem::ContinuousSystem(const TaylorModelVec & ode_input, const Flowpipe & initialSet_input)
{
	int rangeDim = ode_input.tms.size();
	Interval intZero;

	initialSet = initialSet_input;
	tmvOde = ode_input;

	for(int i=0; i<rangeDim; ++i)
	{
		HornerForm hf;
		tmvOde.tms[i].expansion.toHornerForm(hf);
		hfOde.push_back(hf);
	}

	tmvOde_centered = ode_input;
	tmvOde_centered.center_nc();

	for(int i=0; i<rangeDim; ++i)
	{
		HornerForm hf;
		tmvOde_centered.tms[i].expansion.toHornerForm(hf);
		hfOde_centered.push_back(hf);
	}
}

ContinuousSystem::ContinuousSystem(const std::vector<std::string> & strOde_input, const Flowpipe & initialSet_input)
{
	strOde = strOde_input;

	std::string prefix(str_prefix_center);
	std::string suffix(str_suffix);

	for(int i=0 ;i<strOde.size(); ++i)
	{
		parseSetting.clear();
		parseSetting.strODE = prefix + strOde[i] + suffix;
		parseODE();

		strOde_centered.push_back(parseResult.strExpansion);
	}

	initialSet = initialSet_input;
}

ContinuousSystem::ContinuousSystem(const ContinuousSystem & system)
{
	tmvOde				= system.tmvOde;
	hfOde				= system.hfOde;
	initialSet			= system.initialSet;
	strOde				= system.strOde;
	tmvOde_centered		= system.tmvOde_centered;
	hfOde_centered		= system.hfOde_centered;
	strOde_centered		= system.strOde_centered;
}

ContinuousSystem::~ContinuousSystem()
{
	hfOde.clear();
	hfOde_centered.clear();
	strOde.clear();
	strOde_centered.clear();
}

void ContinuousSystem::reach_linear(std::list<Flowpipe> & results, const double step, const double time, const int order, const bool bPrint, const Interval & cutoff_threshold) const
{
	Interval intZero, intOne(1), intMOne(-1);
	int rangeDim = tmvOde.tms.size();

	std::vector<Interval> intVecTemp;
	intVecTemp.push_back(intZero);

	for(int i=0; i<rangeDim; ++i)
	{
		intVecTemp.push_back(intZero);
	}

	std::vector<std::vector<Interval> > linear_coefficients;
	for(int i=0; i<rangeDim; ++i)
	{
		linear_coefficients.push_back(intVecTemp);
	}

	tmvOde.linearCoefficients(linear_coefficients);

	// find the matrix A in dx/dt = Ax + u
	for(int i=0; i<rangeDim; ++i)
	{
		linear_coefficients[i].erase( linear_coefficients[i].begin() );
	}

	Interval_matrix A(linear_coefficients);

	// identity matrix
	Interval_matrix identity(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		identity.set(intOne, i, i);
	}

	// 2. Compute the first flowpipe
	// compute A^n for 1 <= n <= k
	std::vector<Interval_matrix> A_exp_table;
	compute_int_mat_pow(A_exp_table, A, order+1);

	// compute the expansion for exp(At)
	Polynomial_matrix expAt(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial polyOne(intOne, rangeDim+1);
		expAt[i][i] = polyOne;
	}

	for(int i=1; i<=order; ++i)
	{
		Polynomial_matrix A_i(A_exp_table[i], rangeDim+1);
		Polynomial t_i(0, i, rangeDim+1);

		A_i.mul_assign(t_i);
		A_i.mul_assign(factorial_rec[i]);

		expAt += A_i;
	}

	// compute the expansion for \int_0^t exp(-At)
	Polynomial_matrix int_exp_At(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial poly(0, 1, rangeDim+1);
		int_exp_At[i][i] = poly;
	}

	for(int i=1; i<order; ++i)
	{
		Polynomial_matrix A_i(A_exp_table[i], rangeDim+1);
		Polynomial t_i(0, i+1, rangeDim+1);

		A_i.mul_assign(t_i);
		A_i.mul_assign(factorial_rec[i+1]);

		if(i%2 == 1)
		{
			A_i.mul_assign(intMOne);
		}

		int_exp_At += A_i;
	}

	Interval_matrix constant_part;
	tmvOde.constant(constant_part);

	Polynomial_matrix constant_poly(rangeDim, 1);
	for(int i=0; i<rangeDim; ++i)
	{
		Polynomial poly(constant_part[i][0], rangeDim+1);
		constant_poly[i][0] = poly;
	}

	Polynomial_matrix pm_x0(rangeDim, 1);
	for(int i=1; i<=rangeDim; ++i)
	{
		Polynomial x_i(i, 1, rangeDim+1);
		pm_x0[i-1][0] = x_i;
	}

	// compute the interval Taylor series of the solution
	// x(t) = exp(At) x0 + exp(At) \int_0^t exp(-As) ds u
	Polynomial_matrix x_t = expAt * pm_x0 + expAt * int_exp_At * constant_poly;
	x_t.nctrunc(order);

	// compute the remainder for x
	std::vector<Interval> step_exp_table;
	std::vector<Interval> step_end_exp_table;
	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	double max_A = A.max_norm();
	Interval intTemp(max_A);
	intTemp *= step_end_exp_table[1];
	intTemp.exp_assign();
	double maxNorm = intTemp.mag();

	Interval_matrix expAt_rem = A_exp_table[order+1];
	expAt_rem.mul_assign(step_exp_table[order+1]);
	expAt_rem.mul_assign(factorial_rec[order+1]);

	Interval remainder(-maxNorm, maxNorm);

	Interval_matrix interval_expAt(rangeDim, rangeDim);
	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=0; j<rangeDim; ++j)
		{
			interval_expAt.set(remainder, i, j);
		}
	}

	expAt_rem *= interval_expAt;

	Interval_matrix interval_int_expAt = interval_expAt;
	interval_int_expAt.mul_assign(step_exp_table[1]);

	// compute the remainder for \int_0^t exp(-As) ds wherein t in [0,r]
	Interval_matrix int_expAt_rem = A_exp_table[order+1];
	int_expAt_rem.mul_assign(step_exp_table[order+2]);
	int_expAt_rem.mul_assign(factorial_rec[order+2]);
	int_expAt_rem *= interval_expAt;

	std::vector<Interval> polyRangeOfx0;
	initialSet.tmvPre.polyRangeNormal(polyRangeOfx0, step_exp_table);

	std::vector<Interval> rangeOfx0;
	for(int i=0; i<polyRangeOfx0.size(); ++i)
	{
		rangeOfx0.push_back(polyRangeOfx0[i] + initialSet.tmvPre.tms[i].remainder);
	}

	Interval_matrix imRangeOfx0(rangeOfx0);

	Interval_matrix remainder_x = expAt_rem * imRangeOfx0 + (interval_expAt * int_expAt_rem + expAt_rem * interval_int_expAt) * constant_part;

	TaylorModelVec x;
	for(int i=0; i<rangeDim; ++i)
	{
		TaylorModel tm(x_t[i][0], remainder_x[i][0]);
		x.tms.push_back(tm);
	}

	TaylorModelVec tmvTemp;
	x.insert_ctrunc_normal(tmvTemp, initialSet.tmvPre, polyRangeOfx0, step_exp_table, rangeDim+1, order, cutoff_threshold);

	tmvTemp.cutoff_normal(step_exp_table, cutoff_threshold);

	std::vector<Interval> newDomain = initialSet.domain;
	newDomain[0] = step_exp_table[1];

	Flowpipe firstFlowpipe(tmvTemp, initialSet.tmv, newDomain);

	results.clear();
	results.push_back(initialSet);
	results.push_back(firstFlowpipe);

	// 3. Compute the remaining flowpipes
	// compute the transformation exp(Ar)
	Interval_matrix expAr = identity;
	for(int i=1; i<=order; ++i)
	{
		Interval_matrix M = A_exp_table[i];
		M.mul_assign(step_end_exp_table[i].sup());
		M.mul_assign(factorial_rec[i]);
		expAr += M;
	}

	expAr += expAt_rem;

	// compute the transformation \int_0^r exp(-As) ds
	Interval_matrix int_expAr = identity;
	int_expAr.mul_assign(step_end_exp_table[1].sup());
	for(int i=1; i<order; ++i)
	{
		Interval_matrix M = A_exp_table[i];
		M.mul_assign(step_end_exp_table[i+1].sup());
		M.mul_assign(factorial_rec[i+1]);

		if(i%2 == 1)
		{
			M.mul_assign(intMOne);
		}

		int_expAr += M;
	}

	int_expAr += int_expAt_rem;

	// compute [exp(Ar)]^k for all 1 <= k <= T/r
	int N = time/step;
	std::vector<Interval_matrix> expAr_exp_table;
	compute_int_mat_pow(expAr_exp_table, expAr, N);

	Interval_matrix U = expAr * int_expAr * constant_part;
	Interval_matrix Ui = U;

	if(bPrint)
	{
		printf("time = %f,\t", step);
		printf("step = %f,\t", step);
		printf("order = %d\n", order);
	}

	for(int i=1; i<=N-1; ++i)
	{
		Flowpipe newFlowpipe;

		newFlowpipe.domain = firstFlowpipe.domain;
		newFlowpipe.tmv = firstFlowpipe.tmv;

		TaylorModelVec xi;
		expAr_exp_table[i].linear_trans(xi, firstFlowpipe.tmvPre);

		for(int j=0; j<rangeDim; ++j)
		{
			Interval I1;
			Interval I2 = Ui[j][0];
			I2.remove_midpoint(I1);

			Polynomial P(I1, rangeDim+1);
			xi.tms[j].expansion += P;
			xi.tms[j].remainder += I2;
		}

		Ui += expAr_exp_table[i] * U;

		xi.cutoff_normal(step_exp_table, cutoff_threshold);
		newFlowpipe.tmvPre = xi;

		results.push_back(newFlowpipe);

		if(bPrint)
		{
			printf("time = %f,\t", (i+1)*step);
			printf("step = %f,\t", step);
			printf("order = %d\n", order);
		}
	}
}

void ContinuousSystem::reach_picard(std::list<Flowpipe> & results, const double step, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_picard(newFlowpipe, hfOde, hfOde_centered, precondition, step_exp_table, step_end_exp_table, order, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("order = %d\n", order);
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_picard(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_picard(newFlowpipe, hfOde, hfOde_centered, precondition, step_exp_table, step_end_exp_table, orders, globalMaxOrder, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("orders:\t");
				int num = orders.size()-1;
				for(int i=0; i<num; ++i)
				{
					printf("%s : %d, ", stateVarNames[i].c_str(), orders[i]);
				}
				printf("%s : %d\n", stateVarNames[num].c_str(), orders[num]);
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

// adaptive step sizes and fixed orders
void ContinuousSystem::reach_picard(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	double newStep = 0;

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_picard(newFlowpipe, hfOde, hfOde_centered, precondition, step_exp_table, step_end_exp_table, newStep, miniStep, order, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step_exp_table[1].sup();

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step_exp_table[1].sup());
				printf("order = %d\n", order);
			}

			newStep = step_exp_table[1].sup() * LAMBDA_UP;
			if(newStep > step - THRESHOLD_HIGH)
			{
				newStep = 0;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_picard(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);

	double newStep = 0;

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_picard(newFlowpipe, hfOde, hfOde_centered, precondition, step_exp_table, step_end_exp_table, newStep, miniStep, orders, globalMaxOrder, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step_exp_table[1].sup();

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step_exp_table[1].sup());
				printf("orders:\t");
				int num = orders.size()-1;
				for(int i=0; i<num; ++i)
				{
					printf("%s : %d, ", stateVarNames[i].c_str(), orders[i]);
				}
				printf("%s : %d\n", stateVarNames[num].c_str(), orders[num]);
			}

			newStep = step_exp_table[1].sup() * LAMBDA_UP;
			if(newStep > step - THRESHOLD_HIGH)
			{
				newStep = 0;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

// adaptive orders and fixed step sizes
void ContinuousSystem::reach_picard(std::list<Flowpipe> & results, const double step, const double time, const int order, const int maxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*maxOrder);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	int newOrder = order;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_picard(newFlowpipe, hfOde, hfOde_centered, precondition, step_exp_table, step_end_exp_table, newOrder, maxOrder, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("order = %d\n", newOrder);
			}

			if(newOrder > order)
			{
				--newOrder;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_picard(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const std::vector<int> & maxOrders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);

	std::vector<int> newOrders = orders;

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int localMaxOrder = newOrders[0];
		for(int i=1; i<newOrders.size(); ++i)
		{
			if(localMaxOrder < newOrders[i])
				localMaxOrder = newOrders[i];
		}

		int res = currentFlowpipe.advance_picard(newFlowpipe, hfOde, hfOde_centered, precondition, step_exp_table, step_end_exp_table, newOrders, localMaxOrder, maxOrders, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("orders:\t");
				int num = newOrders.size()-1;
				for(int i=0; i<num; ++i)
				{
					printf("%s : %d, ", stateVarNames[i].c_str(), newOrders[i]);
				}
				printf("%s : %d\n", stateVarNames[num].c_str(), newOrders[num]);
			}

			for(int i=0; i<newOrders.size(); ++i)
			{
				if(newOrders[i] > orders[i])
					--newOrders[i];
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}


// fixed step sizes and orders
void ContinuousSystem::reach_low_degree(std::list<Flowpipe> & results, const double step, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<Polynomial> polyODE;
	for(int i=0; i<tmvOde_centered.tms.size(); ++i)
	{
		polyODE.push_back(tmvOde_centered.tms[i].expansion);
	}

	std::vector<HornerForm> taylorExpansion;
	computeTaylorExpansion(taylorExpansion, polyODE, order);

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_low_degree(newFlowpipe, hfOde, taylorExpansion, precondition, step_exp_table, step_end_exp_table, order, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("order = %d\n", order);
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_low_degree(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<Polynomial> polyODE;
	for(int i=0; i<tmvOde_centered.tms.size(); ++i)
	{
		polyODE.push_back(tmvOde_centered.tms[i].expansion);
	}

	std::vector<HornerForm> taylorExpansion;
	computeTaylorExpansion(taylorExpansion, polyODE, orders);

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_low_degree(newFlowpipe, hfOde, taylorExpansion, precondition, step_exp_table, step_end_exp_table, orders, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("orders:\t");
				int num = orders.size()-1;
				for(int i=0; i<num; ++i)
				{
					printf("%s : %d, ", stateVarNames[i].c_str(), orders[i]);
				}
				printf("%s : %d\n", stateVarNames[num].c_str(), orders[num]);
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

// adaptive step sizes and fixed orders
void ContinuousSystem::reach_low_degree(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	double newStep = 0;

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<Polynomial> polyODE;
	for(int i=0; i<tmvOde_centered.tms.size(); ++i)
	{
		polyODE.push_back(tmvOde_centered.tms[i].expansion);
	}

	std::vector<HornerForm> taylorExpansion;
	computeTaylorExpansion(taylorExpansion, polyODE, order);

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_low_degree(newFlowpipe, hfOde, taylorExpansion, precondition, step_exp_table, step_end_exp_table, newStep, miniStep, order, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step_exp_table[1].sup();

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step_exp_table[1].sup());
				printf("order = %d\n", order);
			}

			newStep = step_exp_table[1].sup() * LAMBDA_UP;

			double tDiffer = time - t;

			if(newStep > tDiffer)
			{
				newStep = tDiffer;
			}

			if(newStep > step - THRESHOLD_HIGH)
			{
				newStep = 0;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_low_degree(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);

	double newStep = 0;

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<Polynomial> polyODE;
	for(int i=0; i<tmvOde_centered.tms.size(); ++i)
	{
		polyODE.push_back(tmvOde_centered.tms[i].expansion);
	}

	std::vector<HornerForm> taylorExpansion;
	computeTaylorExpansion(taylorExpansion, polyODE, orders);

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_low_degree(newFlowpipe, hfOde, taylorExpansion, precondition, step_exp_table, step_end_exp_table, newStep, miniStep, orders, globalMaxOrder, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step_exp_table[1].sup();

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step_exp_table[1].sup());
				printf("orders:\t");
				int num = orders.size()-1;
				for(int i=0; i<num; ++i)
				{
					printf("%s : %d, ", stateVarNames[i].c_str(), orders[i]);
				}
				printf("%s : %d\n", stateVarNames[num].c_str(), orders[num]);
			}

			newStep = step_exp_table[1].sup() * LAMBDA_UP;
			if(newStep > step - THRESHOLD_HIGH)
			{
				newStep = 0;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

// adaptive orders and fixed step sizes
void ContinuousSystem::reach_low_degree(std::list<Flowpipe> & results, const double step, const double time, const int order, const int maxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*maxOrder);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	int newOrder = order;
	int currentMaxOrder = order;

	std::vector<Polynomial> polyODE;
	for(int i=0; i<tmvOde_centered.tms.size(); ++i)
	{
		polyODE.push_back(tmvOde_centered.tms[i].expansion);
	}

	std::vector<HornerForm> taylorExpansionHF;
	std::vector<Polynomial> taylorExpansionMF;
	std::vector<Polynomial> highestTerms;

	computeTaylorExpansion(taylorExpansionHF, taylorExpansionMF, highestTerms, polyODE, order);

	std::vector<std::vector<HornerForm> > expansions;
	expansions.push_back(taylorExpansionHF);

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_low_degree(newFlowpipe, hfOde, hfOde_centered, expansions[newOrder-order], precondition, step_exp_table, step_end_exp_table, newOrder, maxOrder, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("order = %d\n", newOrder);
			}

			if(newOrder > order)
			{
				--newOrder;

				if(newOrder > currentMaxOrder)
				{
					for(int i=currentMaxOrder; i<newOrder; ++i)
					{
						std::vector<HornerForm> newTaylorExpansionHF;
						std::vector<Polynomial> newTaylorExpansionMF;

						increaseExpansionOrder(newTaylorExpansionHF, newTaylorExpansionMF, highestTerms, taylorExpansionMF, polyODE, i);

						expansions.push_back(newTaylorExpansionHF);
						taylorExpansionMF = newTaylorExpansionMF;
					}

					currentMaxOrder = newOrder;
				}
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_low_degree(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const std::vector<int> & maxOrders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<int> newOrders = orders;
	std::vector<int> localMaxOrders = orders;

	std::vector<Polynomial> polyODE;
	for(int i=0; i<tmvOde_centered.tms.size(); ++i)
	{
		polyODE.push_back(tmvOde_centered.tms[i].expansion);
	}

	std::vector<HornerForm> taylorExpansionHF;
	std::vector<Polynomial> taylorExpansionMF;
	std::vector<Polynomial> highestTerms;

	computeTaylorExpansion(taylorExpansionHF, taylorExpansionMF, highestTerms, polyODE, orders);

	std::vector<std::vector<HornerForm> > expansions;
	std::vector<HornerForm> emptySet;
	for(int i=0; i<taylorExpansionHF.size(); ++i)
	{
		expansions.push_back(emptySet);
		expansions[i].push_back(taylorExpansionHF[i]);
	}

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_low_degree(newFlowpipe, hfOde, hfOde_centered, taylorExpansionHF, precondition, step_exp_table, step_end_exp_table, newOrders, maxOrders, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("orders:\t");
				int num = newOrders.size()-1;
				for(int i=0; i<num; ++i)
				{
					printf("%s : %d, ", stateVarNames[i].c_str(), newOrders[i]);
				}
				printf("%s : %d\n", stateVarNames[num].c_str(), newOrders[num]);
			}

			for(int i=0; i<newOrders.size(); ++i)
			{
				if(newOrders[i] > orders[i])
				{
					--newOrders[i];

					if(newOrders[i] > localMaxOrders[i])
					{
						for(int j=localMaxOrders[i]; j<newOrders[i]; ++j)
						{
							HornerForm newTaylorExpansionHF;
							Polynomial newTaylorExpansionMF;

							increaseExpansionOrder(newTaylorExpansionHF, newTaylorExpansionMF, highestTerms[i], taylorExpansionMF[i], polyODE, j);

							expansions[i].push_back(newTaylorExpansionHF);
							taylorExpansionMF[i] = newTaylorExpansionMF;
						}
					}

					localMaxOrders[i] = newOrders[i];

					taylorExpansionHF[i] = expansions[i][newOrders[i]-orders[i]];
				}
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}




// for high-degree ODEs
// fixed step sizes and orders

void ContinuousSystem::reach_high_degree(std::list<Flowpipe> & results, const double step, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_high_degree(newFlowpipe, hfOde, hfOde_centered, precondition, step_exp_table, step_end_exp_table, order, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("order = %d\n", order);
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_high_degree(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_high_degree(newFlowpipe, hfOde, hfOde_centered, precondition, step_exp_table, step_end_exp_table, orders, globalMaxOrder, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("orders:\t");
				int num = orders.size()-1;
				for(int i=0; i<num; ++i)
				{
					printf("%s : %d, ", stateVarNames[i].c_str(), orders[i]);
				}
				printf("%s : %d\n", stateVarNames[num].c_str(), orders[num]);
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

// adaptive step sizes and fixed orders
void ContinuousSystem::reach_high_degree(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	double newStep = 0;

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_high_degree(newFlowpipe, hfOde, hfOde_centered, precondition, step_exp_table, step_end_exp_table, newStep, miniStep, order, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step_exp_table[1].sup();

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step_exp_table[1].sup());
				printf("order = %d\n", order);
			}

			newStep = step_exp_table[1].sup() * LAMBDA_UP;
			if(newStep > step - THRESHOLD_HIGH)
			{
				newStep = 0;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_high_degree(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);

	double newStep = 0;

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_high_degree(newFlowpipe, hfOde, hfOde_centered, precondition, step_exp_table, step_end_exp_table, newStep, miniStep, orders, globalMaxOrder, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step_exp_table[1].sup();

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step_exp_table[1].sup());
				printf("orders:\t");
				int num = orders.size()-1;
				for(int i=0; i<num; ++i)
				{
					printf("%s : %d, ", stateVarNames[i].c_str(), orders[i]);
				}
				printf("%s : %d\n", stateVarNames[num].c_str(), orders[num]);
			}

			newStep = step_exp_table[1].sup() * LAMBDA_UP;
			if(newStep > step - THRESHOLD_HIGH)
			{
				newStep = 0;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

// adaptive orders and fixed step sizes
void ContinuousSystem::reach_high_degree(std::list<Flowpipe> & results, const double step, const double time, const int order, const int maxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*maxOrder);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	int newOrder = order;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_high_degree(newFlowpipe, hfOde, hfOde_centered, precondition, step_exp_table, step_end_exp_table, newOrder, maxOrder, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("order = %d\n", newOrder);
			}

			if(newOrder > order)
			{
				--newOrder;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_high_degree(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const std::vector<int> & maxOrders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);

	std::vector<int> newOrders = orders;

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int localMaxOrder = newOrders[0];
		for(int i=1; i<newOrders.size(); ++i)
		{
			if(localMaxOrder < newOrders[i])
				localMaxOrder = newOrders[i];
		}

		int res = currentFlowpipe.advance_high_degree(newFlowpipe, hfOde, hfOde_centered, precondition, step_exp_table, step_end_exp_table, newOrders, localMaxOrder, maxOrders, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("orders:\t");
				int num = newOrders.size()-1;
				for(int i=0; i<num; ++i)
				{
					printf("%s : %d, ", stateVarNames[i].c_str(), newOrders[i]);
				}
				printf("%s : %d\n", stateVarNames[num].c_str(), newOrders[num]);
			}

			for(int i=0; i<newOrders.size(); ++i)
			{
				if(newOrders[i] > orders[i])
					--newOrders[i];
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}













// for non-polynomial ODEs (using Taylor approximations)
// fixed step sizes and orders
void ContinuousSystem::reach_non_polynomial_taylor(std::list<Flowpipe> & results, const double step, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_non_polynomial_taylor(newFlowpipe, strOde, strOde_centered, precondition, step_exp_table, step_end_exp_table, order, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("order = %d\n", order);
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_non_polynomial_taylor(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_non_polynomial_taylor(newFlowpipe, strOde, strOde_centered, precondition, step_exp_table, step_end_exp_table, orders, globalMaxOrder, estimation, dummy_invariant, cutoff_threshold);

		if(res)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("orders:\t");
				int num = orders.size()-1;
				for(int i=0; i<num; ++i)
				{
					printf("%s : %d, ", stateVarNames[i].c_str(), orders[i]);
				}
				printf("%s : %d\n", stateVarNames[num].c_str(), orders[num]);
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

// adaptive step sizes and fixed orders
void ContinuousSystem::reach_non_polynomial_taylor(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	double newStep = 0;

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_non_polynomial_taylor(newFlowpipe, strOde, strOde_centered, precondition, step_exp_table, step_end_exp_table, newStep, miniStep, order, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step_exp_table[1].sup();

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step_exp_table[1].sup());
				printf("order = %d\n", order);
			}

			newStep = step_exp_table[1].sup() * LAMBDA_UP;
			if(newStep > step - THRESHOLD_HIGH)
			{
				newStep = 0;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_non_polynomial_taylor(std::list<Flowpipe> & results, const double step, const double miniStep, const double time, const std::vector<int> & orders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);

	double newStep = 0;

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_non_polynomial_taylor(newFlowpipe, strOde, strOde_centered, precondition, step_exp_table, step_end_exp_table, newStep, miniStep, orders, globalMaxOrder, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step_exp_table[1].sup();

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step_exp_table[1].sup());
				printf("orders:\t");
				int num = orders.size()-1;
				for(int i=0; i<num; ++i)
				{
					printf("%s : %d, ", stateVarNames[i].c_str(), orders[i]);
				}
				printf("%s : %d\n", stateVarNames[num].c_str(), orders[num]);
			}

			newStep = step_exp_table[1].sup() * LAMBDA_UP;
			if(newStep > step - THRESHOLD_HIGH)
			{
				newStep = 0;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}


// adaptive orders and fixed step sizes
void ContinuousSystem::reach_non_polynomial_taylor(std::list<Flowpipe> & results, const double step, const double time, const int order, const int maxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*maxOrder);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	int newOrder = order;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int res = currentFlowpipe.advance_non_polynomial_taylor(newFlowpipe, strOde, strOde_centered, precondition, step_exp_table, step_end_exp_table, newOrder, maxOrder, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("order = %d\n", newOrder);
			}

			if(newOrder > order)
			{
				--newOrder;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_non_polynomial_taylor(std::list<Flowpipe> & results, const double step, const double time, const std::vector<int> & orders, const std::vector<int> & maxOrders, const int globalMaxOrder, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*globalMaxOrder);

	std::vector<int> newOrders = orders;

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<PolynomialConstraint> dummy_invariant;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		int localMaxOrder = newOrders[0];
		for(int i=1; i<newOrders.size(); ++i)
		{
			if(localMaxOrder < newOrders[i])
				localMaxOrder = newOrders[i];
		}

		int res = currentFlowpipe.advance_non_polynomial_taylor(newFlowpipe, strOde, strOde_centered, precondition, step_exp_table, step_end_exp_table, newOrders, localMaxOrder, maxOrders, estimation, dummy_invariant, cutoff_threshold);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("orders:\t");
				int num = newOrders.size()-1;
				for(int i=0; i<num; ++i)
				{
					printf("%s : %d, ", stateVarNames[i].c_str(), newOrders[i]);
				}
				printf("%s : %d\n", stateVarNames[num].c_str(), newOrders[num]);
			}

			for(int i=0; i<newOrders.size(); ++i)
			{
				if(newOrders[i] > orders[i])
					--newOrders[i];
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_picard_symbolic_remainder(std::list<Flowpipe> & results, const double step, const double time, const int order, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold, const int N) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<Interval_matrix> J;
	std::vector<Interval_matrix> Phi_L;

	Interval intOne(1);
	int rangeDim = tmvOde.tms.size();
	Interval_matrix trans_back(rangeDim, rangeDim);

	for(int i=0; i<rangeDim; ++i)
	{
		trans_back.set(intOne, i, i);
	}

	Interval_matrix initial_trans = trans_back;

	int iterNum = 0;

	for(double t=THRESHOLD_HIGH; t < time; )
	{
		int res = currentFlowpipe.advance_picard_symbolic_remainder(newFlowpipe, hfOde, hfOde_centered, step_exp_table, step_end_exp_table, order, estimation, cutoff_threshold, trans_back, J, Phi_L);

		if(res == 1)
		{
			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("order = %d\n", order);
			}

			++iterNum;

			if(iterNum == N)
			{
				J.clear();
				Phi_L.clear();
				iterNum = 0;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_picard_hybridization_with_decomposition(std::list<Flowpipe> & results, const double step, const double time, const int order, const std::vector<Interval> & estimation, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold, const std::vector<std::vector<bool> > & substitution, const std::vector<int> & vars2sub, const std::vector<std::vector<int> > & subrule, const int M, const double sigma) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<Interval_matrix> J;
	std::vector<Interval_matrix> Phi_L;

	Interval intOne(1), intZero;
	int rangeDim = tmvOde.tms.size();
	Interval_matrix trans_back(rangeDim, rangeDim);

	for(int i=0; i<rangeDim; ++i)
	{
		trans_back.set(intOne, i, i);
	}

	int iterNum = 0;

	int maxIterNum2 = M/20;
	int iterNum2 = 0;

	bool bsimp = false;

	std::vector<Interval> hybridization_domain;

	for(int i=0; i<rangeDim; ++i)
	{
		hybridization_domain.push_back(intZero);
	}

	for(double t=THRESHOLD_HIGH; t < time; )
	{
		// 1.  update the dynamics
		// 1.1 compute an estimation of the hybridization domain
		TaylorModelVec new_tmvPre;
		currentFlowpipe.tmvPre.evaluate_t(new_tmvPre, step_end_exp_table);

		currentFlowpipe.tmvPre = new_tmvPre;

		std::vector<Interval> localInitialCenter;
		currentFlowpipe.tmvPre.constant(localInitialCenter);
		currentFlowpipe.tmvPre.rmConstant();

		std::vector<Interval> boundOfx0;
		std::vector<Interval> boundOfx0_centered;
		currentFlowpipe.intEvalNormal(boundOfx0_centered, step_end_exp_table, cutoff_threshold);

		for(int i=0; i<boundOfx0_centered.size(); ++i)
		{
			boundOfx0.push_back(boundOfx0_centered[i] + localInitialCenter[i]);
		}

		std::vector<Interval> local_domain;
		local_domain.push_back(step_exp_table[1]);
		for(int i=0; i<rangeDim; ++i)
		{
			Interval I = boundOfx0[i];
			I.bloat(sigma);
			local_domain.push_back(I);
		}

		// compute an estimation
		std::vector<Interval> ode_range;
		tmvOde.intEval(ode_range, local_domain);

		std::vector<int> directions;

		for(int i=0; i<vars2sub.size(); ++i)
		{
			int dim = vars2sub[i] - 1;
			Interval I = boundOfx0[dim];

			if(ode_range[dim].inf() > 0)		// derivative is positive
			{
				I.bloat_sup(sigma);
				directions.push_back(1);
			}
			else if(ode_range[dim].sup() < 0)	// derivative is negative
			{
				I.bloat_inf(sigma);
				directions.push_back(-1);
			}
			else
			{
				I.bloat(sigma);
				directions.push_back(0);
			}

			hybridization_domain[dim] = I;
		}

		TaylorModelVec c0(localInitialCenter, currentFlowpipe.domain.size());

		// 1.3 compute the decomposed ODE
		TaylorModelVec newOde;
		std::vector<HornerForm> newHfOde;

		for(int i=0; i<tmvOde.tms.size(); ++i)
		{
			std::vector<Interval> intVals;

			for(int j=0; j<subrule[i].size(); ++j)
			{
				intVals.push_back(hybridization_domain[subrule[i][j]-1]);
			}

			TaylorModel tmTemp;
			tmvOde.tms[i].substitute(tmTemp, subrule[i], intVals);
			newOde.tms.push_back(tmTemp);
		}

		for(int i=0; i<newOde.tms.size(); ++i)
		{
			HornerForm hf;
			newOde.tms[i].expansion.toHornerForm(hf);
			newHfOde.push_back(hf);
		}

		// 1.4 compute the next flowpipe under the decomposed dynamics
		TaylorModelVec x0;
		int res = currentFlowpipe.advance_picard_hybridization_with_decomposition(newFlowpipe, x0, boundOfx0_centered, newHfOde, c0, step_exp_table, step_end_exp_table, substitution, order, estimation, cutoff_threshold, trans_back, J, Phi_L, bsimp);

		if(res == 1)
		{
			// 2. validate the new flowpipe
			std::vector<Interval> box;

			newFlowpipe.intEvalNormal(box, vars2sub, step_exp_table, order, cutoff_threshold);

			bool bvalid = true;
			bool brefine = false;

			for(int i=0; i<vars2sub.size(); ++i)
			{
				int dim = vars2sub[i] - 1;

				if(directions[i] == 1)
				{
					if(! box[i].compare_upper(hybridization_domain[dim]) )
					{
						bvalid = false;
					}
					else
					{
						if((hybridization_domain[dim].sup() - box[i].sup())/hybridization_domain[dim].width() >= 1-ABS_STOP_RATIO)
						{
							brefine = true;
						}

						hybridization_domain[dim].setSup(box[i]);
					}
				}
				else if(directions[i] == -1)
				{
					if(! hybridization_domain[dim].compare_lower(box[i]) )
					{
						bvalid = false;
					}
					else
					{
						if((box[i].inf() - hybridization_domain[dim].inf())/hybridization_domain[dim].width() >= 1-ABS_STOP_RATIO)
						{
							brefine = true;
						}

						hybridization_domain[dim].setInf(box[i]);
					}
				}
				else
				{
					if( !box[i].subseteq(hybridization_domain[dim]) )
					{
						bvalid = false;
					}
					else
					{
						if(hybridization_domain[dim].widthRatio(box[i]) <= ABS_STOP_RATIO)
							brefine = true;

						hybridization_domain[dim] = box[i];
					}
				}
			}

			if(!bvalid)
			{
				printf("Too large stepsize.\n");
				return;
			}
			else if(brefine)
			{
				newFlowpipe.refinement_picard_hybridization_with_decomposition(hfOde, x0, substitution, step_exp_table, order, estimation, cutoff_threshold);
			}

			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("order = %d\n", order);
			}

			++iterNum;

			if(iterNum == M)
			{
				J.clear();
				Phi_L.clear();
				iterNum = 0;
			}

			if(iterNum2 >= maxIterNum2)
			{
				iterNum2 = 0;
				bsimp = true;
			}
			else
			{
				++iterNum2;
				bsimp = false;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}




void ContinuousSystem::reach_picard_dynamical_hybridization2(std::list<Flowpipe> & results, const double step, const double time, const int order, const bool bPrint, const std::vector<std::string> & stateVarNames, const Interval & cutoff_threshold, const std::vector<int> & vars2sub, const std::vector<std::vector<int> > & subrule, const int N, const double epsilon, const int M) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<Interval_matrix> remainders;
	std::vector<Interval_matrix> trans_on_local_initial;

	Interval intZero, intOne(1);
	int rangeDim = tmvOde.tms.size();
	Interval_matrix trans_back(rangeDim, rangeDim);

	for(int i=0; i<rangeDim; ++i)
	{
		trans_back.set(intOne, i, i);
	}

	Interval_matrix initial_trans = trans_back;

	TaylorModelVec initial_tmv = initialSet.tmv;
	int iterNum = 0;

	std::vector<Interval> hyb_domain;

	for(double t=THRESHOLD_HIGH; t < time; )
	{
		// 1.  update the dynamics
		// 1.1 compute an estimation of the hybridization domain
		TaylorModelVec new_tmvPre;
		currentFlowpipe.tmvPre.evaluate_t(new_tmvPre, step_end_exp_table);
		currentFlowpipe.tmvPre = new_tmvPre;

		std::vector<Interval> boundOfx0;

		currentFlowpipe.intEvalNormal(boundOfx0, step_end_exp_table, cutoff_threshold);

		// bloat the range to an abstract domain
		std::vector<Interval> local_domain;
		local_domain.push_back(step_exp_table[1]);
		for(int i=0; i<rangeDim; ++i)
		{
			Interval I = boundOfx0[i];
			I.bloat(epsilon);
			local_domain.push_back(I);
		}

		// evaluate the ODE value
		std::vector<Interval> ode_range;
		tmvOde.intEval(ode_range, local_domain);

		hyb_domain.clear();

		std::vector<int> directions;

		for(int i=0, j=0; i<rangeDim; ++i)
		{
			if(j < vars2sub.size() && i == vars2sub[j]-1)
			{
				Interval I = boundOfx0[i];

				if(ode_range[j].inf() >= 0)
				{
					I.bloat_sup(epsilon);
					directions.push_back(1);
				}
				else if(ode_range[j].sup() <= 0)
				{
					I.bloat_inf(epsilon);
					directions.push_back(-1);
				}
				else
				{
					I.bloat(epsilon);
					directions.push_back(0);
				}

				hyb_domain.push_back(I);
				++j;
			}
			else
			{
				hyb_domain.push_back(intZero);	// dummy
				directions.push_back(-100);
			}
		}


		// 1.2 compute the simplified dynamics
		TaylorModelVec newOde;
		bool brefining = false;
		int numRef = 0;

ODE:	for(int i=0; i<tmvOde.tms.size(); ++i)
		{
			std::vector<Interval> intVals;

			for(int j=0; j<subrule[i].size(); ++j)
			{
				intVals.push_back(hyb_domain[subrule[i][j]-1]);
			}

			TaylorModel tmTemp;
			tmvOde.tms[i].substitute(tmTemp, subrule[i], intVals);
			newOde.tms.push_back(tmTemp);
		}


		// 1.3 compute the next flowpipe under the new dynamics

		int res = currentFlowpipe.advance_picard_suppfunc_remainder_hyb2(newFlowpipe, boundOfx0, newOde, step_exp_table, step_end_exp_table, order, cutoff_threshold, initial_tmv, remainders, trans_on_local_initial);

		if(res == 1)
		{
			// 2. check the validity of the current flowpipe
			std::vector<Interval> box;
			newFlowpipe.intEvalNormal(box, step_exp_table, cutoff_threshold);

			bool bvalid = true;
			bool brefine = false;

			for(int i=0; i<vars2sub.size(); ++i)
			{
				int j = vars2sub[i]-1;

				if(directions[i] == 1)
				{
					double d1 = box[i].sup();
					double d2 = hyb_domain[j].sup();
					if(d1 > d2)
					{
						if(!brefining)
						{
							bvalid = false;
							break;
						}
						else
						{
							break;
						}
					}
					else
					{
						if(std::abs((d2 - d1)/hyb_domain[j].width()) >= 0.01)
							brefine = true;

						hyb_domain[j].setSup(d1);
					}
				}
				else if(directions[i] == -1)
				{
					double d1 = box[i].inf();
					double d2 = hyb_domain[j].inf();
					if(d1 < d2)
					{
						if(!brefining)
						{
							bvalid = false;
							break;
						}
						else
						{
							break;
						}
					}
					else
					{
						if(std::abs((d1 - d2)/hyb_domain[j].width()) >= 0.01)
							brefine = true;

						hyb_domain[j].setInf(d1);
					}
				}
				else
				{
					if( !box[i].subseteq(hyb_domain[j]) )
					{
						if(!brefining)
						{
							bvalid = false;
							break;
						}
						else
						{
							break;
						}
					}
					else
					{
						if(hyb_domain[j].widthRatio(box[i]) <= 0.99)
							brefine = true;

						hyb_domain[j] = box[i];
					}
				}

			}

			if(!bvalid)
			{
				printf("Too large stepsize.\n");
				return;
			}
			else if(brefine && numRef < M)
			{
				// refine the current ODE
				newOde.clear();

				trans_on_local_initial.pop_back();
				remainders.pop_back();
				brefining = true;
				++numRef;

				goto ODE;
			}

			results.push_back(newFlowpipe);
			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("order = %d\n", order);
			}

			brefining = false;
			++iterNum;

			if(iterNum == N)
			{
				// update the initial set and the remainder records
				remainders.clear();
				trans_on_local_initial.clear();

				TaylorModelVec newTmvPre, newTmvFp;

				initial_tmv = currentFlowpipe.tmv;

				iterNum = 0;
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- The remainder estimation is not large enough.\n");
			break;
		}
	}
}





void ContinuousSystem::reach_picard_dyn_ref(std::list<Flowpipe> & results, const std::vector<int> & varIDs, const double epsilon, const double step, const double time, const int order, const int precondition, const std::vector<Interval> & estimation, const bool bPrint, const Interval & cutoff_threshold) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;

	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	results.clear();
	results.push_back(initialSet);
	Flowpipe newFlowpipe, currentFlowpipe = initialSet;

	std::vector<Interval> abs_domain;
double ep = 1e-5;
	initialSet.intEvalNormal(abs_domain, step_exp_table, cutoff_threshold);
//	enlarge(abs_domain, epsilon);

	std::vector<Interval> intVals;
	for(int i=0; i<varIDs.size(); ++i)
	{
		// test
		Interval M;
		abs_domain[varIDs[i]-1].midpoint(M);
		printf("width = %e\n", abs_domain[varIDs[i]-1].width());
		M.bloat(ep);
		intVals.push_back(M);
	}

	std::vector<HornerForm> new_hfOde = hfOde;
	std::vector<HornerForm> new_hfOde_centered = hfOde_centered;

	for(double t=THRESHOLD_HIGH; t < time;)
	{
		// update the dynamics
		TaylorModelVec newOde;

		for(int i=0; i<tmvOde.tms.size(); ++i)
		{
//			if(i != 5)
//			{
				TaylorModel tmTemp;
				tmvOde.tms[i].substitute(tmTemp, varIDs, intVals);
				newOde.tms.push_back(tmTemp);
//			}
//			else
//			{
//				newOde.tms.push_back(tmvOde.tms[i]);
//			}
		}

		for(int i=0; i<newOde.tms.size(); ++i)
		{
			HornerForm hf;
			newOde.tms[i].expansion.toHornerForm(hf);
			new_hfOde[i] = hf;
		}
/*
		std::string vt("t");
		std::string vx("x");
		std::string vy("y");
		std::vector<std::string> vnames;
		vnames.push_back(vt);
		vnames.push_back(vx);
		vnames.push_back(vy);
		newOde.dump_interval(stdout, vnames, vnames);
*/
		newOde.center_nc();

//		newOde.dump_interval(stdout, vnames, vnames);

		for(int i=0; i<newOde.tms.size(); ++i)
		{
			HornerForm hf;
			newOde.tms[i].expansion.toHornerForm(hf);
			new_hfOde_centered[i] = hf;
		}

		bool bfinish = false;
		bool bvalid = true;
		int res;
		for(int k=1; !bfinish && bvalid; ++k)
		{
			// compute a flowpipe in the abstract domain
			std::vector<Interval> new_abs_domain;
			res = currentFlowpipe.advance_picard_dyn_ref(newFlowpipe, new_abs_domain, new_hfOde, new_hfOde_centered, precondition, step_exp_table, step_end_exp_table, order, estimation, cutoff_threshold);

			// refine the abstract domain as well as the dynamics until no big improvement can be made
			bfinish = true;
/*
			if(res == 1)
			{
				// refine the abstract domain
				for(int i=0; i<new_abs_domain.size(); ++i)
				{
					if(new_abs_domain[i].subseteq(abs_domain[i]))
					{
						if(abs_domain[i].widthRatio(new_abs_domain[i]) <= ABS_STOP_RATIO)
						{
							bfinish = false;
						}

						abs_domain[i] = new_abs_domain[i];
					}
					else
					{
//						abs_domain[i].dump(stdout);printf("\t");
//						new_abs_domain[i].dump(stdout);printf("\ndimension = %d\n", i);
						bvalid = false;
						break;
					}
				}

				if(bvalid)
				{
					// update the dynamics
					for(int i=0; i<varIDs.size(); ++i)
					{
						intVals[i] = abs_domain[varIDs[i]-1];
					}

					TaylorModelVec newOde;
					tmvOde.substitute(newOde, varIDs, intVals);

					for(int i=0; i<newOde.tms.size(); ++i)
					{
						HornerForm hf;
						newOde.tms[i].expansion.toHornerForm(hf);
						new_hfOde[i] = hf;
					}

					newOde.center_nc();

					for(int i=0; i<newOde.tms.size(); ++i)
					{
						HornerForm hf;
						newOde.tms[i].expansion.toHornerForm(hf);
						new_hfOde_centered[i] = hf;
					}
				}
			}
			else
			{
				fprintf(stdout, "Terminated -- Too small remainder estimation.\n");
				bvalid = false;
				break;
			}
*/
		}

		if(res != 1)
		{
			fprintf(stdout, "Terminated -- Too small remainder estimation.\n");
			break;
		}

		if(bvalid)
		{
			results.push_back(newFlowpipe);

			// evaluate the initial set in the next step
			TaylorModelVec tmv;
			newFlowpipe.tmvPre.evaluate_t(tmv, step_end_exp_table);
			newFlowpipe.tmvPre = tmv;

			newFlowpipe.intEvalNormal(abs_domain, step_end_exp_table, cutoff_threshold);
//			enlarge(abs_domain, epsilon);
			for(int i=0; i<varIDs.size(); ++i)
			{
				Interval M;
				abs_domain[varIDs[i]-1].midpoint(M);
				printf("width = %e\n", abs_domain[varIDs[i]-1].width());
				M.bloat(ep);
				intVals[i] = M;
//				intVals[i].dump(stdout);printf("\n");
			}

			currentFlowpipe = newFlowpipe;

			t += step;

			if(bPrint)
			{
				printf("time = %f,\t", t);
				printf("step = %f,\t", step);
				printf("order = %d\n", order);
			}
		}
		else
		{
			fprintf(stdout, "Terminated -- Too small remainder estimation.\n");
			break;
		}
	}
}

void ContinuousSystem::reach_linear_dynamical_hybridization(std::vector<TaylorModelVec> & flowpipes, std::vector<Interval> & domain, const std::vector<int> & varIDs, const double step, const double epsilon, const double time, const int order, const bool bPrint, const Interval & cutoff_threshold, const int N) const
{
	// transform the initial set to a Taylor model
	int rangeDim = tmvOde.tms.size();

	TaylorModelVec X0 = initialSet.tmvPre;
	std::vector<Interval> X0_domain = initialSet.domain;

	Polynomial_matrix X0_poly(rangeDim, 1);
	Interval_matrix X0_rem(rangeDim, 1);
	for(int i=0; i<rangeDim; ++i)
	{
		X0_poly[i][0] = X0.tms[i].expansion;
		X0_rem[i][0] = X0.tms[i].remainder;
	}

	flowpipes.push_back(X0);
	domain = X0_domain;
	domain[0].setSup(step);

	Interval intZero, intOne(1), intMOne(-1);

	std::vector<Interval> intVecTemp;
	intVecTemp.push_back(intZero);

	for(int i=0; i<rangeDim; ++i)
	{
		intVecTemp.push_back(intZero);
	}

	std::vector<std::vector<Interval> > coefficients;
	for(int i=0; i<rangeDim; ++i)
	{
		coefficients.push_back(intVecTemp);
	}

	std::vector<Interval> step_exp_table;
	std::vector<Interval> step_end_exp_table;
	construct_step_exp_table(step_exp_table, step_end_exp_table, step, 2*order);

	std::vector<Interval_matrix> remainders;
	std::vector<Interval_matrix> constants;
	std::vector<Interval_matrix> trans_on_local_initial;

	double t = THRESHOLD_HIGH;

	Interval_matrix initial_rem = X0_rem;

	TaylorModelVec local_initial = X0;

	int iterNum = 0;

	while(t < time)
	{
		// 1.  update the ODE according to the local abstraction
		// 1.1 evaluate an interval enclosure of the local initial set
		std::vector<Interval> localInitial_range;
		std::vector<Interval> polyRange;
		local_initial.polyRangeNormal(polyRange, step_exp_table);

		for(int i=0; i<rangeDim; ++i)
		{
			localInitial_range.push_back(polyRange[i] + local_initial.tms[i].remainder);
		}


		// 1.2 bloat the range to an abstract domain
		std::vector<Interval> local_domain;
		local_domain.push_back(step_exp_table[1]);
		for(int i=0; i<rangeDim; ++i)
		{
			Interval I = localInitial_range[i];
			I.bloat(epsilon);
			local_domain.push_back(I);
		}

		// evaluate the ODE value
		std::vector<Interval> ode_range;
		tmvOde.intEval(ode_range, local_domain, varIDs);

		std::vector<int> directions;
		std::vector<Interval> intVals;
		for(int i=0; i<varIDs.size(); ++i)
		{
			// need a more clever way to bloat the interval domain
			Interval I = localInitial_range[varIDs[i]-1];

			if(ode_range[i].inf() >= 0)
			{
				I.bloat_sup(epsilon);
				directions.push_back(1);
			}
			else if(ode_range[i].sup() <= 0)
			{
				I.bloat_inf(epsilon);
				directions.push_back(-1);
			}
			else
			{
				I.bloat(epsilon);
				directions.push_back(0);
			}

			intVals.push_back(I);
		}

		// 1.3 compute the linear ODE for the current abstract domain
		TaylorModelVec newOde;
		std::vector<HornerForm> new_hfOde;



ODE:	for(int i=0; i<tmvOde.tms.size(); ++i)
		{
			TaylorModel tmTemp;
			tmvOde.tms[i].substitute(tmTemp, varIDs, intVals);
			newOde.tms.push_back(tmTemp);
		}

		for(int i=0; i<newOde.tms.size(); ++i)
		{
			HornerForm hf;
			newOde.tms[i].expansion.toHornerForm(hf);
			new_hfOde.push_back(hf);
		}

		// 2.  Compute the local flowpipe
		// 2.1 Obtain the matrix A in dx/dt = Ax + u
		std::vector<std::vector<Interval> > tempCoe = coefficients;
		newOde.linearCoefficients(tempCoe);

		for(int i=0; i<rangeDim; ++i)
		{
			tempCoe[i].erase( tempCoe[i].begin() );
		}

		Interval_matrix A(tempCoe);

		// compute A^n for 1 <= n <= k
		std::vector<Interval_matrix> A_exp_table;
		compute_int_mat_pow(A_exp_table, A, order+1);

		// compute the expansion for exp(At)
		Polynomial_matrix expAt(rangeDim, rangeDim);
		for(int i=0; i<rangeDim; ++i)
		{
			Polynomial polyOne(intOne, rangeDim+1);
			expAt[i][i] = polyOne;
		}

		for(int i=1; i<=order; ++i)
		{
			Polynomial_matrix A_i(A_exp_table[i], rangeDim+1);
			Polynomial t_i(0, i, rangeDim+1);

			A_i.mul_assign(t_i);
			A_i.mul_assign(factorial_rec[i]);

			expAt += A_i;
		}

		// compute the expansion for \int_0^t exp(-At)
		Polynomial_matrix int_exp_At(rangeDim, rangeDim);
		for(int i=0; i<rangeDim; ++i)
		{
			Polynomial poly(0, 1, rangeDim+1);
			int_exp_At[i][i] = poly;
		}

		Interval intMOne(-1);
		for(int i=1; i<order; ++i)
		{
			Polynomial_matrix A_i(A_exp_table[i], rangeDim+1);
			Polynomial t_i(0, i+1, rangeDim+1);

			A_i.mul_assign(t_i);
			A_i.mul_assign(factorial_rec[i+1]);

			if(i%2 == 1)
			{
				A_i.mul_assign(intMOne);
			}

			int_exp_At += A_i;
		}

		Interval_matrix constant_part;
		newOde.constant(constant_part);

		Polynomial_matrix constant_poly(rangeDim, 1);
		for(int i=0; i<rangeDim; ++i)
		{
			Polynomial poly(constant_part[i][0], rangeDim+1);
			constant_poly[i][0] = poly;
		}


		Polynomial_matrix pm_x0(rangeDim, 1);
		for(int i=1; i<=rangeDim; ++i)
		{
			Polynomial x_i(i, 1, rangeDim+1);
			pm_x0[i-1][0] = x_i;
		}

		// compute the interval Taylor series of the solution
		// x(t) = exp(At) x0 + exp(At) \int_0^t exp(-As) ds c
		Polynomial_matrix x_t = expAt * pm_x0 + expAt * int_exp_At * constant_poly;
		x_t.nctrunc(order);

		TaylorModelVec x;
		for(int i=0; i<rangeDim; ++i)
		{
			TaylorModel tm(x_t[i][0]);
			x.tms.push_back(tm);
		}

		// compute the remainder
		// compute the remainder for expAt wherein t in [0,delta]
/*
		Interval_matrix p_expAdelta(rangeDim, rangeDim);
		for(int i=0; i<rangeDim; ++i)
		{
			p_expAdelta[i][i] = intOne;
		}

		for(int i=1; i<=order; ++i)
		{
			Interval_matrix tmpM = A_exp_table[i];
			tmpM.mul_assign(step_end_exp_table[i]);
			p_expAdelta += tmpM;
		}
*/
		double max_A = A.max_norm();
		Interval intTemp(max_A);
		intTemp *= step_end_exp_table[1];
		intTemp.exp_assign();
		double maxNorm = intTemp.mag();

//		double maxNorm2 = p_expAdelta.max_norm();



		Interval_matrix expAt_rem = A_exp_table[order+1];
		expAt_rem.mul_assign(step_exp_table[order+1]);
		expAt_rem.mul_assign(factorial_rec[order+1]);




		Interval remainder(-maxNorm, maxNorm);

		Interval_matrix interval_expAt(rangeDim, rangeDim);
		for(int i=0; i<rangeDim; ++i)
		{
			for(int j=0; j<rangeDim; ++j)
			{
				interval_expAt.set(remainder, i, j);
			}
		}

		expAt_rem *= interval_expAt;

		Interval_matrix interval_int_expAt = interval_expAt;
		interval_int_expAt.mul_assign(step_exp_table[1]);

		// compute the remainder for int_0^t exp(-As) ds wherein t in [0,delta]
		Interval_matrix int_expAt_rem = A_exp_table[order+1];
		int_expAt_rem.mul_assign(step_exp_table[order+2]);
		int_expAt_rem.mul_assign(factorial_rec[order+2]);
		int_expAt_rem *= interval_expAt;


		Interval_matrix remainder_x = expAt_rem * localInitial_range + (interval_expAt * int_expAt_rem + expAt_rem * interval_int_expAt) * constant_part;





		// compute the TM with all coefficients narrowed down
/*
		for(int i=0; i<rangeDim; ++i)
		{
			x.tms[i].remainder = remainder_x[i][0];
		}
*/
		// compute the local linear mapping
		TaylorModelVec next_initial;
		x.evaluate_t(next_initial, step_end_exp_table);

		std::vector<std::vector<Interval> > step_trans = coefficients;
		next_initial.linearCoefficients(step_trans);

		for(int i=0; i<rangeDim; ++i)
		{
			step_trans[i].erase( step_trans[i].begin() );
		}

		Interval_matrix stepTrans(step_trans);

		std::vector<Interval> tmpdomain;
		tmpdomain.push_back(step_end_exp_table[1]);
		for(int i=0; i<localInitial_range.size(); ++i)
		{
			tmpdomain.push_back(localInitial_range[i]);
		}

//		x.cutoff(tmpdomain, cutoff_threshold);

//		Interval_matrix rem;
//		x.Remainder(rem);

		for(int i=0; i<rangeDim; ++i)
		{
			x.tms[i].remainder += remainder_x[i][0];
		}

		// compute the the normalized TM flowpipe
		TaylorModelVec normalized_x;
//		x.insert_ctrunc_normal_no_cutoff(normalized_x, local_initial, polyRange, step_exp_table, rangeDim+1, order);
		x.insert_ctrunc_normal(normalized_x, local_initial, polyRange, step_exp_table, rangeDim+1, order, cutoff_threshold);

		Interval_matrix rem;
		x.Remainder(rem);
//		normalized_x.cutoff_normal(rem, step_exp_table, cutoff_threshold);

		// check the validity of the TM flowpipe
		bool bvalid = true;
		bool brefine = false;
		for(int i=0; i<varIDs.size(); ++i)
		{
			Interval I;
			normalized_x.tms[varIDs[i]-1].intEvalNormal(I, step_exp_table);

			if(directions[i] == 1)
			{
				double d1 = I.sup();
				double d2 = intVals[i].sup();
				if(d1 > d2)
				{
					bvalid = false;
					break;
				}
				else
				{
					if(std::abs((d2 - d1)/intVals[i].width()) >= 0.01)
						brefine = true;

					intVals[i].setSup(d1);
				}
			}
			else if(directions[i] == -1)
			{
				double d1 = I.inf();
				double d2 = intVals[i].inf();
				if(d1 < d2)
				{
					bvalid = false;
					break;
				}
				else
				{
					if(std::abs((d1 - d2)/intVals[i].width()) >= 0.01)
						brefine = true;

					intVals[i].setInf(d1);
				}
			}
			else
			{
				if( !I.subseteq(intVals[i]) )
				{
					bvalid = false;
					break;
				}
				else
				{
					if(intVals[i].widthRatio(I) <= 0.99)
						brefine = true;

					intVals[i] = I;
				}
			}
		}

		if(!bvalid)
		{
			printf("Too large stepsize.\n");
			return;
		}
		else if(brefine)
		{
			// refine the current ODE
			newOde.clear();
			new_hfOde.clear();
			goto ODE;
		}

		Interval_matrix local_remainder;
		local_remainder = remainder_x;

		remainders.push_back(local_remainder);
		flowpipes.push_back(normalized_x);

		// update the transformation queue

		for(int i=0; i<trans_on_local_initial.size(); ++i)
		{
			trans_on_local_initial[i] = stepTrans * trans_on_local_initial[i];
		}

		// compute the next local initial set

		TaylorModelVec next_local;
		normalized_x.evaluate_t(next_local, step_end_exp_table);

		Interval_matrix local_initial_rem;
		if(trans_on_local_initial.size() > 0)
		{
			local_initial_rem = trans_on_local_initial[0] * initial_rem;
		}
		else
		{
			local_initial_rem = stepTrans * initial_rem;
		}

		trans_on_local_initial.push_back(stepTrans);

		for(int i=1; i<trans_on_local_initial.size(); ++i)
		{
			local_initial_rem += trans_on_local_initial[i] * remainders[i-1];
		}

		local_initial_rem += remainders[remainders.size()-1];
		local_initial_rem += rem;

		for(int i=0; i<rangeDim; ++i)
		{
			next_local.tms[i].remainder = local_initial_rem[i][0];
		}

		local_initial = next_local;

		t += step;
		++iterNum;

		if(iterNum >= N)
		{
			// simplify the remainder
			trans_on_local_initial.clear();
			remainders.clear();

			local_initial.Remainder(initial_rem);

			iterNum = 0;
		}

		if(bPrint)
		{
			printf("time = %f,\t", t);
			printf("step = %f,\t", step);
			printf("order = %d\n", order);
		}
	}
}


ContinuousSystem & ContinuousSystem::operator = (const ContinuousSystem & system)
{
	if(this == &system)
		return *this;

	tmvOde				= system.tmvOde;
	hfOde				= system.hfOde;
	initialSet			= system.initialSet;
	strOde				= system.strOde;
	tmvOde_centered		= system.tmvOde_centered;
	hfOde_centered		= system.hfOde_centered;
	strOde_centered		= system.strOde_centered;

	return *this;
}




































// class ContinuousReachability

ContinuousReachability::ContinuousReachability()
{
}

ContinuousReachability::~ContinuousReachability()
{
	outputAxes.clear();
	flowpipes.clear();
	orders.clear();
	maxOrders.clear();
	flowpipesCompo.clear();
	domains.clear();
	unsafeSet.clear();
	stateVarTab.clear();
	stateVarNames.clear();
	tmVarTab.clear();
	tmVarNames.clear();
	parTab.clear();
	parNames.clear();
}

void ContinuousReachability::dump(FILE *fp) const
{
	fprintf(fp,"state var ");
	for(int i=0; i<stateVarNames.size()-1; ++i)
	{
		fprintf(fp, "%s,", stateVarNames[i].c_str());
	}

	fprintf(fp, "%s\n\n", stateVarNames[stateVarNames.size()-1].c_str());

	switch(plotFormat)
	{
	case PLOT_GNUPLOT:
		switch(plotSetting)
		{
		case PLOT_INTERVAL:
			fprintf(fp, "gnuplot interval %s , %s\n\n", stateVarNames[outputAxes[0]].c_str(), stateVarNames[outputAxes[1]].c_str());
			break;
		case PLOT_OCTAGON:
			fprintf(fp, "gnuplot octagon %s , %s\n\n", stateVarNames[outputAxes[0]].c_str(), stateVarNames[outputAxes[1]].c_str());
			break;
		case PLOT_GRID:
			fprintf(fp, "gnuplot grid %d %s , %s\n\n", numSections, stateVarNames[outputAxes[0]].c_str(), stateVarNames[outputAxes[1]].c_str());
			break;
		}
		break;
	case PLOT_MATLAB:
		switch(plotSetting)
		{
		case PLOT_INTERVAL:
			fprintf(fp, "matlab interval %s , %s\n\n", stateVarNames[outputAxes[0]].c_str(), stateVarNames[outputAxes[1]].c_str());
			break;
		case PLOT_OCTAGON:
			fprintf(fp, "matlab octagon %s , %s\n\n", stateVarNames[outputAxes[0]].c_str(), stateVarNames[outputAxes[1]].c_str());
			break;
		case PLOT_GRID:
			fprintf(fp, "matlab grid %d %s , %s\n\n", numSections, stateVarNames[outputAxes[0]].c_str(), stateVarNames[outputAxes[1]].c_str());
			break;
		}
		break;
	}

	fprintf(fp, "cutoff %e\n\n", cutoff_threshold.sup());

	fprintf(fp, "output %s\n\n", outputFileName);

	if(bSafetyChecking)
	{
		// dump the unsafe set
		fprintf(fp, "unsafe set\n{\n");

		for(int i=0; i<unsafeSet.size(); ++i)
		{
			unsafeSet[i].dump(fp, stateVarNames);
		}

		fprintf(fp, "}\n\n");
	}

	fprintf(fp, "continuous flowpipes\n{\n");

	fprintf(fp, "tm var ");
	for(int i=1; i<tmVarNames.size()-1; ++i)
	{
		fprintf(fp, "%s,", tmVarNames[i].c_str());
	}
	fprintf(fp, "%s\n\n", tmVarNames[tmVarNames.size()-1].c_str());

	std::list<TaylorModelVec>::const_iterator fpIter = flowpipesCompo.begin();
	std::list<std::vector<Interval> >::const_iterator doIter = domains.begin();

	// every Taylor model flowpipe is enclosed by braces

	for(; fpIter != flowpipesCompo.end(); ++fpIter, ++doIter)
	{
		fprintf(fp, "{\n");
		fpIter->dump_interval(fp, stateVarNames, tmVarNames);

		//dump the domain
		for(int i=0; i<doIter->size(); ++i)
		{
			fprintf(fp, "%s in ", tmVarNames[i].c_str());
			(*doIter)[i].dump(fp);
			fprintf(fp, "\n");
		}

		fprintf(fp, "}\n\n");
	}

	fprintf(fp, "}\n");

/*
	// ======================= test begin ========================
	fpIter = flowpipesCompo.end();
	--fpIter;
	doIter = domains.end();
	--doIter;
	TaylorModelVec tmvTemp;
	vector<Interval> step_exp_table;
	Interval intStep((*doIter)[0].sup());
	construct_step_exp_table(step_exp_table, intStep, globalMaxOrder);
	fpIter->evaluate_t(tmvTemp, step_exp_table);
	vector<Interval> enclosure;
	tmvTemp.intEval(enclosure, *doIter);

	for(int i=0; i<enclosure.size(); ++i)
	{
		enclosure[i].dump(stdout);
		printf("\n");
	}
	// ======================= test end ========================
*/
}

void ContinuousReachability::run()
{
	compute_factorial_rec(globalMaxOrder+2);
	compute_power_4(globalMaxOrder+2);
	compute_double_factorial(2*globalMaxOrder+4);

/*
	if(integrationScheme != NONPOLY_TAYLOR)
	{
		bool bLinear = true;
		int rangeDim = system.tmvOde.tms.size();

		for(int i=0; i<rangeDim; ++i)
		{
			if(system.tmvOde.tms[i].expansion.degree() > 1)
			{
				bLinear = false;
				break;
			}
		}

		if(bLinear)
		{
			integrationScheme = LINEAR;
			system.reach_linear(flowpipes, step, time, orders[0], bPrint);
			return;
		}
	}
*/

	switch(integrationScheme)
	{
	case ONLY_PICARD:
	{
		switch(orderType)
		{
		case UNIFORM:
			if(bAdaptiveSteps)
			{
				system.reach_picard(flowpipes, step, miniStep, time, orders[0], precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else if(bAdaptiveOrders)
			{
				system.reach_picard(flowpipes, step, time, orders[0], maxOrders[0], precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else
			{
				system.reach_picard(flowpipes, step, time, orders[0], precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			break;
		case MULTI:
			if(bAdaptiveSteps)
			{
				system.reach_picard(flowpipes, step, miniStep, time, orders, globalMaxOrder, precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else if(bAdaptiveOrders)
			{
				system.reach_picard(flowpipes, step, time, orders, maxOrders, globalMaxOrder, precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else
			{
				system.reach_picard(flowpipes, step, time, orders, globalMaxOrder, precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			break;
		}
		break;
	}

	case LOW_DEGREE:
	{
		switch(orderType)
		{
		case UNIFORM:
			if(bAdaptiveSteps)
			{
				system.reach_low_degree(flowpipes, step, miniStep, time, orders[0], precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else if(bAdaptiveOrders)
			{
				system.reach_low_degree(flowpipes, step, time, orders[0], maxOrders[0], precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else
			{
				system.reach_low_degree(flowpipes, step, time, orders[0], precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			break;
		case MULTI:
			if(bAdaptiveSteps)
			{
				system.reach_low_degree(flowpipes, step, miniStep, time, orders, globalMaxOrder, precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else if(bAdaptiveOrders)
			{
				system.reach_low_degree(flowpipes, step, time, orders, maxOrders, globalMaxOrder, precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else
			{
				system.reach_low_degree(flowpipes, step, time, orders, globalMaxOrder, precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			break;
		}
		break;
	}

	case HIGH_DEGREE:
	{
		switch(orderType)
		{
		case UNIFORM:
			if(bAdaptiveSteps)
			{
				system.reach_high_degree(flowpipes, step, miniStep, time, orders[0], precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else if(bAdaptiveOrders)
			{
				system.reach_high_degree(flowpipes, step, time, orders[0], maxOrders[0], precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else
			{
				system.reach_high_degree(flowpipes, step, time, orders[0], precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			break;
		case MULTI:
			if(bAdaptiveSteps)
			{
				system.reach_high_degree(flowpipes, step, miniStep, time, orders, globalMaxOrder, precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else if(bAdaptiveOrders)
			{
				system.reach_high_degree(flowpipes, step, time, orders, maxOrders, globalMaxOrder, precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else
			{
				system.reach_high_degree(flowpipes, step, time, orders, globalMaxOrder, precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			break;
		}
		break;
	}

	case NONPOLY_TAYLOR:
	{
		switch(orderType)
		{
		case UNIFORM:
			if(bAdaptiveSteps)
			{
				system.reach_non_polynomial_taylor(flowpipes, step, miniStep, time, orders[0], precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else if(bAdaptiveOrders)
			{
				system.reach_non_polynomial_taylor(flowpipes, step, time, orders[0], maxOrders[0], precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else
			{
				system.reach_non_polynomial_taylor(flowpipes, step, time, orders[0], precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			break;
		case MULTI:
			if(bAdaptiveSteps)
			{
				system.reach_non_polynomial_taylor(flowpipes, step, miniStep, time, orders, globalMaxOrder, precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else if(bAdaptiveOrders)
			{
				system.reach_non_polynomial_taylor(flowpipes, step, time, orders, maxOrders, globalMaxOrder, precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			else
			{
				system.reach_non_polynomial_taylor(flowpipes, step, time, orders, globalMaxOrder, precondition, estimation, bPrint, stateVarNames, cutoff_threshold);
			}
			break;
		}
		break;
	}

	case LINEAR:
	{
		system.reach_linear(flowpipes, step, time, orders[0], bPrint, cutoff_threshold);
		break;
	}
	}
}

void ContinuousReachability::composition()
{
	flowpipesCompo.clear();
	domains.clear();

	if(integrationScheme == LINEAR)
	{
		std::list<Flowpipe>::const_iterator iter;

		for(iter = flowpipes.begin(); iter != flowpipes.end(); ++iter)
		{
			flowpipesCompo.push_back(iter->tmvPre);
			domains.push_back(iter->domain);
		}
	}
	else
	{
		std::vector<Interval> step_exp_table;
		Interval intStep;

		std::list<Flowpipe>::const_iterator iter;

		for(iter = flowpipes.begin(); iter != flowpipes.end(); ++iter)
		{
			if(step_exp_table.size() == 0 || intStep != iter->domain[0])
			{
				construct_step_exp_table(step_exp_table, iter->domain[0], globalMaxOrder);
				intStep = iter->domain[0];
			}

			TaylorModelVec tmvTemp;

			iter->composition_normal(tmvTemp, step_exp_table, cutoff_threshold);

			flowpipesCompo.push_back(tmvTemp);
			domains.push_back(iter->domain);
		}
	}
}

int ContinuousReachability::safetyChecking() const
{
	if(unsafeSet.size() == 0)
	{
		return UNSAFE;	// since the whole state space is unsafe, the system is not safe
	}

	bool bDumpCounterexamples = true;

	int mkres = mkdir(counterexampleDir, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	if(mkres < 0 && errno != EEXIST)
	{
		printf("Can not create the directory for counterexamples.\n");
		bDumpCounterexamples = false;
	}

	char filename_counterexamples[NAME_SIZE+10];
	FILE *fpDumpCounterexamples;

	if(bDumpCounterexamples)
	{
		sprintf(filename_counterexamples, "%s%s%s", counterexampleDir, outputFileName, str_counterexample_dumping_name_suffix);
		fpDumpCounterexamples = fopen(filename_counterexamples, "w");
	}

	std::list<TaylorModelVec>::const_iterator tmvIter = flowpipesCompo.begin();
	std::list<std::vector<Interval> >::const_iterator doIter = domains.begin();

	std::vector<Interval> step_exp_table;

	int rangeDim = tmvIter->tms.size();
	int domainDim = doIter->size();

	int result = SAFE;

	std::list<TaylorModelVec> unsafe_flowpipes;
	std::list<std::vector<Interval> > unsafe_flowpipe_domains;
	std::list<Interval> globalTimes;

	int maxOrder = 0;
	Interval globalTime;

	for(; tmvIter!=flowpipesCompo.end(); ++tmvIter, ++doIter)
	{
		int tmp = maxOrder;
		for(int i=0; i<tmvIter->tms.size(); ++i)
		{
			int order = tmvIter->tms[i].expansion.degree();
			if(maxOrder < order)
			{
				maxOrder = order;
			}
		}

		if(step_exp_table.size() == 0 || step_exp_table[1] != (*doIter)[0] || maxOrder > tmp)
		{
			construct_step_exp_table(step_exp_table, (*doIter)[0], 2*maxOrder);
		}

		bool bsafe = false;

		std::vector<Interval> tmvPolyRange;
		tmvIter->polyRangeNormal(tmvPolyRange, step_exp_table);

		for(int i=0; i<unsafeSet.size(); ++i)
		{
			TaylorModel tmTemp;

			// interval evaluation on the constraint
			unsafeSet[i].hf.insert_normal(tmTemp, *tmvIter, tmvPolyRange, step_exp_table, domainDim, cutoff_threshold);

			Interval intTemp;
			tmTemp.intEvalNormal(intTemp, step_exp_table);

			if(intTemp > unsafeSet[i].B)
			{
				// no intersection with the unsafe set
				bsafe = true;
				break;
			}
			else
			{
				continue;
			}
		}

		if(!bsafe)
		{
			unsafe_flowpipes.push_back(*tmvIter);
			unsafe_flowpipe_domains.push_back(*doIter);
			globalTimes.push_back(globalTime);

			result = UNKNOWN;
		}

		globalTime += (*doIter)[0];
	}

	if(bDumpCounterexamples && unsafe_flowpipes.size() > 0)
	{
		dump_potential_counterexample(fpDumpCounterexamples, unsafe_flowpipes, unsafe_flowpipe_domains, globalTimes);
	}

	if(bDumpCounterexamples)
	{
		fclose(fpDumpCounterexamples);
	}

	return result;
}

unsigned long ContinuousReachability::numOfFlowpipes() const
{
	return (unsigned long)(flowpipes.size()) - 1;
}

void ContinuousReachability::dump_potential_counterexample(FILE *fp, const std::list<TaylorModelVec> & flowpipes, const std::list<std::vector<Interval> > & domains, const std::list<Interval> & globalTimes) const
{
	std::list<TaylorModelVec>::const_iterator fpIter = flowpipes.begin();
	std::list<std::vector<Interval> >::const_iterator doIter = domains.begin();
	std::list<Interval>::const_iterator timeIter = globalTimes.begin();

	for(; fpIter!=flowpipes.end(); ++fpIter, ++doIter, ++timeIter)
	{
		fprintf(fp, "starting time %lf\n{\n", timeIter->sup());

		fpIter->dump_interval(fp, stateVarNames, tmVarNames);

		for(int i=0; i<doIter->size(); ++i)
		{
			fprintf(fp, "%s in ", tmVarNames[i].c_str());
			(*doIter)[i].dump(fp);
			fprintf(fp, "\n");
		}

		fprintf(fp, "}\n\n\n");
	}
}

void ContinuousReachability::plot_2D() const
{
	char filename[NAME_SIZE+10];

	switch(plotFormat)
	{
	case PLOT_GNUPLOT:
		sprintf(filename, "%s%s.plt", outputDir, outputFileName);
		break;
	case PLOT_MATLAB:
		sprintf(filename, "%s%s.m", outputDir, outputFileName);
		break;
	}

	FILE *fpPlotting = fopen(filename, "w");

	if(fpPlotting == NULL)
	{
		printf("Can not create the plotting file.\n");
		exit(1);
	}

	printf("Generating the plotting file...\n");
	switch(plotFormat)
	{
	case PLOT_GNUPLOT:
		plot_2D_GNUPLOT(fpPlotting);
		break;
	case PLOT_MATLAB:
		plot_2D_MATLAB(fpPlotting);
		break;
	}
	printf("Done.\n");

	fclose(fpPlotting);
}

void ContinuousReachability::plot_2D_GNUPLOT(FILE *fp) const
{
	switch(plotSetting)
	{
	case PLOT_INTERVAL:
		plot_2D_interval_GNUPLOT(fp);
		break;
	case PLOT_OCTAGON:
		plot_2D_octagon_GNUPLOT(fp);
		break;
	case PLOT_GRID:
		plot_2D_grid_GNUPLOT(fp);
		break;
	}
}

void ContinuousReachability::plot_2D_interval_GNUPLOT(FILE *fp) const
{
	fprintf(fp, "set terminal postscript enhanced color\n");

	char filename[NAME_SIZE+10];
	sprintf(filename, "%s%s.eps", imageDir, outputFileName);
	fprintf(fp, "set output '%s'\n", filename);

	fprintf(fp, "set style line 1 linecolor rgb \"blue\"\n");
	fprintf(fp, "set autoscale\n");
	fprintf(fp, "unset label\n");
	fprintf(fp, "set xtic auto\n");
	fprintf(fp, "set ytic auto\n");
	fprintf(fp, "set xlabel \"%s\"\n", stateVarNames[outputAxes[0]].c_str());
	fprintf(fp, "set ylabel \"%s\"\n", stateVarNames[outputAxes[1]].c_str());
	fprintf(fp, "plot '-' notitle with lines ls 1\n");

	std::vector<Interval> step_exp_table, step_end_exp_table;
	Interval intStep;

	std::list<TaylorModelVec>::const_iterator tmvIter = flowpipesCompo.begin();
	std::list<std::vector<Interval> >::const_iterator doIter = domains.begin();

	std::vector<int> varIDs;
	varIDs.push_back(outputAxes[0]+1);
	varIDs.push_back(outputAxes[1]+1);

	int maxOrder = 0;

	for(; tmvIter != flowpipesCompo.end() && doIter != domains.end(); ++tmvIter, ++doIter)
	{
		int tmp = maxOrder;
		for(int i=0; i<tmvIter->tms.size(); ++i)
		{
			int order = tmvIter->tms[i].expansion.degree();
			if(maxOrder < order)
			{
				maxOrder = order;
			}
		}

		if(step_exp_table.size() == 0 || intStep != (*doIter)[0] || maxOrder > tmp)
		{
			construct_step_exp_table(step_exp_table, step_end_exp_table, (*doIter)[0].sup(), maxOrder);
			intStep = (*doIter)[0];
		}

		std::vector<Interval> box;
		tmvIter->intEvalNormal(box, step_exp_table, varIDs);

		Interval X(box[0]), Y(box[1]);

		// output the vertices
		fprintf(fp, "%lf %lf\n", X.inf(), Y.inf());
		fprintf(fp, "%lf %lf\n", X.sup(), Y.inf());
		fprintf(fp, "%lf %lf\n", X.sup(), Y.sup());
		fprintf(fp, "%lf %lf\n", X.inf(), Y.sup());
		fprintf(fp, "%lf %lf\n", X.inf(), Y.inf());
		fprintf(fp, "\n\n");
	}

	fprintf(fp, "e\n");
}

void ContinuousReachability::plot_2D_octagon_GNUPLOT(FILE *fp) const
{
	int x = outputAxes[0];
	int y = outputAxes[1];

	int rangeDim = stateVarNames.size();
	Matrix output_poly_temp(8, rangeDim);

	output_poly_temp.set(1, 0, x);
	output_poly_temp.set(1, 1, y);
	output_poly_temp.set(-1, 2, x);
	output_poly_temp.set(-1, 3, y);
	output_poly_temp.set(1/sqrt(2), 4, x);
	output_poly_temp.set(1/sqrt(2), 4, y);
	output_poly_temp.set(1/sqrt(2), 5, x);
	output_poly_temp.set(-1/sqrt(2), 5, y);
	output_poly_temp.set(-1/sqrt(2), 6, x);
	output_poly_temp.set(1/sqrt(2), 6, y);
	output_poly_temp.set(-1/sqrt(2), 7, x);
	output_poly_temp.set(-1/sqrt(2), 7, y);

	// Construct the 2D template matrix.
	int rows = 8;
	int cols = rangeDim;

	Matrix sortedTemplate(rows, cols);
	RowVector rowVec(cols);
	std::vector<RowVector> sortedRows;
	std::vector<RowVector>::iterator iterp, iterq;

	output_poly_temp.row(rowVec, 0);
	sortedRows.push_back(rowVec);

	bool bInserted;

	// Sort the row vectors in the template by anti-clockwise order (only in the x-y space).
	for(int i=1; i<rows; ++i)
	{
		iterp = sortedRows.begin();
		iterq = iterp;
		++iterq;
		bInserted = false;

		for(; iterq != sortedRows.end();)
		{
			double tmp1 = output_poly_temp.get(i,x) * iterp->get(y) - output_poly_temp.get(i,y) * iterp->get(x);
			double tmp2 = output_poly_temp.get(i,x) * iterq->get(y) - output_poly_temp.get(i,y) * iterq->get(x);

			if(tmp1 < 0 && tmp2 > 0)
			{
				output_poly_temp.row(rowVec, i);
				sortedRows.insert(iterq, rowVec);
				bInserted = true;
				break;
			}
			else
			{
				++iterp;
				++iterq;
			}
		}

		if(!bInserted)
		{
			output_poly_temp.row(rowVec, i);
			sortedRows.push_back(rowVec);
		}
	}

	iterp = sortedRows.begin();
	for(int i=0; i<rows; ++i, ++iterp)
	{
		for(int j=0; j<cols; ++j)
		{
			sortedTemplate.set(iterp->get(j), i, j);
		}
	}

	ColVector b(rows);
	Polyhedron polyTemplate(sortedTemplate, b);

	fprintf(fp, "set terminal postscript enhanced color\n");

	char filename[NAME_SIZE+10];
	sprintf(filename, "%s%s.eps", imageDir, outputFileName);
	fprintf(fp, "set output '%s'\n", filename);

	fprintf(fp, "set style line 1 linecolor rgb \"blue\"\n");
	fprintf(fp, "set autoscale\n");
	fprintf(fp, "unset label\n");
	fprintf(fp, "set xtic auto\n");
	fprintf(fp, "set ytic auto\n");
	fprintf(fp, "set xlabel \"%s\"\n", stateVarNames[outputAxes[0]].c_str());
	fprintf(fp, "set ylabel \"%s\"\n", stateVarNames[outputAxes[1]].c_str());
	fprintf(fp, "plot '-' notitle with lines ls 1\n");

	// Compute the intersections of two facets.
	// The vertices are ordered clockwisely.

	gsl_matrix *C = gsl_matrix_alloc(2,2);
	gsl_vector *d = gsl_vector_alloc(2);
	gsl_vector *vertex = gsl_vector_alloc(2);

	std::vector<Interval> step_exp_table, step_end_exp_table;
	Interval intStep;

	std::list<TaylorModelVec>::const_iterator tmvIter = flowpipesCompo.begin();
	std::list<std::vector<Interval> >::const_iterator doIter = domains.begin();

	int maxOrder = 0;

	for(; tmvIter != flowpipesCompo.end() && doIter != domains.end(); ++tmvIter, ++doIter)
	{
		int tmp = maxOrder;
		for(int i=0; i<tmvIter->tms.size(); ++i)
		{
			int order = tmvIter->tms[i].expansion.degree();
			if(maxOrder < order)
			{
				maxOrder = order;
			}
		}

		if(step_exp_table.size() == 0 || intStep != (*doIter)[0] || maxOrder > tmp)
		{
			construct_step_exp_table(step_exp_table, step_end_exp_table, (*doIter)[0].sup(), maxOrder);
			intStep = (*doIter)[0];
		}


		templatePolyhedronNormal(polyTemplate, *tmvIter, step_exp_table);

		double f1, f2;

		std::vector<LinearConstraint>::iterator iterp, iterq;
		iterp = iterq = polyTemplate.constraints.begin();
		++iterq;

		for(; iterq != polyTemplate.constraints.end(); ++iterp, ++iterq)
		{
			gsl_matrix_set(C, 0, 0, iterp->A[x].midpoint());
			gsl_matrix_set(C, 0, 1, iterp->A[y].midpoint());
			gsl_matrix_set(C, 1, 0, iterq->A[x].midpoint());
			gsl_matrix_set(C, 1, 1, iterq->A[y].midpoint());

			gsl_vector_set(d, 0, iterp->B.midpoint());
			gsl_vector_set(d, 1, iterq->B.midpoint());

			gsl_linalg_HH_solve(C, d, vertex);

			double v1 = gsl_vector_get(vertex, 0);
			double v2 = gsl_vector_get(vertex, 1);

			if(iterp == polyTemplate.constraints.begin())
			{
				f1 = v1;
				f2 = v2;
			}

			fprintf(fp, "%lf %lf\n", v1, v2);
		}

		iterp = polyTemplate.constraints.begin();
		--iterq;

		gsl_matrix_set(C, 0, 0, iterp->A[x].midpoint());
		gsl_matrix_set(C, 0, 1, iterp->A[y].midpoint());
		gsl_matrix_set(C, 1, 0, iterq->A[x].midpoint());
		gsl_matrix_set(C, 1, 1, iterq->A[y].midpoint());

		gsl_vector_set(d, 0, iterp->B.midpoint());
		gsl_vector_set(d, 1, iterq->B.midpoint());

		gsl_linalg_HH_solve(C, d, vertex);

		double v1 = gsl_vector_get(vertex, 0);
		double v2 = gsl_vector_get(vertex, 1);

		fprintf(fp, "%lf %lf\n", v1, v2);

		fprintf(fp, "%lf %lf\n", f1, f2);
		fprintf(fp, "\n\n");
	}

	fprintf(fp, "e\n");

	gsl_matrix_free(C);
	gsl_vector_free(d);
	gsl_vector_free(vertex);
}

void ContinuousReachability::plot_2D_grid_GNUPLOT(FILE *fp) const
{
	fprintf(fp, "set terminal postscript enhanced color\n");

	char filename[NAME_SIZE+10];
	sprintf(filename, "%s%s.eps", imageDir, outputFileName);
	fprintf(fp, "set output '%s'\n", filename);

	fprintf(fp, "set style line 1 linecolor rgb \"blue\"\n");
	fprintf(fp, "set autoscale\n");
	fprintf(fp, "unset label\n");
	fprintf(fp, "set xtic auto\n");
	fprintf(fp, "set ytic auto\n");
	fprintf(fp, "set xlabel \"%s\"\n", stateVarNames[outputAxes[0]].c_str());
	fprintf(fp, "set ylabel \"%s\"\n", stateVarNames[outputAxes[1]].c_str());
	fprintf(fp, "plot '-' notitle with lines ls 1\n");

	std::list<TaylorModelVec>::const_iterator tmvIter = flowpipesCompo.begin();
	std::list<std::vector<Interval> >::const_iterator doIter = domains.begin();
//	int domainDim = doIter->size();

	for(; tmvIter != flowpipesCompo.end() && doIter != domains.end(); ++tmvIter, ++doIter)
	{
		// decompose the domain
		std::list<std::vector<Interval> > grids;

		gridBox(grids, *doIter, numSections);

		// we only consider the output dimensions
		HornerForm hfOutputX;
		Interval remainderX;
		tmvIter->tms[outputAxes[0]].toHornerForm(hfOutputX, remainderX);

		HornerForm hfOutputY;
		Interval remainderY;
		tmvIter->tms[outputAxes[1]].toHornerForm(hfOutputY, remainderY);

		// evaluate the images from all of the grids
		std::list<std::vector<Interval> >::const_iterator gIter = grids.begin();
		for(; gIter!=grids.end(); ++gIter)
		{
			Interval X;
			hfOutputX.intEval(X, *gIter);
			X += remainderX;

			Interval Y;
			hfOutputY.intEval(Y, *gIter);
			Y += remainderY;

			// output the vertices
			fprintf(fp, "%lf %lf\n", X.inf(), Y.inf());
			fprintf(fp, "%lf %lf\n", X.sup(), Y.inf());
			fprintf(fp, "%lf %lf\n", X.sup(), Y.sup());
			fprintf(fp, "%lf %lf\n", X.inf(), Y.sup());
			fprintf(fp, "%lf %lf\n", X.inf(), Y.inf());
			fprintf(fp, "\n\n");
		}
	}

	fprintf(fp, "e\n");
}

void ContinuousReachability::plot_2D_MATLAB(FILE *fp) const
{
	switch(plotSetting)
	{
	case PLOT_INTERVAL:
		plot_2D_interval_MATLAB(fp);
		break;
	case PLOT_OCTAGON:
		plot_2D_octagon_MATLAB(fp);
		break;
	case PLOT_GRID:
		plot_2D_grid_MATLAB(fp);
		break;
	}
}

void ContinuousReachability::plot_2D_interval_MATLAB(FILE *fp) const
{
	std::vector<Interval> step_exp_table, step_end_exp_table;
	Interval intStep;

	std::list<TaylorModelVec>::const_iterator tmvIter = flowpipesCompo.begin();
	std::list<std::vector<Interval> >::const_iterator doIter = domains.begin();

	std::vector<int> varIDs;
	varIDs.push_back(outputAxes[0]+1);
	varIDs.push_back(outputAxes[1]+1);

	int maxOrder = 0;

	for(; tmvIter != flowpipesCompo.end() && doIter != domains.end(); ++tmvIter, ++doIter)
	{
		int tmp = maxOrder;
		for(int i=0; i<tmvIter->tms.size(); ++i)
		{
			int order = tmvIter->tms[i].expansion.degree();
			if(maxOrder < order)
			{
				maxOrder = order;
			}
		}

		if(step_exp_table.size() == 0 || intStep != (*doIter)[0] || maxOrder > tmp)
		{
			construct_step_exp_table(step_exp_table, step_end_exp_table, (*doIter)[0].sup(), maxOrder);
			intStep = (*doIter)[0];
		}

		std::vector<Interval> box;
		tmvIter->intEvalNormal(box, step_exp_table, varIDs);

		Interval X(box[0]), Y(box[1]);

		// output all the vertices
		fprintf(fp,"fill( [%e,%e,%e,%e,%e] , [%e,%e,%e,%e,%e] , [1,0.6,0.6], 'EdgeColor', [1,0.6,0.6]);\nhold on;\nclear;\n",
				X.inf(), X.sup(), X.sup(), X.inf(), X.inf(), Y.inf(), Y.inf(), Y.sup(), Y.sup(), Y.inf());
	}
}

void ContinuousReachability::plot_2D_octagon_MATLAB(FILE *fp) const
{
	int x = outputAxes[0];
	int y = outputAxes[1];

	int rangeDim = stateVarNames.size();
	Matrix output_poly_temp(8, rangeDim);

	output_poly_temp.set(1, 0, x);
	output_poly_temp.set(1, 1, y);
	output_poly_temp.set(-1, 2, x);
	output_poly_temp.set(-1, 3, y);
	output_poly_temp.set(1/sqrt(2), 4, x);
	output_poly_temp.set(1/sqrt(2), 4, y);
	output_poly_temp.set(1/sqrt(2), 5, x);
	output_poly_temp.set(-1/sqrt(2), 5, y);
	output_poly_temp.set(-1/sqrt(2), 6, x);
	output_poly_temp.set(1/sqrt(2), 6, y);
	output_poly_temp.set(-1/sqrt(2), 7, x);
	output_poly_temp.set(-1/sqrt(2), 7, y);

	// Construct the 2D template matrix.
	int rows = 8;
	int cols = rangeDim;

	Matrix sortedTemplate(rows, cols);
	RowVector rowVec(cols);
	std::vector<RowVector> sortedRows;
	std::vector<RowVector>::iterator iterp, iterq;

	output_poly_temp.row(rowVec, 0);
	sortedRows.push_back(rowVec);

	bool bInserted;

	// Sort the row vectors in the template by anti-clockwise order (only in the x-y space).
	for(int i=1; i<rows; ++i)
	{
		iterp = sortedRows.begin();
		iterq = iterp;
		++iterq;
		bInserted = false;

		for(; iterq != sortedRows.end();)
		{
			double tmp1 = output_poly_temp.get(i,x) * iterp->get(y) - output_poly_temp.get(i,y) * iterp->get(x);
			double tmp2 = output_poly_temp.get(i,x) * iterq->get(y) - output_poly_temp.get(i,y) * iterq->get(x);

			if(tmp1 < 0 && tmp2 > 0)
			{
				output_poly_temp.row(rowVec, i);
				sortedRows.insert(iterq, rowVec);
				bInserted = true;
				break;
			}
			else
			{
				++iterp;
				++iterq;
			}
		}

		if(!bInserted)
		{
			output_poly_temp.row(rowVec, i);
			sortedRows.push_back(rowVec);
		}
	}

	iterp = sortedRows.begin();
	for(int i=0; i<rows; ++i, ++iterp)
	{
		for(int j=0; j<cols; ++j)
		{
			sortedTemplate.set(iterp->get(j), i, j);
		}
	}

	ColVector b(rows);
	Polyhedron polyTemplate(sortedTemplate, b);

	// Compute the intersections of two facets.
	// The vertices are ordered clockwisely.

	gsl_matrix *C = gsl_matrix_alloc(2,2);
	gsl_vector *d = gsl_vector_alloc(2);
	gsl_vector *vertex = gsl_vector_alloc(2);

	std::vector<Interval> step_exp_table, step_end_exp_table;
	Interval intStep;

	std::list<TaylorModelVec>::const_iterator tmvIter = flowpipesCompo.begin();
	std::list<std::vector<Interval> >::const_iterator doIter = domains.begin();

	int maxOrder = 0;

	for(; tmvIter != flowpipesCompo.end() && doIter != domains.end(); ++tmvIter, ++doIter)
	{
		int tmp = maxOrder;
		for(int i=0; i<tmvIter->tms.size(); ++i)
		{
			int order = tmvIter->tms[i].expansion.degree();
			if(maxOrder < order)
			{
				maxOrder = order;
			}
		}

		if(step_exp_table.size() == 0 || intStep != (*doIter)[0] || maxOrder > tmp)
		{
			construct_step_exp_table(step_exp_table, step_end_exp_table, (*doIter)[0].sup(), maxOrder);
			intStep = (*doIter)[0];
		}


		templatePolyhedronNormal(polyTemplate, *tmvIter, step_exp_table);

		double f1, f2;

		std::vector<LinearConstraint>::iterator iterp, iterq;
		iterp = iterq = polyTemplate.constraints.begin();
		++iterq;

		std::vector<double> vertices_x, vertices_y;

		for(; iterq != polyTemplate.constraints.end(); ++iterp, ++iterq)
		{
			gsl_matrix_set(C, 0, 0, iterp->A[x].midpoint());
			gsl_matrix_set(C, 0, 1, iterp->A[y].midpoint());
			gsl_matrix_set(C, 1, 0, iterq->A[x].midpoint());
			gsl_matrix_set(C, 1, 1, iterq->A[y].midpoint());

			gsl_vector_set(d, 0, iterp->B.midpoint());
			gsl_vector_set(d, 1, iterq->B.midpoint());

			gsl_linalg_HH_solve(C, d, vertex);

			double v1 = gsl_vector_get(vertex, 0);
			double v2 = gsl_vector_get(vertex, 1);

			if(iterp == polyTemplate.constraints.begin())
			{
				f1 = v1;
				f2 = v2;
			}

			vertices_x.push_back(v1);
			vertices_y.push_back(v2);
		}

		iterp = polyTemplate.constraints.begin();
		--iterq;

		gsl_matrix_set(C, 0, 0, iterp->A[x].midpoint());
		gsl_matrix_set(C, 0, 1, iterp->A[y].midpoint());
		gsl_matrix_set(C, 1, 0, iterq->A[x].midpoint());
		gsl_matrix_set(C, 1, 1, iterq->A[y].midpoint());

		gsl_vector_set(d, 0, iterp->B.midpoint());
		gsl_vector_set(d, 1, iterq->B.midpoint());

		gsl_linalg_HH_solve(C, d, vertex);

		double v1 = gsl_vector_get(vertex, 0);
		double v2 = gsl_vector_get(vertex, 1);

		vertices_x.push_back(v1);
		vertices_y.push_back(v2);
		vertices_x.push_back(f1);
		vertices_y.push_back(f2);

		fprintf(fp, "plot( ");

		fprintf(fp, "[ ");
		for(int i=0; i<vertices_x.size()-1; ++i)
		{
			fprintf(fp, "%e , ", vertices_x[i]);
		}
		fprintf(fp, "%e ] , ", vertices_x.back());

		fprintf(fp, "[ ");
		for(int i=0; i<vertices_y.size()-1; ++i)
		{
			fprintf(fp, "%e , ", vertices_y[i]);
		}
		fprintf(fp, "%e ] , ", vertices_y.back());

		fprintf(fp, "'b');\nhold on;\nclear;\n");
	}

	gsl_matrix_free(C);
	gsl_vector_free(d);
	gsl_vector_free(vertex);
}

void ContinuousReachability::plot_2D_grid_MATLAB(FILE *fp) const
{
	std::list<TaylorModelVec>::const_iterator tmvIter = flowpipesCompo.begin();
	std::list<std::vector<Interval> >::const_iterator doIter = domains.begin();
//	int domainDim = doIter->size();

	for(; tmvIter != flowpipesCompo.end() && doIter != domains.end(); ++tmvIter, ++doIter)
	{
		// decompose the domain
		std::list<std::vector<Interval> > grids;

		gridBox(grids, *doIter, numSections);

		// we only consider the output dimensions
		HornerForm hfOutputX;
		Interval remainderX;
		tmvIter->tms[outputAxes[0]].toHornerForm(hfOutputX, remainderX);

		HornerForm hfOutputY;
		Interval remainderY;
		tmvIter->tms[outputAxes[1]].toHornerForm(hfOutputY, remainderY);

		// evaluate the images from all of the grids
		std::list<std::vector<Interval> >::const_iterator gIter = grids.begin();
		for(; gIter!=grids.end(); ++gIter)
		{
			Interval X;
			hfOutputX.intEval(X, *gIter);
			X += remainderX;

			Interval Y;
			hfOutputY.intEval(Y, *gIter);
			Y += remainderY;

			// output the vertices
			fprintf(fp,"plot( [%e,%e,%e,%e,%e] , [%e,%e,%e,%e,%e] , 'b');\nhold on;\nclear;\n",
					X.inf(), X.sup(), X.sup(), X.inf(), X.inf(), Y.inf(), Y.inf(), Y.sup(), Y.sup(), Y.inf());
		}
	}
}

bool ContinuousReachability::declareStateVar(const std::string & vName)
{
	std::map<std::string,int>::const_iterator iter;

	if((iter = stateVarTab.find(vName)) == stateVarTab.end())
	{
		stateVarTab[vName] = stateVarNames.size();
		stateVarNames.push_back(vName);
		return true;
	}
	else
	{
		return false;
	}
}

int ContinuousReachability::getIDForStateVar(const std::string & vName) const
{
	std::map<std::string,int>::const_iterator iter;
	if((iter = stateVarTab.find(vName)) == stateVarTab.end())
	{
		return -1;
	}

	return iter->second;
}

bool ContinuousReachability::getStateVarName(std::string & vName, int id) const
{
	if(id >= 0 && id < stateVarNames.size())
	{
		vName = stateVarNames[id];
		return true;
	}
	else
	{
		return false;
	}
}



bool ContinuousReachability::declareTMVar(const std::string & vName)
{
	std::map<std::string,int>::const_iterator iter;

	if((iter = tmVarTab.find(vName)) == tmVarTab.end())
	{
		tmVarTab[vName] = tmVarNames.size();
		tmVarNames.push_back(vName);
		return true;
	}
	else
	{
		return false;
	}
}

int ContinuousReachability::getIDForTMVar(const std::string & vName) const
{
	std::map<std::string,int>::const_iterator iter;
	if((iter = tmVarTab.find(vName)) == tmVarTab.end())
	{
		return -1;
	}

	return iter->second;
}

bool ContinuousReachability::getTMVarName(std::string & vName, int id) const
{
	if(id >= 0 && id < tmVarNames.size())
	{
		vName = tmVarNames[id];
		return true;
	}
	else
	{
		return false;
	}
}




bool ContinuousReachability::declarePar(const std::string & pName, const Interval & range)
{
	std::map<std::string,int>::const_iterator iter;

	if((iter = parTab.find(pName)) == parTab.end())
	{
		parTab[pName] = parNames.size();
		parNames.push_back(pName);
		parRanges.push_back(range);
		return true;
	}
	else
	{
		return false;
	}
}

int ContinuousReachability::getIDForPar(const std::string & pName) const
{
	std::map<std::string,int>::const_iterator iter;
	if((iter = parTab.find(pName)) == parTab.end())
	{
		return -1;
	}

	return iter->second;
}

bool ContinuousReachability::getParName(std::string & pName, int id) const
{
	if(id >= 0 && id < parNames.size())
	{
		pName = parNames[id];
		return true;
	}
	else
	{
		return false;
	}
}

bool ContinuousReachability::getRangeForPar(Interval & range, const std::string & pName) const
{
	int id = getIDForPar(pName);

	if(id == -1)
	{
		return false;
	}
	else
	{
		range = parRanges[id];
		return true;
	}
}











namespace flowstar
{

void computeTaylorExpansion(TaylorModelVec & result, const TaylorModelVec & first_order_deriv, const TaylorModelVec & ode, const int order, const Interval & cutoff_threshold)
{
	std::vector<Interval> intVecZero;
	Interval intZero, intOne(1,1);

	intVecZero.push_back(intOne);
	intVecZero.push_back(intZero);

	// we compute the Taylor expansion (without the 0-order term)
	TaylorModelVec taylorExpansion;
	first_order_deriv.evaluate_t(taylorExpansion, intVecZero);
//	taylorExpansion.nctrunc(order - 1);
	taylorExpansion.mul_assign(0, 1);

	TaylorModelVec tmvLieDeriv_n = first_order_deriv;

	for(int i=2; i<=order; ++i)
	{
		TaylorModelVec tmvTemp;
		tmvLieDeriv_n.LieDerivative_no_remainder(tmvTemp, ode, order - i, cutoff_threshold);

		TaylorModelVec tmvTemp2;
		tmvTemp.evaluate_t(tmvTemp2, intVecZero);
		tmvTemp2.mul_assign(factorial_rec[i]);
		tmvTemp2.mul_assign(0,i);			// multiplied by t^i
//		tmvTemp2.nctrunc(order);

		taylorExpansion.add_assign(tmvTemp2);

		tmvLieDeriv_n = tmvTemp;
	}
/*
	for(int i=0; i<taylorExpansion.tms.size(); ++i)
	{
		taylorExpansion.tms[i].cutoff(cutoff_threshold);
	}
*/
	result = taylorExpansion;
}

void computeTaylorExpansion(TaylorModelVec & result, const TaylorModelVec & first_order_deriv, const TaylorModelVec & ode, const std::vector<int> & orders, const Interval & cutoff_threshold)
{
	int rangeDim = ode.tms.size();
	std::vector<Interval> intVecZero;
	Interval intZero, intOne(1,1);

	intVecZero.push_back(intOne);
	intVecZero.push_back(intZero);

	// we compute the Taylor expansion (without the 0-order term)
	TaylorModelVec taylorExpansion;
	first_order_deriv.evaluate_t(taylorExpansion, intVecZero);
/*
	for(int i=0; i<rangeDim; ++i)
	{
		taylorExpansion.tms[i].nctrunc(orders[i] - 1);
	}
*/
	taylorExpansion.mul_assign(0, 1);

	TaylorModelVec tmvLieDeriv_n = first_order_deriv;

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=2; j<=orders[i]; ++j)
		{
			TaylorModel tmTemp;
			tmvLieDeriv_n.tms[i].LieDerivative_no_remainder(tmTemp, ode, orders[i] - j, cutoff_threshold);

			TaylorModel tmTemp2;
			tmTemp.evaluate_t(tmTemp2, intVecZero);
			tmTemp2.mul_assign(factorial_rec[j]);
			tmTemp2.mul_assign(0,j);
//			tmTemp2.nctrunc(orders[i]);

			taylorExpansion.tms[i].add_assign(tmTemp2);

			tmvLieDeriv_n.tms[i] = tmTemp;
		}
	}
/*
	for(int i=0; i<taylorExpansion.tms.size(); ++i)
	{
		taylorExpansion.tms[i].cutoff(cutoff_threshold);
	}
*/
	result = taylorExpansion;
}

void construct_step_exp_table(std::vector<Interval> & step_exp_table, std::vector<Interval> & step_end_exp_table, const double step, const int order)
{
	step_exp_table.clear();
	step_end_exp_table.clear();

	Interval intProd(1), intStep(0,step);

	for(int i=0; i<=order; ++i)
	{
		step_exp_table.push_back(intProd);

		Interval intTend(intProd.sup());
		step_end_exp_table.push_back(intTend);

		intProd *= intStep;
	}
}

void construct_step_exp_table(std::vector<Interval> & step_exp_table, const Interval & step, const int order)
{
	step_exp_table.clear();

	Interval intProd(1);
	step_exp_table.push_back(intProd);

	for(int i=1; i<=order; ++i)
	{
		intProd *= step;
		step_exp_table.push_back(intProd);
	}
}

void preconditionQR(Matrix & result, const TaylorModelVec & x0, const int rangeDim, const int domainDim)
{
	Interval intZero;
	std::vector<std::vector<Interval> > intCoefficients;

	std::vector<Interval> intVecTemp;
	for(int i=0; i<domainDim; ++i)
	{
		intVecTemp.push_back(intZero);
	}

	for(int i=0; i<rangeDim; ++i)
	{
		intCoefficients.push_back(intVecTemp);
	}

	x0.linearCoefficients(intCoefficients);
	Matrix matCoefficients(rangeDim, rangeDim);

	for(int i=0; i<rangeDim; ++i)
	{
		for(int j=1; j<=rangeDim; ++j)
		{
			matCoefficients.set(intCoefficients[i][j].midpoint(), i, j-1);
		}
	}

	matCoefficients.sortColumns();
	matCoefficients.QRfactor(result);
}

Interval rho(const TaylorModelVec & tmv, const std::vector<Interval> & l, const std::vector<Interval> & domain)
{
	int d = l.size();
	TaylorModel tmObj;

	for(int i=0; i<d; ++i)
	{
		TaylorModel tmTemp;
		tmv.tms[i].mul(tmTemp, l[i]);
		tmObj.add_assign(tmTemp);
	}

	Interval intRange;
	tmObj.intEval(intRange, domain);

	Interval S;
	intRange.sup(S);

	return S;
}

Interval rhoNormal(const TaylorModelVec & tmv, const std::vector<Interval> & l, const std::vector<Interval> & step_exp_table)
{
	int d = l.size();
	TaylorModel tmObj;

	for(int i=0; i<d; ++i)
	{
		TaylorModel tmTemp;
		tmv.tms[i].mul(tmTemp, l[i]);
		tmObj.add_assign(tmTemp);
	}

	Interval intRange;
	tmObj.intEvalNormal(intRange, step_exp_table);

	Interval S;
	intRange.sup(S);

	return S;
}

Interval rho(const TaylorModelVec & tmv, const RowVector & l, const std::vector<Interval> & domain)
{
	int d = l.size();
	TaylorModel tmObj;

	for(int i=0; i<d; ++i)
	{
		TaylorModel tmTemp;
		Interval intTemp(l.get(i));
		tmv.tms[i].mul(tmTemp, intTemp);
		tmObj.add_assign(tmTemp);
	}

	Interval intRange;
	tmObj.intEval(intRange, domain);

	Interval S;
	intRange.sup(S);

	return S;
}

Interval rhoNormal(const TaylorModelVec & tmv, const RowVector & l, const std::vector<Interval> & step_exp_table)
{
	int d = l.size();
	TaylorModel tmObj;

	for(int i=0; i<d; ++i)
	{
		TaylorModel tmTemp;
		Interval intTemp(l.get(i));
		tmv.tms[i].mul(tmTemp, intTemp);
		tmObj.add_assign(tmTemp);
	}

	Interval intRange;
	tmObj.intEvalNormal(intRange, step_exp_table);

	Interval S;
	intRange.sup(S);

	return S;
}

void templatePolyhedron(Polyhedron & result, const TaylorModelVec & tmv, const std::vector<Interval> & domain)
{
	for(int i=0; i<result.constraints.size(); ++i)
	{
		result.constraints[i].B = rho(tmv, result.constraints[i].A, domain);
	}
}

void templatePolyhedronNormal(Polyhedron & result, const TaylorModelVec & tmv, std::vector<Interval> & step_exp_table)
{
	for(int i=0; i<result.constraints.size(); ++i)
	{
		result.constraints[i].B = rhoNormal(tmv, result.constraints[i].A, step_exp_table);
	}
}

/*
int intersection_check_interval_arithmetic(const vector<PolynomialConstraint> & pcs, const vector<HornerForm> & objFuncs, const vector<Interval> & remainders, const vector<Interval> & domain, vector<bool> & bNeeded)
{
	int counter = 0;
	bNeeded.clear();

	for(int i=0; i<pcs.size(); ++i)
	{
		Interval intTemp;
		objFuncs[i].intEval(intTemp, domain);
		intTemp += remainders[i];

		if(intTemp > pcs[i].B)
		{
			// no intersection
			return -1;
		}
		else if(intTemp.smallereq(pcs[i].B))
		{
			// the flowpipe is entirely contained in the guard, domain contraction is not needed
			bNeeded.push_back(false);
			++counter;
		}
		else
		{
			bNeeded.push_back(true);
		}
	}

	return counter;
}

bool boundary_intersected_collection(const vector<PolynomialConstraint> & pcs, const vector<HornerForm> & objFuncs, const vector<Interval> & remainders, const vector<Interval> & domain, vector<bool> & boundary_intersected)
{
	boundary_intersected.clear();

	for(int i=0; i<pcs.size(); ++i)
	{
		Interval intTemp;
		objFuncs[i].intEval(intTemp, domain);
		intTemp += remainders[i];

		if(intTemp > pcs[i].B)
		{
			// no intersection
			return false;
		}
		else if(intTemp < pcs[i].B)
		{
			// do not intersect the boundary
			boundary_intersected.push_back(false);
		}
		else
		{
			boundary_intersected.push_back(true);
		}
	}

	return true;
}
*/

int contract_interval_arithmetic(TaylorModelVec & flowpipe, std::vector<Interval> & domain, const std::vector<PolynomialConstraint> & pcs, std::vector<bool> & boundary_intersected, const Interval & cutoff_threshold)
{
	int rangeDim = flowpipe.tms.size();
	int domainDim = domain.size();

	// contract the remainder firstly
	bool bvalid = true;
	bool bcontinue = true;
	Interval W;
	Interval intZero;

	std::vector<bool> bNeeded;
	int counter = 0;

	for(int i=0; i<pcs.size(); ++i)
	{
		bNeeded.push_back(true);
	}

	boundary_intersected.clear();

	std::vector<Interval> remainders;
	for(int i=0; i<rangeDim; ++i)
	{
		remainders.push_back(flowpipe.tms[i].remainder);
	}

	std::vector<Interval> flowpipePolyRange;
	flowpipe.polyRange(flowpipePolyRange, domain);

	std::vector<Interval> intVecTemp = flowpipePolyRange;
	intVecTemp.insert(intVecTemp.begin(), intZero);

	// 1: we check the intersection with every constraint
	for(int i=0; i<rangeDim; ++i)
	{
		intVecTemp[i+1] = flowpipePolyRange[i] + remainders[i];
	}

	for(int i=0; i<pcs.size(); ++i)
	{
		Interval intTemp;

		pcs[i].hf.intEval(intTemp, intVecTemp);

		if(intTemp > pcs[i].B)
		{
			// no intersection on the left half
			bvalid = false;
			break;
		}
		else if(intTemp.smallereq(pcs[i].B))
		{
			// do not need to apply domain contraction w.r.t. the current constraint
			boundary_intersected.push_back(false);
			bNeeded[i] = false;
			++counter;
		}
		else
		{
			boundary_intersected.push_back(true);
			bNeeded[i] = true;
			continue;
		}
	}

	if(!bvalid)
	{
		boundary_intersected.clear();
		return -1;	// no intersection is detected
	}
	else if(counter == pcs.size())
	{
		return 0;	// no need to do contraction
	}


	// 2: remainder contraction
	for(; bcontinue; )
	{
		std::vector<Interval> oldRemainders = remainders;

		for(int i=0; i<rangeDim; ++i)
		{
			Interval newInt = remainders[i];
			std::vector<bool> localNeeded = bNeeded;
			int localCounter = counter;

			for(int k=0; k<rangeDim; ++k)
			{
				if(k != i)
				{
					intVecTemp[k+1] = flowpipePolyRange[k] + remainders[k];
				}
				else
				{
					intVecTemp[k+1] = flowpipePolyRange[k];
				}
			}

			newInt.width(W);

			// search an approximation for the lower bound
			for(; W >= DC_THRESHOLD_SEARCH;)
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);

				for(int j=0; j<pcs.size(); ++j)
				{
					if(localNeeded[j])
					{
						Interval intTemp;
						std::vector<Interval> newIntVecTemp = intVecTemp;
						newIntVecTemp[i+1] += intLeft;

						pcs[j].hf.intEval(intTemp, newIntVecTemp);

						if(intTemp > pcs[j].B)
						{
							// no intersection on the left half
							newInt = intRight;
							newInt.width(W);
							break;
						}
						else if(intTemp.smallereq(pcs[j].B))
						{
							// do not need to apply domain contraction w.r.t. the current constraint
							newInt = intLeft;
							newInt.width(W);
							localNeeded[j] = false;
							++localCounter;
						}
						else
						{
							// refine the interval
							newInt = intLeft;
							newInt.width(W);

							continue;
						}
					}
				}

				if(localCounter == pcs.size())
				{
					break;
				}
			}

			// set the lower bound
			Interval Inf;
			newInt.inf(Inf);
			remainders[i].setInf(Inf);

			newInt = remainders[i];
			newInt.width(W);

			localNeeded = bNeeded;
			localCounter = counter;

			// search an approximation for the upper bound
			for(; W >= DC_THRESHOLD_SEARCH;)
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);

				for(int j=0; j<pcs.size(); ++j)
				{
					if(localNeeded[j])
					{
						Interval intTemp;
						std::vector<Interval> newIntVecTemp = intVecTemp;
						newIntVecTemp[i+1] += intRight;

						pcs[j].hf.intEval(intTemp, newIntVecTemp);

						if(intTemp > pcs[j].B)
						{
							// no intersection on the right half
							newInt = intLeft;
							newInt.width(W);
							break;
						}
						else if(intTemp.smallereq(pcs[j].B))
						{
							// do not need to apply domain contraction w.r.t. the current constraint
							newInt = intRight;
							newInt.width(W);
							localNeeded[j] = false;
							++localCounter;
						}
						else
						{
							// refine the interval
							newInt = intRight;
							newInt.width(W);
							continue;
						}
					}
				}

				if(localCounter == pcs.size())
				{
					break;
				}
			}

			Interval Sup;
			newInt.sup(Sup);
			remainders[i].setSup(Sup);	// set the upper bound

			if(!remainders[i].valid())
			{
				bvalid = false;
				break;
			}
		}

		if(!bvalid)
		{
			break;
		}

		bcontinue = false;
		for(int i=0; i<rangeDim; ++i)
		{
			if(oldRemainders[i].widthRatio(remainders[i]) <= DC_THRESHOLD_IMPROV)
			{
				bcontinue = true;
				break;
			}
		}
	}

	if(!bvalid)
	{
		boundary_intersected.clear();
		return -1;	// no intersection is detected
	}

	for(int i=0; i<rangeDim; ++i)
	{
		flowpipe.tms[i].remainder = remainders[i];
	}


	// the Horner forms of p(T(x))
	std::vector<HornerForm> objHF;

	std::vector<Interval> eval_remainders;

	for(int i=0; i<pcs.size(); ++i)
	{
		TaylorModel tmTemp;
		pcs[i].hf.insert(tmTemp, flowpipe, flowpipePolyRange, domain, cutoff_threshold);

		HornerForm hf;
		Interval remainder;
		tmTemp.toHornerForm(hf, remainder);
		objHF.push_back(hf);
		eval_remainders.push_back(remainder);
	}

	Interval intTime = domain[0];

	bvalid = true;
	bcontinue = true;

	// 3: domain contraction
	for(; bcontinue; )
	{
		std::vector<Interval> oldDomain = domain;

		// contract the domain
		for(int i=0; i<domainDim; ++i)
		{
			Interval newInt = domain[i];
			std::vector<bool> localNeeded = bNeeded;
			int localCounter = counter;

			newInt.width(W);

			// search an approximation for the lower bound
			for(; W >= DC_THRESHOLD_SEARCH;)
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);

				for(int j=0; j<pcs.size(); ++j)
				{
					if(localNeeded[j])
					{
						std::vector<Interval> newDomain = domain;
						newDomain[i] = intLeft;

						Interval intTemp;
						objHF[j].intEval(intTemp, newDomain);
						intTemp += eval_remainders[j];

						if(intTemp > pcs[j].B)
						{
							// no intersection on the left half
							newInt = intRight;
							newInt.width(W);
							break;
						}
						else if(intTemp.smallereq(pcs[j].B))
						{
							// do not need to apply domain contraction w.r.t. the current constraint
							newInt = intLeft;
							newInt.width(W);
							localNeeded[j] = false;
							++localCounter;
						}
						else
						{
							// refine the interval
							newInt = intLeft;
							newInt.width(W);

							continue;
						}
					}
				}

				if(localCounter == pcs.size())
				{
					break;
				}
			}

			// set the lower bound
			Interval Inf;
			newInt.inf(Inf);
			domain[i].setInf(Inf);

			newInt = domain[i];

			localNeeded = bNeeded;
			localCounter = counter;

			newInt.width(W);

			// search an approximation for the upper bound
			for(; W >= DC_THRESHOLD_SEARCH;)
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);

				for(int j=0; j<pcs.size(); ++j)
				{
					if(localNeeded[j])
					{
						std::vector<Interval> newDomain = domain;
						newDomain[i] = intRight;

						Interval intTemp;
						objHF[j].intEval(intTemp, newDomain);
						intTemp += eval_remainders[j];

						if(intTemp > pcs[j].B)
						{
							// no intersection on the right half
							newInt = intLeft;
							newInt.width(W);
							break;
						}
						else if(intTemp.smallereq(pcs[j].B))
						{
							// do not need to apply domain contraction w.r.t. the current constraint
							newInt = intRight;
							newInt.width(W);
							localNeeded[j] = false;
							++localCounter;
						}
						else
						{
							// refine the interval
							newInt = intRight;
							newInt.width(W);
							continue;
						}
					}
				}

				if(localCounter == pcs.size())
				{
					break;
				}
			}

			Interval Sup;
			newInt.sup(Sup);
			domain[i].setSup(Sup);	// set the upper bound

			if(!domain[i].valid())
			{
				bvalid = false;
				break;
			}
		}

		if(!bvalid)
		{
			break;
		}

		bcontinue = false;
		for(int i=0; i<domainDim; ++i)
		{
			if(oldDomain[i].widthRatio(domain[i]) <= DC_THRESHOLD_IMPROV)
			{
				bcontinue = true;
				break;
			}
		}

		if(bcontinue)
		{
			objHF.clear();
			eval_remainders.clear();

			flowpipe.polyRange(flowpipePolyRange, domain);

			for(int i=0; i<pcs.size(); ++i)
			{
				TaylorModel tmTemp;

				pcs[i].hf.insert(tmTemp, flowpipe, flowpipePolyRange, domain, cutoff_threshold);

				HornerForm hf;
				Interval remainder;
				tmTemp.toHornerForm(hf, remainder);
				objHF.push_back(hf);
				eval_remainders.push_back(remainder);
			}
		}
	}

	if(!bvalid)
	{
		boundary_intersected.clear();
		return -1;
	}

	if(intTime != domain[0])
	{
		return 2;
	}
	else
	{
		return 1;
	}
}
/*
int contract_interval_arithmetic(TaylorModel & tm, std::vector<Interval> & domain, const Interval & intMax, const Interval & intMin, const Interval & cutoff_threshold)
{
	int domainDim = domain.size();

	bool bvalid = true;
	bool bcontinue = true;
	Interval W;
	Interval intZero;

	bool bIntersectMax, bIntersectMin;

	HornerForm hf_expansion;
	tm.expansion.toHornerForm(hf_expansion);

	Interval tmPolyRange;
	hf_expansion.intEval(tmPolyRange, domain);

	Interval intRange = tmPolyRange + tm.remainder;

	if(intRange.smallereq(intMin))
	{
		return -2;			// the whole flowpipe is smaller than min
	}

	if(intMax.smallereq(intRange))
	{
		return -1;			// the whole flowpipe is larger than max
	}

	if(intRange.smallereq(intMax))
	{
		bIntersectMax = false;
	}
	else
	{
		bIntersectMax = true;
	}

	if(intMin.smallereq(intRange))
	{
		bIntersectMin = false;
	}
	else
	{
		bIntersectMin = true;
	}

	if(!bIntersectMax && !bIntersectMin)
	{
		return 0;			// the whole flowpipe is between the max and min values
	}

	// we first contract the remainder part
	if(bIntersectMax)
	{
		double diff_max = intMax.inf() - tmPolyRange.sup();

		if(diff_max <= 0)
		{
			tm.remainder.setSup(0);
		}
		else if(tm.remainder.sup() > diff_max)
		{
			tm.remainder.setSup(diff_max);
		}
	}

	if(bIntersectMin)
	{
		double diff_min = intMin.sup() - tmPolyRange.inf();

		if(diff_min >= 0)
		{
			tm.remainder.setInf(0);
		}
		if(tm.remainder.inf() < diff_min)
		{
			tm.remainder.setInf(diff_min);
		}
	}


	// we then contract the domain
	bool bContinue = true;
	for(; bContinue; )
	{
		std::vector<Interval> oldDomain = domain;

		for(int i=1; i<domainDim; ++i)		// time dimenion is neglected
		{
			Interval newInt = domain[i];
			newInt.width(W);

			// search an approximation for the lower bound
			for(; W.sup() >= DC_THRESHOLD_SEARCH;)
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);
				domain_test[i] = intLeft;

				Interval I;
				hf_expansion.intEval(I, domain);
				I += tm.remainder;

				if((bIntersectMax && I > intMax) || (bIntersectMin && intMin > I))
				{
					newInt = intRight;
					newInt.width(W);
				}
				else if((bIntersectMax && I.smallereq(intMax)) && (bIntersectMin && intMin.smallereq(I)))
				{
					newInt = intLeft;
					break;
				}
				else
				{
					newInt = intLeft;
					newInt.width(W);
				}
			}

			Interval Inf;
			newInt.inf(Inf);
			domain[i].setInf(Inf);

			newInt = domain[i];

			newInt.width(W);

			// search an approximation for the upper bound
			for(; W.sup() >= DC_THRESHOLD_SEARCH;)
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);
				domain[i] = intRight;

				Interval I;
				hf_expansion.intEval(I, domain);
				I += tm.remainder;

				if((bIntersectMax && I > intMax) || (bIntersectMin && intMin > I))
				{
					newInt = intLeft;
					newInt.width(W);
				}
				else if((bIntersectMax && I.smallereq(intMax)) && (bIntersectMin && intMin.smallereq(I)))
				{
					newInt = intRight;
					break;
				}
				else
				{
					newInt = intRight;
					newInt.width(W);
				}
			}

			Interval Sup;
			newInt.sup(Sup);
			domain[i].setSup(Sup);

			if(!domain[i].valid())
			{
				if(bIntersectMax)
				{
					return -1;
				}
				else
				{
					return -2;
				}
			}
		}

		bContinue = false;
		for(int i=0; i<domainDim; ++i)
		{
			if(oldDomain[i].widthRatio(domain[i]) <= DC_THRESHOLD_IMPROV)
			{
				bContinue = true;
				break;
			}
		}
	}

	if(bIntersectMax && bIntersectMin)
	{
		return 3;
	}
	else if(bIntersectMax)
	{
		return 2;
	}
	else
	{
		return 1;
	}
}
*/

int contract_interval_arithmetic(TaylorModel & tm_mid, std::vector<Interval> & domain_mid,
		TaylorModel & tm_max, std::vector<Interval> & domain_max, TaylorModel & tm_min, std::vector<Interval> & domain_min,
		const TaylorModel & tm, const std::vector<Interval> & domain, const Interval & intMax, const Interval & intMin, const Interval & cutoff_threshold)
{
	int domainDim = domain.size();

	bool bvalid = true;
	bool bcontinue = true;
	Interval W;
	Interval intZero;

	bool bIntersectMax, bIntersectMin;

	HornerForm hf_expansion;
	tm.expansion.toHornerForm(hf_expansion);

	Interval tmPolyRange;
	hf_expansion.intEval(tmPolyRange, domain);

	Interval intRange = tmPolyRange + tm.remainder;

	if(intRange.smallereq(intMin))
	{
		return -2;			// the whole flowpipe is smaller than min
	}

	if(intMax.smallereq(intRange))
	{
		return -1;			// the whole flowpipe is larger than max
	}

	if(intRange.smallereq(intMax))
	{
		bIntersectMax = false;
	}
	else
	{
		bIntersectMax = true;
	}

	if(intMin.smallereq(intRange))
	{
		bIntersectMin = false;
	}
	else
	{
		bIntersectMin = true;
	}

	if(!bIntersectMax && !bIntersectMin)
	{
		return 0;			// the whole flowpipe is between the max and min values
	}

	tm_mid = tm;
	domain_mid = domain;

	if(bIntersectMax)
	{
		tm_max = tm;
		domain_max = domain;
	}

	if(bIntersectMin)
	{
		tm_min = tm;
		domain_min = domain;
	}

	// we first contract the remainder part
	if(bIntersectMax)
	{
		double diff_max = intMax.sup() - tmPolyRange.sup();

		if(diff_max <= 0)
		{
			tm_mid.remainder.setSup(0);
		}
		else if(tm_mid.remainder.sup() > diff_max)
		{
			tm_mid.remainder.setSup(diff_max);
		}

		diff_max = intMax.inf() - tmPolyRange.inf();

		if(diff_max >= 0)
		{
			tm_max.remainder.setInf(0);
		}
		else if(tm_max.remainder.inf() < diff_max)
		{
			tm_max.remainder.setInf(diff_max);
		}
	}

	if(bIntersectMin)
	{
		double diff_min = intMin.inf() - tmPolyRange.inf();

		if(diff_min >= 0)
		{
			tm_mid.remainder.setInf(0);
		}
		else if(tm_mid.remainder.inf() < diff_min)
		{
			tm_mid.remainder.setInf(diff_min);
		}

		diff_min = intMin.sup() - tmPolyRange.sup();

		if(diff_min <= 0)
		{
			tm_min.remainder.setSup(0);
		}
		else if(tm_min.remainder.sup() > diff_min)
		{
			tm_min.remainder.setSup(diff_min);
		}
	}

/*
	for(int i=0; i<domain.size(); ++i)
	{
		domain[i].dump(stdout);
		printf("\t");
	}

	printf("\n");
*/
	// domain contraction for intMin <= x <= intMax
	bool bContinue = true;
	for(; bContinue; )
	{
		std::vector<Interval> oldDomain = domain_mid;
		std::vector<Interval> domain_test = domain_mid;

		for(int i=1; i<domainDim; ++i)		// time dimenion is neglected
		{
			Interval newInt = domain_test[i];
			newInt.width(W);

			// search an approximation for the lower bound
			for(; W.sup() >= DC_THRESHOLD_SEARCH; )
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);
				domain_test[i] = intLeft;

				Interval I;
				hf_expansion.intEval(I, domain_test);
				I += tm_mid.remainder;

				if((bIntersectMax && I > intMax) || (bIntersectMin && intMin > I))
				{
					newInt = intRight;
					newInt.width(W);
				}
				else if((bIntersectMax && I.smallereq(intMax)) && (bIntersectMin && intMin.smallereq(I)))
				{
					newInt = intLeft;
					break;
				}
				else
				{
					newInt = intLeft;
					newInt.width(W);
				}
			}

			Interval Inf;
			newInt.inf(Inf);

			domain_mid[i].setInf(Inf);

			newInt = domain_mid[i];

			newInt.width(W);

			// search an approximation for the upper bound
			for(; W.sup() >= DC_THRESHOLD_SEARCH;)
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);
				domain_test[i] = intRight;

				Interval I;
				hf_expansion.intEval(I, domain_test);
				I += tm_mid.remainder;

				if((bIntersectMax && I > intMax) || (bIntersectMin && intMin > I))
				{
					newInt = intLeft;
					newInt.width(W);
				}
				else if((bIntersectMax && I.smallereq(intMax)) && (bIntersectMin && intMin.smallereq(I)))
				{
					newInt = intRight;
					break;
				}
				else
				{
					newInt = intRight;
					newInt.width(W);
				}
			}

			Interval Sup;
			newInt.sup(Sup);
			domain_mid[i].setSup(Sup);

			if(!domain_mid[i].valid())
			{
				if(bIntersectMax)
				{
					return -1;
				}
				else
				{
					return -2;
				}
			}

			domain_test[i] = domain_mid[i];
		}

		bContinue = false;
		for(int i=0; i<domainDim; ++i)
		{
			if(oldDomain[i].widthRatio(domain_mid[i]) <= DC_THRESHOLD_IMPROV)
			{
				bContinue = true;
				break;
			}
		}
	}
/*
	for(int i=0; i<domain_mid.size(); ++i)
	{
		domain_mid[i].dump(stdout);
		printf("\t");
	}

	printf("\n");
*/
	if(bIntersectMax)
	{
		// domain contraction for x >= intMax
		bContinue = true;
		for(; bContinue; )
		{
			std::vector<Interval> oldDomain = domain_max;
			std::vector<Interval> domain_test = domain_max;

			for(int i=1; i<domainDim; ++i)		// time dimenion is neglected
			{
				Interval newInt = domain_test[i];
				newInt.width(W);

				// search an approximation for the lower bound
				for(; W.sup() >= DC_THRESHOLD_SEARCH;)
				{
					Interval intLeft;
					Interval intRight;
					newInt.split(intLeft, intRight);
					domain_test[i] = intLeft;

					Interval I;
					hf_expansion.intEval(I, domain_test);
					I += tm_max.remainder;

					if(I < intMax)
					{
						newInt = intRight;
						newInt.width(W);
					}
					else if(intMax.smallereq(I))
					{
						newInt = intLeft;
						break;
					}
					else
					{
						newInt = intLeft;
						newInt.width(W);
					}
				}

				Interval Inf;
				newInt.inf(Inf);
				domain_max[i].setInf(Inf);

				newInt = domain_max[i];

				newInt.width(W);

				// search an approximation for the upper bound
				for(; W.sup() >= DC_THRESHOLD_SEARCH;)
				{
					Interval intLeft;
					Interval intRight;
					newInt.split(intLeft, intRight);
					domain_test[i] = intRight;

					Interval I;
					hf_expansion.intEval(I, domain_test);
					I += tm_max.remainder;

					if(I < intMax)
					{
						newInt = intLeft;
						newInt.width(W);
					}
					else if(intMax.smallereq(I))
					{
						newInt = intRight;
						break;
					}
					else
					{
						newInt = intRight;
						newInt.width(W);
					}
				}

				Interval Sup;
				newInt.sup(Sup);
				domain_max[i].setSup(Sup);

				if(!domain_max[i].valid())
				{
					if(bIntersectMin)
					{
						return 1;
					}
					else
					{
						return 0;
					}
				}

				domain_test[i] = domain_max[i];
			}

			bContinue = false;
			for(int i=0; i<domainDim; ++i)
			{
				if(oldDomain[i].widthRatio(domain_max[i]) <= DC_THRESHOLD_IMPROV)
				{
					bContinue = true;
					break;
				}
			}
		}
	}
/*
	for(int i=0; i<domain_max.size(); ++i)
	{
		domain_max[i].dump(stdout);
		printf("\t");
	}

	printf("\n");
*/
	if(bIntersectMin)
	{
		// domain contraction for x <= intMin
		bContinue = true;
		for(; bContinue; )
		{
			std::vector<Interval> oldDomain = domain_min;
			std::vector<Interval> domain_test = domain_min;

			for(int i=1; i<domainDim; ++i)		// time dimenion is neglected
			{
				Interval newInt = domain_test[i];
				newInt.width(W);

				// search an approximation for the lower bound
				for(; W.sup() >= DC_THRESHOLD_SEARCH;)
				{
					Interval intLeft;
					Interval intRight;
					newInt.split(intLeft, intRight);
					domain_test[i] = intLeft;

					Interval I;
					hf_expansion.intEval(I, domain_test);
					I += tm_min.remainder;

					if(I > intMin)
					{
						newInt = intRight;
						newInt.width(W);
					}
					else if(I.smallereq(intMin))
					{
						newInt = intLeft;
						break;
					}
					else
					{
						newInt = intLeft;
						newInt.width(W);
					}
				}

				Interval Inf;
				newInt.inf(Inf);
				domain_min[i].setInf(Inf);

				newInt = domain_min[i];

				newInt.width(W);

				// search an approximation for the upper bound
				for(; W.sup() >= DC_THRESHOLD_SEARCH;)
				{
					Interval intLeft;
					Interval intRight;
					newInt.split(intLeft, intRight);
					domain_test[i] = intRight;

					Interval I;
					hf_expansion.intEval(I, domain_test);
					I += tm_min.remainder;

					if(I > intMin)
					{
						newInt = intLeft;
						newInt.width(W);
					}
					else if(I.smallereq(intMin))
					{
						newInt = intRight;
						break;
					}
					else
					{
						newInt = intRight;
						newInt.width(W);
					}
				}

				Interval Sup;
				newInt.sup(Sup);
				domain_min[i].setSup(Sup);

				if(!domain_min[i].valid())
				{
					if(bIntersectMax)
					{
						return 2;
					}
					else
					{
						return 0;
					}
				}

				domain_test[i] = domain_min[i];
			}

			bContinue = false;
			for(int i=0; i<domainDim; ++i)
			{
				if(oldDomain[i].widthRatio(domain_min[i]) <= DC_THRESHOLD_IMPROV)
				{
					bContinue = true;
					break;
				}
			}
		}
	}
/*
	for(int i=0; i<domain_min.size(); ++i)
	{
		domain_min[i].dump(stdout);
		printf("\t");
	}

	printf("\n");
*/
	if(bIntersectMax && bIntersectMin)
	{
		return 3;
	}
	else if(bIntersectMax)
	{
		return 2;
	}
	else
	{
		return 1;
	}
}











int contract_interval_arithmetic(TaylorModelVec & flowpipe, std::vector<Interval> & domain, const Polyhedron & inv, std::vector<bool> & boundary_intersected)
{
	int rangeDim = flowpipe.tms.size();
	int domainDim = domain.size();

	boundary_intersected.clear();

	// contract the remainder firstly
	bool bvalid = true;
	bool bcontinue = true;
	Interval W;
	Interval intZero;

	std::vector<bool> bNeeded;
	int counter = 0;

	for(int i=0; i<inv.constraints.size(); ++i)
	{
		bNeeded.push_back(true);
	}

	boundary_intersected.clear();

	std::vector<Interval> remainders;
	for(int i=0; i<rangeDim; ++i)
	{
		remainders.push_back(flowpipe.tms[i].remainder);
	}

	std::vector<Interval> flowpipePolyRange;
	flowpipe.polyRange(flowpipePolyRange, domain);

	std::vector<Interval> intVecTemp = flowpipePolyRange;

	// 1: we check the intersection with every constraint
	for(int i=0; i<rangeDim; ++i)
	{
		intVecTemp[i] = flowpipePolyRange[i] + remainders[i];
	}

	for(int i=0; i<inv.constraints.size(); ++i)
	{
		Interval intTemp;

		for(int j=0; j<inv.constraints[i].A.size(); ++j)
		{
			intTemp += inv.constraints[i].A[j] * intVecTemp[j];
		}

		if(intTemp > inv.constraints[i].B)
		{
			// no intersection on the left half
			bvalid = false;
			break;
		}
		else if(intTemp.smallereq(inv.constraints[i].B))
		{
			// do not need to apply domain contraction w.r.t. the current constraint
			boundary_intersected.push_back(false);
			bNeeded[i] = false;
			++counter;
		}
		else
		{
			boundary_intersected.push_back(true);
			bNeeded[i] = true;
			continue;
		}
	}

	if(!bvalid)
	{
		boundary_intersected.clear();
		return -1;	// no intersection is detected
	}
	else if(counter == inv.constraints.size())
	{
		return 0;	// no need to do contraction
	}


	// 2: remainder contraction
	for(; bcontinue; )
	{
		std::vector<Interval> oldRemainders = remainders;

		for(int i=0; i<rangeDim; ++i)
		{
			Interval newInt = remainders[i];
			std::vector<bool> localNeeded = bNeeded;
			int localCounter = counter;

			for(int k=0; k<rangeDim; ++k)
			{
				if(k != i)
				{
					intVecTemp[k] = flowpipePolyRange[k] + remainders[k];
				}
				else
				{
					intVecTemp[k] = flowpipePolyRange[k];
				}
			}

			newInt.width(W);

			// search an approximation for the lower bound
			for(; W >= DC_THRESHOLD_SEARCH;)
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);

				for(int j=0; j<inv.constraints.size(); ++j)
				{
					if(localNeeded[j])
					{
						Interval intTemp;
						std::vector<Interval> newIntVecTemp = intVecTemp;
						newIntVecTemp[i] += intLeft;

						for(int k=0; k<inv.constraints[j].A.size(); ++k)
						{
							intTemp += inv.constraints[j].A[k] * newIntVecTemp[k];
						}

						if(intTemp > inv.constraints[j].B)
						{
							// no intersection on the left half
							newInt = intRight;
							newInt.width(W);
							break;
						}
						else if(intTemp.smallereq(inv.constraints[j].B))
						{
							// do not need to apply domain contraction w.r.t. the current constraint
							newInt = intLeft;
							newInt.width(W);
							localNeeded[j] = false;
							++localCounter;
						}
						else
						{
							// refine the interval
							newInt = intLeft;
							newInt.width(W);

							continue;
						}
					}
				}

				if(localCounter == inv.constraints.size())
				{
					break;
				}
			}

			// set the lower bound
			Interval Inf;
			newInt.inf(Inf);
			remainders[i].setInf(Inf);

			newInt = remainders[i];
			newInt.width(W);

			localNeeded = bNeeded;
			localCounter = counter;

			// search an approximation for the upper bound
			for(; W >= DC_THRESHOLD_SEARCH;)
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);

				for(int j=0; j<inv.constraints.size(); ++j)
				{
					if(localNeeded[j])
					{
						Interval intTemp;
						std::vector<Interval> newIntVecTemp = intVecTemp;
						newIntVecTemp[i] += intRight;

						for(int k=0; k<inv.constraints[j].A.size(); ++k)
						{
							intTemp += inv.constraints[j].A[k] * newIntVecTemp[k];
						}

						if(intTemp > inv.constraints[j].B)
						{
							// no intersection on the right half
							newInt = intLeft;
							newInt.width(W);
							break;
						}
						else if(intTemp.smallereq(inv.constraints[j].B))
						{
							// do not need to apply domain contraction w.r.t. the current constraint
							newInt = intRight;
							newInt.width(W);
							localNeeded[j] = false;
							++localCounter;
						}
						else
						{
							// refine the interval
							newInt = intRight;
							newInt.width(W);
							continue;
						}
					}
				}

				if(localCounter == inv.constraints.size())
				{
					break;
				}
			}

			Interval Sup;
			newInt.sup(Sup);
			remainders[i].setSup(Sup);	// set the upper bound

			if(!remainders[i].valid())
			{
				bvalid = false;
				break;
			}
		}

		if(!bvalid)
		{
			break;
		}

		bcontinue = false;
		for(int i=0; i<rangeDim; ++i)
		{
			if(oldRemainders[i].widthRatio(remainders[i]) <= DC_THRESHOLD_IMPROV)
			{
				bcontinue = true;
				break;
			}
		}
	}

	if(!bvalid)
	{
		boundary_intersected.clear();
		return -1;	// no intersection is detected
	}

	for(int i=0; i<rangeDim; ++i)
	{
		flowpipe.tms[i].remainder = remainders[i];
	}


	// the Horner forms of p(T(x))
	std::vector<HornerForm> objHF;
	std::vector<Interval> eval_remainders;

	for(int i=0; i<inv.constraints.size(); ++i)
	{
		TaylorModel tmTemp;

		for(int j=0; j<inv.constraints[i].A.size(); ++j)
		{
			TaylorModel tmTemp2;
			flowpipe.tms[j].mul(tmTemp2, inv.constraints[i].A[j]);
			tmTemp.add_assign(tmTemp2);
		}

		HornerForm hf;
		Interval remainder;
		tmTemp.toHornerForm(hf, remainder);

		objHF.push_back(hf);
		eval_remainders.push_back(remainder);
	}

	Interval intTime = domain[0];

	bvalid = true;
	bcontinue = true;

	// 3: domain contraction

	for(; bcontinue; )
	{
		std::vector<Interval> oldDomain = domain;

		// contract the domain
		for(int i=0; i<domainDim; ++i)
		{
			Interval newInt = domain[i];
			std::vector<bool> localNeeded = bNeeded;
			int localCounter = counter;

			newInt.width(W);

			// search an approximation for the lower bound
			for(; W >= DC_THRESHOLD_SEARCH;)
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);

				for(int j=0; j<inv.constraints.size(); ++j)
				{
					if(localNeeded[j])
					{
						std::vector<Interval> newDomain = domain;
						newDomain[i] = intLeft;

						Interval intTemp;
						objHF[j].intEval(intTemp, newDomain);
						intTemp += eval_remainders[j];

						if(intTemp > inv.constraints[j].B)
						{
							// no intersection on the left half
							newInt = intRight;
							newInt.width(W);
							break;
						}
						else if(intTemp.smallereq(inv.constraints[j].B))
						{
							// do not need to apply domain contraction w.r.t. the current constraint
							newInt = intLeft;
							newInt.width(W);
							localNeeded[j] = false;
							++localCounter;
						}
						else
						{
							// refine the interval
							newInt = intLeft;
							newInt.width(W);

							continue;
						}
					}
				}

				if(localCounter == inv.constraints.size())
				{
					break;
				}
			}

			// set the lower bound
			Interval Inf;
			newInt.inf(Inf);
			domain[i].setInf(Inf);

			newInt = domain[i];

			localNeeded = bNeeded;
			localCounter = counter;

			newInt.width(W);

			// search an approximation for the upper bound
			for(; W >= DC_THRESHOLD_SEARCH;)
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);

				for(int j=0; j<inv.constraints.size(); ++j)
				{
					if(localNeeded[j])
					{
						std::vector<Interval> newDomain = domain;
						newDomain[i] = intRight;

						Interval intTemp;
						objHF[j].intEval(intTemp, newDomain);
						intTemp += eval_remainders[j];

						if(intTemp > inv.constraints[j].B)
						{
							// no intersection on the right half
							newInt = intLeft;
							newInt.width(W);
							break;
						}
						else if(intTemp.smallereq(inv.constraints[j].B))
						{
							// do not need to apply domain contraction w.r.t. the current constraint
							newInt = intRight;
							newInt.width(W);
							localNeeded[j] = false;
							++localCounter;
						}
						else
						{
							// refine the interval
							newInt = intRight;
							newInt.width(W);
							continue;
						}
					}
				}

				if(localCounter == inv.constraints.size())
				{
					break;
				}
			}

			Interval Sup;
			newInt.sup(Sup);
			domain[i].setSup(Sup);	// set the upper bound

			if(!domain[i].valid())
			{
				bvalid = false;
				break;
			}
		}

		if(!bvalid)
		{
			break;
		}

		bcontinue = false;
		for(int i=0; i<domainDim; ++i)
		{
			if(oldDomain[i].widthRatio(domain[i]) <= DC_THRESHOLD_IMPROV)
			{
				bcontinue = true;
				break;
			}
		}
	}

	if(!bvalid)
	{
		boundary_intersected.clear();
		return -1;
	}

	if(intTime != domain[0])
	{
		return 2;
	}
	else
	{
		return 1;
	}
}

int contract_remainder(const std::vector<Interval> & polyRange, std::vector<Interval> & remainders, const std::vector<HornerForm> & hfs, const std::vector<Interval> & b)
{
	bool bvalid = true;
	bool bcontinue = true;
	Interval W;
	Interval intZero;

	std::vector<bool> bNeeded;
	int counter = 0;

	for(int i=0; i<hfs.size(); ++i)
	{
		bNeeded.push_back(true);
	}

	int rangeDim = polyRange.size();

	std::vector<Interval> intVecTemp = polyRange;
	intVecTemp.insert(intVecTemp.begin(), intZero);		// range of the dummy time variable

	// 1: we check the intersection with every constraint
	for(int i=0; i<rangeDim; ++i)
	{
		intVecTemp[i+1] = polyRange[i] + remainders[i];
	}

	for(int i=0; i<hfs.size(); ++i)
	{
		Interval intTemp;

		hfs[i].intEval(intTemp, intVecTemp);

		if(intTemp > b[i])
		{
			// no intersection on the left half
			bvalid = false;
			break;
		}
		else if(intTemp.smallereq(b[i]))
		{
			// do not need to apply domain contraction w.r.t. the current constraint
			bNeeded[i] = false;
			++counter;
		}
		else
		{
			bNeeded[i] = true;
			continue;
		}
	}

	if(!bvalid)
	{
		return -1;	// no intersection is detected
	}
	else if(counter == hfs.size())
	{
		return 0;	// no need to do contraction
	}

	// 2: contract the remainder
	for(; bcontinue; )
	{
		std::vector<Interval> oldRemainders = remainders;

		for(int i=0; i<rangeDim; ++i)
		{
			Interval newInt = remainders[i];
			std::vector<bool> localNeeded = bNeeded;
			int localCounter = counter;

			for(int k=0; k<rangeDim; ++k)
			{
				if(k != i)
				{
					intVecTemp[k+1] = polyRange[k] + remainders[k];
				}
				else
				{
					intVecTemp[k+1] = polyRange[k];
				}
			}

			newInt.width(W);

			// search an approximation for the lower bound
			for(; W >= DC_THRESHOLD_SEARCH;)
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);

				for(int j=0; j<hfs.size(); ++j)
				{
					if(localNeeded[j])
					{
						Interval intTemp;
						std::vector<Interval> newIntVecTemp = intVecTemp;
						newIntVecTemp[i+1] += intLeft;

						hfs[j].intEval(intTemp, newIntVecTemp);

						if(intTemp > b[j])
						{
							// no intersection on the left half
							newInt = intRight;
							newInt.width(W);
							break;
						}
						else if(intTemp.smallereq(b[j]))
						{
							// do not need to apply domain contraction w.r.t. the current constraint
							newInt = intLeft;
							newInt.width(W);
							localNeeded[j] = false;
							++localCounter;
						}
						else
						{
							// refine the interval
							newInt = intLeft;
							newInt.width(W);

							continue;
						}
					}
				}

				if(localCounter == hfs.size())
				{
					break;
				}
			}

			// set the lower bound
			Interval Inf;
			newInt.inf(Inf);
			remainders[i].setInf(Inf);

			newInt = remainders[i];
			newInt.width(W);

			localNeeded = bNeeded;
			localCounter = counter;

			// search an approximation for the upper bound
			for(; W >= DC_THRESHOLD_SEARCH;)
			{
				Interval intLeft;
				Interval intRight;
				newInt.split(intLeft, intRight);

				for(int j=0; j<hfs.size(); ++j)
				{
					if(localNeeded[j])
					{
						Interval intTemp;
						std::vector<Interval> newIntVecTemp = intVecTemp;
						newIntVecTemp[i+1] += intRight;

						hfs[j].intEval(intTemp, newIntVecTemp);

						if(intTemp > b[j])
						{
							// no intersection on the right half
							newInt = intLeft;
							newInt.width(W);
							break;
						}
						else if(intTemp.smallereq(b[j]))
						{
							// do not need to apply domain contraction w.r.t. the current constraint
							newInt = intRight;
							newInt.width(W);
							localNeeded[j] = false;
							++localCounter;
						}
						else
						{
							// refine the interval
							newInt = intRight;
							newInt.width(W);
							continue;
						}
					}
				}

				if(localCounter == hfs.size())
				{
					break;
				}
			}

			Interval Sup;
			newInt.sup(Sup);
			remainders[i].setSup(Sup);	// set the upper bound

			if(!remainders[i].valid())
			{
				bvalid = false;
				break;
			}
		}

		if(!bvalid)
		{
			break;
		}

		bcontinue = false;
		for(int i=0; i<rangeDim; ++i)
		{
			if(oldRemainders[i].widthRatio(remainders[i]) <= DC_THRESHOLD_IMPROV)
			{
				bcontinue = true;
				break;
			}
		}
	}

	if(!bvalid)
	{
		return -1;	// no intersection is detected
	}
	else
	{
		return 1;
	}
}

void gridBox(std::list<std::vector<Interval> > & grids, const std::vector<Interval> & box, const int num)
{
	grids.clear();
	grids.push_back(box);

	for(int i=0; i<box.size(); ++i)
	{
		std::list<std::vector<Interval> >::iterator gridIter;
		std::list<std::vector<Interval> > newGrids;

		for(; grids.size() > 0;)
		{
			gridIter = grids.begin();

			std::list<Interval> queue;
			(*gridIter)[i].split(queue, num);

			std::list<Interval>::iterator iterComponent = queue.begin();
			for(; iterComponent != queue.end(); ++iterComponent)
			{
				std::vector<Interval> tmpBox = *gridIter;
				tmpBox[i] = *iterComponent;
				newGrids.push_back(tmpBox);
			}

			grids.pop_front();
		}

		grids = newGrids;
	}
}

void exp_int_mat(Interval_matrix & result_ts, Interval_matrix & result_rem, const Interval_matrix & A, const int order)
{
	int dim = A.cols();

	Interval_matrix identity(dim, dim), zero(dim, dim);
	Interval intOne(1);

	for(int i=0; i<dim; ++i)
	{
		identity.set(intOne, i, i);
	}

	if(order == 0)
	{
		result_ts = identity;
		result_rem = zero;
	}
	else
	{
		result_ts = identity;

		for(int k=order; k>0; --k)
		{
			result_ts.div_assign(k);
			result_ts *= A;
			result_ts += identity;
		}

		Interval_matrix matTemp = A;
		matTemp.pow_assign(order+1);
		matTemp.mul_assign(factorial_rec[order+1]);

		double max_A = A.max_norm();
		Interval intTemp(max_A);
		intTemp.exp_assign();

		double maxNorm = intTemp.mag();
		Interval remainder(-maxNorm, maxNorm);

		Interval_matrix matRemainder(dim, dim);
		for(int i=0; i<dim; ++i)
		{
			for(int j=0; j<dim; ++j)
			{
				matRemainder.set(remainder, i, j);
			}
		}

		matTemp *= matRemainder;
		result_rem = matTemp;
	}
}

void int_exp_int_mat(Interval_matrix & result_ts, Interval_matrix & result_rem, const Interval_matrix & A, const double step, const int order)
{
	int dim = A.cols();

	Interval_matrix identity(dim, dim), zero(dim, dim);
	Interval intOne(1);

	for(int i=0; i<dim; ++i)
	{
		identity.set(intOne, i, i);
	}

	if(order == 0)
	{
		result_ts = identity;
		result_rem = zero;
	}
	else
	{
		Interval_matrix R(dim, dim);
		Interval intStep(step);

		for(int i=0; i<dim; ++i)
		{
			R.set(intStep, i, i);
		}

		result_ts = R;
		Interval_matrix mA = A;
		mA.mul_assign(-1.0);

		for(int k=order+1; k>1; --k)
		{
			result_ts.div_assign(k);
			result_ts *= mA;
			result_ts += R;
		}

		Interval_matrix matTemp = mA;
		matTemp.pow_assign(order+1);
		matTemp.mul_assign(factorial_rec[order+2]);
		matTemp.mul_assign(step);

		double max_mA = mA.max_norm();
		Interval intTemp(max_mA);
		intTemp.exp_assign();

		double maxNorm = intTemp.mag();
		Interval remainder(-maxNorm, maxNorm);

		Interval_matrix matRemainder(dim, dim);
		for(int i=0; i<dim; ++i)
		{
			for(int j=0; j<dim; ++j)
			{
				matRemainder.set(remainder, i, j);
			}
		}

		matTemp *= matRemainder;
		result_rem = matTemp;
	}
}

void compute_int_mat_pow(std::vector<Interval_matrix> & result, const Interval_matrix & A, const int order)
{
	int d = A.rows();
	Interval intOne(1);

	Interval_matrix identity(d, d);

	// identity matrix
	for(int i=0; i<d; ++i)
	{
		identity.set(intOne, i, i);
	}

	std::vector<bool> pow_A_computed;

	Interval_matrix imEmpty;

	for(int i=0; i<=order; ++i)
	{
		pow_A_computed.push_back(false);
		result.push_back(imEmpty);
	}

	pow_A_computed[0] = true;
	result[0] = identity;

	if(order < 1)
	{
		return;
	}

	pow_A_computed[1] = true;
	result[1] = A;

	for(int i=order; i>1; --i)
	{
		if(!pow_A_computed[i])
		{
			Interval_matrix temp = A;
			Interval_matrix temp2 = A;
			int pos_temp = 1;
			int pos_result = 1;

			for(int d=i-1; d > 0;)
			{
				if(d & 1)
				{
					pos_result += pos_temp;

					if(pow_A_computed[pos_result])
					{
						temp2 = result[pos_result];
					}
					else
					{
						temp2 *= temp;
						pow_A_computed[pos_result] = true;
						result[pos_result] = temp2;
					}
				}

				d >>= 1;

				if(d > 0)
				{
					pos_temp <<= 1;

					if(pow_A_computed[pos_temp])
					{
						temp = result[pos_temp];
					}
					else
					{
						temp *= temp;
						pow_A_computed[pos_temp] = true;
						result[pos_temp] = temp;
					}
				}
			}
		}
	}
}


void enlarge(std::vector<Interval> & box, const double e)
{
	for(int i=0; i<box.size(); ++i)
	{
		double newLo = box[i].inf() - e;
		double newUp = box[i].sup() + e;
		box[i].setInf(newLo);
		box[i].setSup(newUp);
	}
}

/*
void build_constraint_template(const int d)
{
	vector<Interval> intVecZero;
	Interval intZero, intOne(1), intMOne(-1);

	for(int i=0; i<d; ++i)
	{
		intVecZero.push_back(intZero);
	}

	for(int i=0; i<d; ++i)
	{
		vector<Interval> A = intVecZero;
		A[i] = intOne;

		LinearConstraint lc(A, intZero);
		constraint_template.push_back(lc);
	}

	for(int i=0; i<d; ++i)
	{
		vector<Interval> A = intVecZero;
		A[i] = intMOne;

		LinearConstraint lc(A, intZero);
		constraint_template.push_back(lc);
	}
}
*/

}








